/************************************************************************
 * FILE:  autom.c                                                       *
 *        Functions to handle weighted finite transducers and automata  *
 *
 * Author: Peter Rajcani, March 1995.
 ************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include "wft.h"

#define TRUE	1
#define FALSE	0
#define EPS     .1
#define LEFT	1
#define RIGHT	2
#define BIG	9999
#define INITIAL 0
#define FINAL   1

extern int	debug;
extern FILE	*fd;
extern int	rows, columns, scaling_flag, syn_flag, cutoff_flag;

int	quiet_mode = TRUE;

/***********************************
 * Functions defined in nutils.c :
 **********************************/

 extern float	*alloc_vect ();
 extern float   **alloc_matrix ();
 extern float   **alloc_sq_matrix ();
 extern int 	*alloc_int_array ();
 extern imgtype **alloc_image ();
 extern void    free_image ();
 extern void    vec_matix_mult ();
 extern float   compute_norm ();
 extern float   inner_prod ();
 extern void	copy_vect ();
 extern void	orthogonalize ();
 extern void 	free_matrix ();
 extern int	ludcmp ();
 extern void	lubksb ();
 extern void	print_vector ();
 extern void	print_matrix ();

 

/**********************************************
 * Function: read_automaton                   *
 * 	     Reads the input automaton        *
 **********************************************/

int read_automaton(infile, autom, ntapes) 
FILE	  *infile;
automaton *autom;
int	  ntapes;
{
 int state,i;
 int nstates, into;
 int label;
 float weight, greyness, norm;
 edgetype *ptr;
 char c;
 void check_automaton ();

  /* skip the comments */
  while ((c = fgetc (infile)) == '#')
	while ((c = fgetc (infile)) != '\n');
  ungetc (c, infile);
  
  autom->ntapes = ntapes;
  fscanf(infile, "%d", &(autom->states));
  printf("number of states = %d\n", autom->states);
  nstates = autom->states;

  for(i=0; i<autom->states; i++){
    fscanf(infile, "%f", &(autom->state[i].initial));
    if (debug) {
       fprintf(fd, "init_distr[%d] = %f\n",i,autom->state[i].initial);
       printf("initial_distr[%d] = %f\n",i,autom->state[i].initial);
    }
   }

  for(i=0; i<autom->states; i++) {
    fscanf(infile, "%f", &(autom->state[i].final)); 
    if (debug) {
       fprintf(fd, "final_distr[%d] = %f\n",i,autom->state[i].final);
       printf("final_distr[%d] = %f\n",i,autom->state[i].final);
    }
  }
  

  while( fscanf(infile, "%d",&state) >0  && state >= 0 ){
         if (state >= nstates) return (FALSE);
         if (debug) {
	    fprintf(fd, "state = %d:\n", state);
	    printf("state = %d:\n", state);
         }
	 autom->state[state].new=NULL;
	 autom->state[state].edges=NULL;

        if (ntapes == 1) {
	    
	    while( fscanf(infile, "%d", &label) , label != -1) {
		 ptr=(edgetype *) malloc(sizeof(edgetype));
		 if (label < 4)
		    ptr->label[0]= (unsigned char) (label + '0');
		 else if (label == 4)
		    ptr->label[0] = 'e';
                 else
		    return (FALSE);

		 fscanf(infile, "%d", &into);
		 if (into < 0 || into >= nstates) return (FALSE);
		 ptr->into=into;
		 fscanf(infile, "%f", &weight);
		 ptr->weight = weight;
                 if (debug) {
                    fprintf (fd,"edge from state %d to state %d, label: %c, weight: %f\n", 
		    state, ptr->into, ptr->label[0], ptr->weight);
                 printf ("edge from state %d to state %d, label: %c, weight: %f\n", 
		    state, ptr->into, ptr->label[0], ptr->weight);
                 }
		 ptr->next=autom->state[state].edges;
		 autom->state[state].edges=ptr;
	    }
        }

	if (ntapes == 2) {
	    c=fgetc(infile);
	    while( (label = fgetc(infile)) != '-') {
		 ptr=(edgetype *) malloc(sizeof(edgetype));
	         c=fgetc(infile);
                 if ((label >= '0' && label <= '3') || label == 'e') {
		    ptr->label[0]= (unsigned char) label;
		    ptr->label[1]= (unsigned char) fgetc(infile);
		 } else
		    return (FALSE);
		 
		 fscanf(infile, "%d", &into);
		 if (into < 0 || into >= nstates) return (FALSE);
		 ptr->into=into;
		 fscanf(infile, "%f", &weight);
		 ptr->weight = weight;
		 if (debug) {
                    fprintf (fd,"edge from state %d to state %d, labels: %c %c, weight: %f\n", 
		    state, ptr->into, ptr->label[0], ptr->label[1], ptr->weight);
                 printf ("edge from state %d to state %d, labels: %c %c, weight: %f\n", 
		 state, ptr->into, ptr->label[0], ptr->label[1], ptr->weight);
                 }
		 ptr->next=autom->state[state].edges;
		 autom->state[state].edges=ptr;
		 c = fgetc(infile);
	     }
	}
  }     
  if (ntapes == 2) check_automaton (autom);
  return (TRUE);
}

int read_color_automaton(infile, aut_red, aut_green, aut_blue, ntapes) 
FILE	  *infile;
automaton *aut_red, *aut_green, *aut_blue;
int	  ntapes;
{
 int state,i;
 int into, nstates;
 int label;
 float weight;
 edgetype *red_ptr, *green_ptr, *blue_ptr;
 char c;

  /* skip the comments */
  while ((c = fgetc (infile)) == '#')
	while ((c = fgetc (infile)) != '\n');
  ungetc (c, infile);
  
  aut_red->ntapes = ntapes;
  aut_green->ntapes = ntapes;
  aut_blue->ntapes = ntapes;
  fscanf(infile, "%d", &nstates);
  printf("number of states = %d\n", nstates);
  aut_blue->states = nstates;
  aut_red->states = nstates;
  aut_green->states = nstates;

  for(i=0; i<nstates; i++){
    fscanf(infile, "%f", &(aut_red->state[i].initial));
    if (debug == 2) {
       fprintf(fd, "init_distr[%d] = %f\n",i,aut_red->state[i].initial);
       printf("initial_distr[%d] = %f\n",i,aut_red->state[i].initial);
    }
   }

  for(i=0; i<nstates; i++){
    fscanf(infile, "%f", &(aut_green->state[i].initial));
    if (debug == 2) {
       fprintf(fd, "init_distr[%d] = %f\n",i,aut_green->state[i].initial);
       printf("initial_distr[%d] = %f\n",i,aut_green->state[i].initial);
    }
   }

  for(i=0; i<nstates; i++){
    fscanf(infile, "%f", &(aut_blue->state[i].initial));
    if (debug == 2) {
       fprintf(fd, "init_distr[%d] = %f\n",i,aut_blue->state[i].initial);
       printf("initial_distr[%d] = %f\n",i,aut_blue->state[i].initial);
    }
   }

  for(i=0; i<nstates; i++){
    fscanf(infile, "%f", &(aut_red->state[i].final));
    if (debug == 2) {
       fprintf(fd, "final_distr[%d] = %f\n",i,aut_red->state[i].final);
       printf("final_distr[%d] = %f\n",i,aut_red->state[i].final);
    }
   }

  for(i=0; i<nstates; i++){
    fscanf(infile, "%f", &(aut_green->state[i].final));
    if (debug == 2) {
       fprintf(fd, "final_distr[%d] = %f\n",i,aut_green->state[i].final);
       printf("final_distr[%d] = %f\n",i,aut_green->state[i].final);
    }
   }

  for(i=0; i<nstates; i++){
    fscanf(infile, "%f", &(aut_blue->state[i].final));
    if (debug) {
       fprintf(fd, "final_distr[%d] = %f\n",i,aut_blue->state[i].final);
       printf("final_distr[%d] = %f\n",i,aut_blue->state[i].final);
    }
   }

  

  while( fscanf(infile, "%d",&state) >0  && state >= 0 ){
	 if (state >= nstates ) return (FALSE);
         if (debug == 2) {
	    fprintf(fd, "state = %d:\n", state);
	    printf("state = %d:\n", state);
         }
	 aut_red->state[state].new=NULL;
	 aut_red->state[state].edges=NULL;
	 aut_green->state[state].new=NULL;
	 aut_green->state[state].edges=NULL;
	 aut_blue->state[state].new=NULL;
	 aut_blue->state[state].edges=NULL;

        if (ntapes == 1) {
	    
	    while( fscanf(infile, "%d", &label) , label != -1) {
		 red_ptr=(edgetype *) malloc(sizeof(edgetype));
		 green_ptr=(edgetype *) malloc(sizeof(edgetype));
		 blue_ptr=(edgetype *) malloc(sizeof(edgetype));
		 if (red_ptr == NULL || blue_ptr == NULL ||
                     green_ptr == NULL) {
			printf ("Memory allocation error"); exit (1);
                 }
		 if (label < 4) {
		    red_ptr->label[0]= (unsigned char) (label + '0');
		    green_ptr->label[0]= (unsigned char) (label + '0');
		    blue_ptr->label[0]= (unsigned char) (label + '0');
		 } else if (label == 4) {
		    red_ptr->label[0] = 'e';
		    green_ptr->label[0] = 'e';
		    blue_ptr->label[0] = 'e';
                 } else
		    return (FALSE);

		 fscanf(infile, "%d", &into);
		 if (into < 0 || into >= nstates) return (FALSE);
		 red_ptr->into=into;
		 green_ptr->into=into;
		 blue_ptr->into=into;
		 fscanf(infile, "%f", &weight);
		 red_ptr->weight = weight;
		 green_ptr->weight = weight;
		 blue_ptr->weight = weight;
/*
                 if (debug == 2) {
                    fprintf (fd,"edge from state %d to state %d, label: %c, weight: %f\n", 
		    state, ptr->into, ptr->label[0], ptr->weight);
                 printf ("edge from state %d to state %d, label: %c, weight: %f\n", 
		    state, ptr->into, ptr->label[0], ptr->weight);
                 }
*/
		 red_ptr->next=aut_red->state[state].edges;
		 aut_red->state[state].edges=red_ptr;
		 blue_ptr->next=aut_blue->state[state].edges;
		 aut_blue->state[state].edges=blue_ptr;
		 green_ptr->next=aut_green->state[state].edges;
		 aut_green->state[state].edges=green_ptr;
	    }
        }
    }
    return (TRUE);
}

/**************************************************************************
 *  Function:	write_automaton                                           *
 *              Write the output wfa into a file                          *
 **************************************************************************/

void write_automaton (outfile, autom, ntapes)
FILE       *outfile;
automaton  *autom;
int	   ntapes;
{
  int       i, state;
  edgetype  *ptr;
  float     greyness = 0.0, norm = 1.0;

 fprintf (outfile, "%d\n", autom->states);
 
  /* for the case of a wfa write out greyness and norm (dummy values) */

/*
  if (ntapes == 1) 
	fprintf (outfile, "%f %f\n", greyness, norm);
*/

  for (i=0; i<autom->states; i++)
	fprintf (outfile, "%f ", autom->state[i].initial);
  fprintf (outfile, "\n");
 
  for (i=0; i<autom->states; i++)
	fprintf (outfile, "%f ", autom->state[i].final);
  fprintf (outfile, "\n");

  for (state = 0; state < autom->states; state++) {
	fprintf (outfile, "%d\n", state);

	/* output edges */
	for (ptr = autom->state[state].edges; ptr != NULL; ptr = ptr->next) {
	     switch (ntapes) {
		case 1: if (ptr->label[0] != 'e')
                           fprintf (outfile,"%d ", ptr->label[0]-'0'); 
			else
			   fprintf (outfile,"%d ", 4);
			break;
		case 2: fprintf (outfile,"%c %c ", ptr->label[0], ptr->label[1]); 
			break;
	     }

	     fprintf (outfile, "%d %f\n", ptr->into, ptr->weight);
       	}
        /* mark the end of the edge list */

        switch (ntapes) {
	    case 1: fprintf (outfile, "%d \n", -1); break;
	    case 2: fprintf (outfile, "- \n"); break;
       }
    }
}


void display_aut (autom, fp)
automaton  *autom;
FILE	*fp;
{
  int       i, state;
  edgetype  *ptr;

if (fp == NULL) {
 	printf ("Number of states: %d\n", autom->states);
 

  	printf ("Initial distribution:\n");
  	for (i=0; i<autom->states; i++)
		printf ("%5.3f ", autom->state[i].initial);
  	printf ("\n\n");
 
  	printf ("Final distribution:\n");
  	for (i=0; i<autom->states; i++)
		printf ("%5.3f ", autom->state[i].final);
  	printf ("\n\n");

  	printf ("Edges:\n");
  	for (state = 0; state < autom->states; state++) {

		/* output edges */
		for (ptr = autom->state[state].edges; 
		     ptr != NULL; ptr = ptr->next) {
	     	     printf ("From state: % d to state: %d : ", 
				state, ptr->into);
             	     if (autom->ntapes == 1) 
                 	printf ("label: %c, ", ptr->label[0]); 
	             else
                 	printf ("labels: %c %c, ", ptr->label[0], 
				 ptr->label[1]); 

	             printf ("weight: %f\n", ptr->weight);
       	        }
            }

  } else {
     	fprintf (fp,"Number of states: %d\n", autom->states);
 

  	fprintf (fp,"Initial distribution:\n");
  	for (i=0; i<autom->states; i++)
		fprintf (fp,"%5.3f ", autom->state[i].initial);
  	fprintf (fp,"\n\n");
 
  	fprintf (fp,"Final distribution:\n");
  	for (i=0; i<autom->states; i++)
		fprintf (fp,"%5.3f ", autom->state[i].final);
  	fprintf (fp,"\n\n");

  	fprintf (fp,"Edges:\n");
  	for (state = 0; state < autom->states; state++) {

	     /* output edges */
	     for (ptr = autom->state[state].edges; ptr != NULL; ptr = ptr->next) {
	          fprintf (fp,"From state: % d to state: %d : ", state, ptr->into);
                  if (autom->ntapes == 1) 
                      fprintf (fp,"label: %c, ", ptr->label[0]); 
	          else
                      fprintf (fp,"labels: %c %c, ", ptr->label[0], ptr->label[1]); 

	          fprintf (fp,"weight: %f\n", ptr->weight);
       	          }
         }
    }
}

void write_color_aut (outfile, aut_red, aut_green, aut_blue, ntapes)
FILE       *outfile;
automaton  *aut_red, *aut_green, *aut_blue;
int	   ntapes;
{
  int       i, state;
  edgetype  *ptr;
  float     greyness = 0.0, norm = 1.0;

  fprintf (outfile, "c\n");
  fprintf (outfile, "%d\n", aut_red->states);
 

/*
  if (ntapes == 1) 
	fprintf (outfile, "%f %f\n", greyness, norm);
*/

  for (i=0; i<aut_red->states; i++)
	fprintf (outfile, "%f ", aut_red->state[i].initial);
  fprintf (outfile, "\n");
 
  for (i=0; i<aut_green->states; i++)
	fprintf (outfile, "%f ", aut_green->state[i].initial);
  fprintf (outfile, "\n");
 
  for (i=0; i<aut_blue->states; i++)
	fprintf (outfile, "%f ", aut_blue->state[i].initial);
  fprintf (outfile, "\n");
 
  for (i=0; i<aut_red->states; i++)
	fprintf (outfile, "%f ", aut_red->state[i].final);
  fprintf (outfile, "\n");

  for (i=0; i<aut_green->states; i++)
	fprintf (outfile, "%f ", aut_green->state[i].final);
  fprintf (outfile, "\n");

  for (i=0; i<aut_blue->states; i++)
	fprintf (outfile, "%f ", aut_blue->state[i].final);
  fprintf (outfile, "\n");

  for (state = 0; state < aut_red->states; state++) {
	fprintf (outfile, "%d\n", state);

	/* output edges */
	for (ptr = aut_red->state[state].edges; ptr != NULL; ptr = ptr->next) {
	     switch (ntapes) {
		case 1: if (ptr->label[0] != 'e')
                           fprintf (outfile,"%d ", ptr->label[0]-'0'); 
			else
			   fprintf (outfile,"%d ", 4);
			break;
		case 2: fprintf (outfile,"%c %c ", ptr->label[0], ptr->label[1]); 
			break;
	     }

	     fprintf (outfile, "%d %f\n", ptr->into, ptr->weight);
       	}
        /* mark the end of the edge list */

        switch (ntapes) {
	    case 1: fprintf (outfile, "%d \n", -1); break;
	    case 2: fprintf (outfile, "- \n"); break;
       }
    }
}



/************************************************************
 * Function: compute_depths()                               *
 *           Compute the distance of each state from the    *
 *           initial state                                  *
 ************************************************************/

void compute_depths (autom)
automaton *autom;
{
  char d, depth_changed;
  int  i;
  edgetype *ptr;

  for (i=0; i<autom->states; i++) {
    if (autom->state[i].initial != 0)
      autom->state[i].depth=0;
     else
      autom->state[i].depth=255;
  }

  /* compute the depths */
  for (depth_changed=TRUE,d=0; depth_changed; d++){
       depth_changed = FALSE;
       for (i=0; i<autom->states; i++)
	  if (autom->state[i].depth == d)  
	     for (ptr=autom->state[i].edges; ptr!=NULL; ptr=ptr->next)
	       if (autom->state[ptr->into].depth == 255) {
		  if (autom->ntapes > 1 && ptr->label[0] == 'e' && ptr->label[1] == 'e')
		     autom->state[ptr->into].depth = d;
		  else
		     autom->state[ptr->into].depth = d+1;
		  depth_changed = TRUE;
	        }
  }
}



/************************************************************
 * Function: add_new_vector                                 *
 *           Performs one step in the unrolling computation *
 *           for the 1-tape WFA                             *
 ************************************************************/

void add_new_vector (source,dest,size,weight,label)
real source[],dest[];
unsigned int size;
real weight;
unsigned char label;
{
 register real *x_s,*y_s,*x_d,*y_d;
 real		w;

 if (label != 'e') {
    switch (label) {
		case '3': dest += 2*size*size; break;
		case '0': dest += size; break;
		case '1': dest += size + 2*size*size; break;
    }

    for (x_s=source,x_d=dest; x_s!=source+size*size; x_s+=size, x_d+=2*size)
         for (y_s=x_s, y_d=x_d; y_s!=x_s+size; y_s++, y_d++) 
             (*y_d)+=weight*(*y_s);
 } else {
    for (x_s=source,x_d=dest; x_s!=source+size*size; x_s+=size, x_d+=4*size)
         for (y_s=x_s, y_d=x_d; y_s!=x_s+size; y_s++, y_d += 2) {
	      w = weight*(*y_s);
	      *y_d += w; y_d[1] += w; y_d[2*size] += w; y_d[2*size+1] += w; 
	 }
 }
}

/************************************************************
 * Function:	append (str1, label, str2)                  *
 *              str2 = label.str1 (. denotes concatenation) *
 ************************************************************/

void append (str1, label, str2)
char *str1, label, *str2;
{
   *str2++ = label;
   while (*str1 != '\0')
        *str2++ = *str1++;
     *str2 = '\0';
}

/*********************************************************
 * Function: update tape                                 *
 *           Updates the input tape of the automaton     *
 *           for the 2-tape WFA (WFT)                    *
 *********************************************************/

void update_tape (source_tape,dest_tape,size,length, weight,label)
tape     *source_tape[],*dest_tape[];
unsigned int size, length;
float    weight;
unsigned char *label;
{
 tape **x_s,**y_s,**x_d,**y_d, 
      *source, **dest,**dest1, **dest2, **dest3;
 int  i;
 float w;
 void install ();

 if (label[1] != 'e') {
     switch (label[1]) {
		case '3': dest_tape += 2*size*size; break;
		case '0': dest_tape += size; break;
		case '1': dest_tape += size + 2*size*size; break;
     }

     for (x_s=source_tape,x_d=dest_tape; x_s!=source_tape+size*size; 
          x_s+=size, x_d+=2*size)
          for (y_s=x_s, y_d=x_d; y_s!=x_s+size; y_s++, y_d++)
               for (source = *y_s, dest = y_d; 
                    source != NULL; source = source->next)  
                    install (source, dest, label[0], length, weight);
	      
 } else {
     for (x_s=source_tape,x_d=dest_tape; x_s!=source_tape+size*size; 
          x_s+=size, x_d+=4*size)
          for (y_s=x_s, y_d=x_d; y_s!=x_s+size; y_s++, y_d += 2) {
               dest1 = y_d+1; dest2 = y_d+2*size;
	       dest3 = y_d+2*size+1; dest = y_d;
               for (source = *y_s; source != NULL; source = source->next) { 
                      install (source, dest, label[0], length, weight);
                      install (source, dest1, label[0], length, weight);
                      install (source, dest2, label[0], length, weight);
                      install (source, dest3, label[0], length, weight);
               } 
	  }     /* end for */
  }
}


/*****************************************************
 * Install an element into the linked list           *
 *****************************************************/

void install (source, dest, label, length, weight)
tape	 *source, **dest;
char     label;
int	 length;
float    weight;
{
   tape	*p;
   char *c, addr[MAXDEPTH];
   int  found = FALSE;

   append (source->address, label, addr);
   for ( p = *dest; p != NULL ; p = p->next) {
         if (strcmp( p->address, addr) == 0) {
             found = TRUE;
	     p->weight += weight*(source->weight);
             break;
	}
   }

   if (!found) {

      if ((p = (tape *) malloc (sizeof (tape))) == NULL) {
	printf ("Unable to allocate memory for a new tape entry! \n");
	exit (1);
      }

   /*
      if ((c = (char *) malloc ((length+1)*sizeof (char))) == NULL) {
	   printf ("Unable to allocate memory for a new tape entry! \n");
	   exit (1);
      }
      p->address = c;
   */

      strcpy (p->address, addr);
      p->weight = weight*(source->weight);
      p->next = *dest;
      *dest = p;
     }
                     
    if (debug == 2) 
      fprintf (fd,"source string %s, label: %c result %s, weight: %f\n",
               source->address, label, p->address, p->weight); 
}


/*****************************************************************
 * Function: update address                                      *
 *           Updates the input address in simple WFT unrolling   *
 *****************************************************************/

void update_address (source,dest,size,length,label)
char source[],dest[];
unsigned int size;
int	length;
unsigned char *label;
{
 char *x_s,*y_s,*x_d,*y_d;

 if (label[1] != 'e') {
     switch (label[1]) {
		case '3': dest += 2*size*size*(length+1); break;
		case '0': dest += size*(length+1); break;
		case '1': dest += (size + 2*size*size)*(length+1); break;
     }

     for (x_s=source,x_d=dest; x_s!=source+size*size*length; 
          x_s+=size*length, x_d+=2*size*(length+1))
          for (y_s=x_s, y_d=x_d; y_s!=x_s+size*length; 
               y_s += length, y_d += length+1){
	       if ((*y_s != '\0' && length > 1) || length == 1) {
                  append (y_s, label[0], y_d);
 	          if (debug == 2) 
                     fprintf (fd, "source string %s, label: %c result %s\n", 
                              y_s, label[0], y_d);
		}
          }
 } else {
     for (x_s=source,x_d=dest; x_s!=source+size*size*length; 
          x_s+=size*length, x_d+=4*size*(length+1))
          for (y_s=x_s, y_d=x_d; y_s!=x_s+size*length; 
               y_s += length, y_d += 2*length+2){
		if ((*y_s != '\0' && length > 1) || length == 1) {
	           append (y_s, label[0], y_d);
	           append (y_s, label[0], y_d+length+1);
	           append (y_s, label[0], y_d+2*size*(length+1));
	           append (y_s, label[0], y_d+(2*size+1)*(length+1));
 	           if (debug == 2) 
                      fprintf (fd, "source string %s, label: %c result %s\n", 
				y_s, label[0], y_d);
		}
          }
 }
}

/******************************************************
 * Function: unroll_automaton                         *
 *           Unrolls the automaton backwards to the   *
 *           specified depth                          *
 ******************************************************/

void unroll_automaton (autom, depth)
automaton *autom;
int	depth;
{
 int	  i,j,k,length,size;
 register edgetype *ptr;
 tape     *p;
 void     free_list ();

 for (j = 0; j < autom->states; j++) {
     if (autom->ntapes == 1 || autom->method == SIMPLE) {
         autom->state[j].old = (real *) malloc(sizeof(real));
         autom->state[j].old[0] = autom->state[j].final;
     }

     if (autom->ntapes > 1) {
	if (autom->method == SIMPLE) {
	    autom->state[j].old_address = (char *) malloc (sizeof (char));
	    autom->state[j].old_address[0] = '\0';
        }

	if (autom->method == GENERAL) {
            /* initialize the tapes */
            autom->state[j].old_tape = (tape **) malloc(sizeof(tape *));
            autom->state[j].old_tape[0] = NULL;
	
        
            p = (tape *) malloc (sizeof (tape));
            if (p == NULL) {
	        printf ("Unable to allocate memory for a new tape entry! \n");
	        exit (1);
            }
/*
            if ((p->address = (char *) malloc(sizeof(char))) == NULL) {
	        printf ("Unable to allocate memory for a new tape entry! \n");
	        exit (1);
            }
*/
            p->address[0] = '\0';
            p->weight = autom->state[j].final;
            p->next = autom->state[j].old_tape[0];
            autom->state[j].old_tape[0] = p;
        }
     }
 }
    
/*
 * Main loop
 */

    for (i=depth-1, size=1, length = 1; i >= 0 ; i--, size *= 2, length++) {

	if(debug == 2) 
           fprintf (fd, "size = %d\n", 2*size);
	/*printf ("size = %d\n", 2*size); fflush (stdout);*/
        for (j=0; j<autom->states; j++) {
	 if (autom->state[j].depth <= i) {
	     if (autom->ntapes == 1 || autom->method == SIMPLE) {
	         if ((autom->state[j].new 
		        = (real *) calloc(size*size*4, sizeof(real)))
		        == NULL) { 
				   printf("Error in allocating new table\n");
				   exit (-1);
				  }
	     }

	     if (autom->ntapes > 1) {
		if (autom->method == SIMPLE) {
		   if ((autom->state[j].new_address 
                        = (char *) calloc (size*size*4*(length+1), sizeof (char)))
			== NULL) {
				   printf("Error in allocating new tape\n");
				   exit (-1);
				  }
                }

		if (autom->method == GENERAL) {
                    if ((autom->state[j].new_tape 
			= (tape **) calloc (size*size*4, 
                                                sizeof(tape *)))
			== NULL) {
				   printf("Error in allocating new tape\n");
				   exit (-1);
				  }
                    for (k=0; k < size*size*4; k++) 
                         autom->state[j].new_tape[k] = NULL;
		}
	     }

	   /*
	    * Visit all neighbors of the state j
            */

           for( ptr = autom->state[j].edges; ptr!= NULL; ptr=ptr->next) {
	       if(debug == 2) 
                 fprintf (fd, 
                 "Now considering edge from state %d to state %d: weight %f,labels: %c %c\n", 
                       j, ptr->into, ptr->weight, ptr->label[0], ptr->label[1]);
                 /*printf ( 
                 "Now considering edge from state %d to state %d: weight %f,labels: %c %c\n", 
                       j, ptr->into, ptr->weight, ptr->label[0], ptr->label[1]);
                 fflush (stdout);*/

               if (autom->ntapes == 1 || autom->method == SIMPLE) {
                   add_new_vector(autom->state[ptr->into].old,
		                  autom->state[j].new,
		       	          size,
                                  ptr->weight,
			          ptr->label[autom->ntapes-1]);
		}

	       if (autom->ntapes > 1) {
		   if (autom->method == SIMPLE) {
		       update_address (autom->state[ptr->into].old_address,
				       autom->state[j].new_address,
					size, length,
					ptr->label);
		   }

		   if (autom->method == GENERAL) {
                       update_tape(autom->state[ptr->into].old_tape,
		                   autom->state[j].new_tape,
		       	           size, length,
                                   ptr->weight,
			           ptr->label);
		   }
		}
	   }       /* end for (ptr)    */
        }	   /* end if (depth)   */
    }		   /* end for (states) */

   /*
    * Clean up the old table
    */

    for (j=0; j<autom->states; j++)  {
	 if (autom->state[j].depth <= i) {
            if (autom->ntapes == 1 || autom->method == SIMPLE) {
	       if (autom->state[j].old != NULL) free(autom->state[j].old);
               autom->state[j].old = autom->state[j].new;
               autom->state[j].new=NULL;
            }

	    if (autom->ntapes > 1) {
               if (autom->method == SIMPLE) {
	          if (autom->state[j].old_address != NULL) 
                      free(autom->state[j].old_address);
                  autom->state[j].old_address = autom->state[j].new_address;
                  autom->state[j].new_address = NULL;
		}

		if (autom->method == GENERAL) {
                    for (k=0; k<size*size; k++)
	                if (autom->state[j].old_tape[k] != NULL) 
                           free_list (&(autom->state[j].old_tape[k]));
                    free (autom->state[j].old_tape);
                    autom->state[j].old_tape = autom->state[j].new_tape;
                    autom->state[j].new_tape=NULL;
		}
	    }
          }
     }
  }		/* i >= 0 */
}


/*****************************************************
 * Free memory space occupied by a WFA               *
 *****************************************************/

free_automaton (autom)
automaton	*autom;
{
  int	i;
  edgetype *p, *q;

  for (i=0; i<autom->states; i++) {
	for (p = autom->state[i].edges; p != NULL; p = q) {
	     q = p->next; p->next = NULL; free(p); p = NULL; }
	if (autom->state[i].edges != NULL) { 
	     (autom->state[i].edges)->next = NULL; 
	     free (autom->state[i].edges); 
	     autom->state[i].edges = NULL; 
	}
  }

  autom->states = 0;
}

/*****************************************************
 * Free memory space occupied by a linked list       *
 *****************************************************/

void free_list (t)
tape     **t;
{
  tape *p, *q;

  for (p = *t; p != NULL; p = q) {
       q = p->next; p->next = NULL; 
       /*free (p->address);*/ free(p); p = NULL; 
  }

  if (*t != NULL) { 
     (*t)->next = NULL; free(*t); *t = NULL; 
  } 
}

/**************************************************
 * Function: decode automaton                     *
 *           Decode automaton to specified depth  *
 **************************************************/

void decode_automaton (autom, depth)
automaton *autom;
int	  depth;
{
  printf ("Decoding the automaton...\n"); fflush (stdout);
  compute_depths (autom);
  unroll_automaton (autom, depth);
  printf ("Done\n"); fflush(stdout);
}


/***************************************************************
 * Function: find_value                                        *
 *           Finds the greyness value specified by the address *
 ***************************************************************/

float find_value (img, address, size)
imgtype **img;
char	*address;
int	 size;
{
  int	i, j, s,x=0,y=0;
  float avg;

  if (*address == '\0') return(0);

  for(s=size; *address != '\0'; address++) {
	switch (*address) {
		case '3': y += s/2; break;
		case '0': x += s/2; break;
		case '2': x += s/2; y += s/2; break;
        }
	if (*address != 'e') s /= 2;
  }

  if (s <= 1) {
            return ((float) img[x][y]);
  } else {
     if (s > 8) s = 8;
     for (avg=0, i=0; i<s; i++)
	 for (j=0; j<s; j++)
              avg += img[x+i][y+j];

     avg /= s*s;
     return (avg);
  }
}


/****************************************************************
 * Function:	output_image					*
 *		Copmputes the result and puts it into the       *
 *              output image                                    *
 ****************************************************************/

void output_image (raw_img, autom, input_size, output_size, depth)
imgtype       ***raw_img;
automaton     *autom;
int	      input_size, *output_size;
int	      depth;
{
  int	i, j, k, N;
  int	ntapes, size, step, ix, iy;
  float res, greyness, weight;
  char  *address;
  tape  *tcell;
  float *output, max, min, sc, x;

  ntapes = autom->ntapes;
  for (size = 1, i=0; i<depth; i++, size *= 2);
  N = size * size;
  if (syn_flag)
     *output_size = size;
  else
     *output_size = (size >= DSIZE) ? size : DSIZE;

  step = *output_size/size;
  output = alloc_vect (N);

  if (scaling_flag) {
      min = 9999;
      max = -9999;
  }

  for (i=N-1; i >= 0; i--) {
    res = 0;
    for (j=0; j < autom->states; j++) {
	if(autom->state[j].depth==0) {
           if (ntapes > 1) {
              if (autom->method == GENERAL) {
	         for ( weight=0, tcell  = autom->state[j].old_tape[i]; 
                     tcell != NULL; tcell = tcell->next ) {
                     address = tcell->address;
                     if (address[0] != '\0') {
                        greyness = find_value(*raw_img, address, input_size);
	                weight += tcell->weight*greyness;
		     }
                  }
                  res += weight*autom->state[j].initial;
		}

	      if (autom->method == SIMPLE) {
		  address = autom->state[j].old_address + i*(depth+1);
                  greyness = find_value(*raw_img, address, input_size);
	          res += autom->state[j].old[i]*
                         autom->state[j].initial*greyness;
	      }
	   } else 
	       res += autom->state[j].old[i]*autom->state[j].initial;
         }
    }

    output[N-i-1] =  res;
    if (scaling_flag) {
       if (res > max) max = res;
       if (res < min) min = res;
   }
  }

  
  if (scaling_flag) {
     sc = 1.0/(max - min);
     printf ("max = %f, min = %f, scaling factor = %f\n", max, min, sc);
  } else sc = 1.0;

  if (ntapes == 1) sc *= 255;

  if (*raw_img != NULL) free_image (*raw_img, input_size);

  *raw_img = alloc_image (*output_size, *output_size);

  for (i=k=0; i < *output_size; i += step) {
	for (j=0; j<*output_size; j += step) { 
		x = sc*output[k++];
                if (cutoff_flag) {
		    if (x > 255) x = 255;
		    if (x < 0) x = 0;
                }
                if (step == 1) 
                    (*raw_img)[i][j] = (imgtype) x;
                else
                    for (ix=0; ix < step; ix++)
			for (iy = 0; iy < step; iy++)
                             (*raw_img)[i+ix][j+iy] = (imgtype) x;
       }
  }

  free(output);
}

/****************************************************************
 * Function:	output_matrix					*
 *		Copmutes the matrix form of a WFT relation      *
 *              output image                                    *
 * Note: The WFT must be epsilon-free!!                         *
 ****************************************************************/

void output_matrix (fp, autom, input_size, output_size, depth)
FILE	      *fp;
automaton     *autom;
int	      input_size, output_size;
int	      depth;
{
  int	i, j, k, N, size;
  char  *address;
  tape  *tcell;
  float weight;

  for (size = 1, i=0; i<depth; i++, size *= 2);
  N = size * size;

  if (fp != NULL)
     fprintf (fp, "source address: | destination pixel: | weight: \n");
  else
     printf ("source address: | destination pixel: | weight: \n");

  for (i=N-1; i >= 0; i--) {
    for (j=0; j < autom->states; j++) {
	if(autom->state[j].depth==0) {
              if (autom->method == GENERAL) {
	         for (tcell  = autom->state[j].old_tape[i]; 
                     tcell != NULL; tcell = tcell->next ) {
                     address = tcell->address;
		     weight = tcell->weight * autom->state[j].initial;
                     if (address[0] != '\0' && weight != 0) {
			if (fp != NULL)
			    fprintf (fp, "%s       |  %d           | %5.3f\n", 
					address, i, weight);
			else
			    printf ("%s       |  %d           | %5.3f\n", 
					address, i, weight);
		     }
                  }
		}

	      if (autom->method == SIMPLE) {
		  address = autom->state[j].old_address + i*(depth+1);
		  weight = autom->state[j].old[i] * autom->state[j].initial;
                  if (address[0] != '\0' && weight != 0) {
			if (fp != NULL)
			    fprintf (fp, "%s       |  %d           | %5.3f\n", 
				address, i, weight);
			else
			    printf ("%s       |  %d           | %5.3f\n", 
				address, i, weight);

		   }
	      }
         }
    }
   }
}


/***************************************************************
 * Function: find_position                                     *
 *           Finds the location in the source image            *
 *	     specified by the address 			       *
 ***************************************************************/

int find_position (address, size)
char	*address;
int	 size;
{
  int	s,x=0,y=0;

  for(s=size; *address != '\0'; address++) {
	switch (*address) {
		case '3': y += s/2; break;
		case '0': x += s/2; break;
		case '2': x += s/2; y += s/2; break;
        }
	if (*address != 'e') s /= 2;
  }

  return (x + y*size);
}

/*******************************************************
 * Function: compose ()                                *
 *           Composition of WFT                        *
 *******************************************************/

void compose (aut1, aut2)
automaton     *aut1, *aut2;
{
  automaton     autom, a2;
  edgetype 	*ptr1, *ptr2, *newptr, *ptr;
  int	   	nstates, ntapes, i, j, k, l, p, q, new;
  int		nstates1, nstates2;
  char	   	u, v, z, label1, label2, label;
  float    	**weight;
  void		copy_automaton ();


  nstates = aut1->states * aut2->states;
  ntapes = aut1->ntapes + aut2->ntapes - 2;

  /* allocate weights */

  if ((weight = (float **) calloc (aut1->states, sizeof (float*))) == NULL) {
	printf ("Cannot allocate the weights array"); exit (1);
  }
  for (k = 0; k < aut1->states; k++) {
      if ((weight[k] = (float*) calloc (aut2->states, sizeof (float))) == NULL) {
	   printf ("Cannot allocate the weights array"); exit (1);
       }
  }

   /* find the composition */

   /* make a copy of the second WFT */
   copy_automaton (aut2, &a2);

   /* adjust the weight matrices */

   for (i=0; i<aut1->states; i++) {
    	ptr = (edgetype *) malloc (sizeof (edgetype));
     	if (ptr == NULL) {
	   printf ("Error in allocating a new edge\n");
	   exit(1);
	}

	ptr->into = i;
	ptr->weight = 1;
	for (k=0; k<aut1->ntapes; k++) ptr->label[k] = 'e';
	ptr->next = aut1->state[i].edges;
	aut1->state[i].edges = ptr;
    }

   for (i=0; i<a2.states; i++) {
    	ptr = (edgetype *) malloc (sizeof (edgetype));
     	if (ptr == NULL) {
	   printf ("Error in allocating a new edge\n");
	   exit(1);
	}

	ptr->into = i;
	ptr->weight = 1;
	for (k=0; k<a2.ntapes; k++) ptr->label[k] = 'e';
	ptr->next = a2.state[i].edges;
	a2.state[i].edges = ptr;
    }

   for (i=0; i < aut1->states; i++) {
      for (j=0; j < a2.states; j++) {
	   /* create new state of the composition */
	   new = j+i*(a2.states);
           if (debug == 2)
	      fprintf (fd, "aut1 -  state: %d,\n aut2 - state: %d,\n new aut - state: %d \n", i, j, new);
           
	   autom.state[new].new = NULL;
	   autom.state[new].old = NULL;
	   autom.state[new].edges = NULL;

           if (ntapes == 2) {

	   for (label1 = '0'; label1 < '5'; label1++) {
	        for (label2 = '0'; label2 < '5'; label2++) {

                     if (debug == 2)
		        fprintf (fd, "Now considering labels %c,%c:\n", label1, label2);
	             /* initialze weights */

	   	     for (p=0; p<aut1->states; p++)
                          for (q=0; q<a2.states; q++)
                               weight[p][q] = 0;

	             /* find edges */

	             for (ptr1 = aut1->state[i].edges; 
                          ptr1 != NULL; ptr1 = ptr1->next) {

                          u = ptr1->label[0];
                          if (debug == 2)
			     fprintf (fd, "aut1: edge from %d to %d, labels %c %c, weight %f\n", i, p, u, ptr1->label[1], ptr1->weight);
			  if (label1 == u || (label1 == '4' && u == 'e')) {
		             z = ptr1->label[1];
	                     for (ptr2 = a2.state[j].edges; 
                                  ptr2 != NULL; ptr2 = ptr2->next) {
			          v = ptr2->label[1];
                               if (debug == 2)
			          fprintf (fd, "aut2: edge from %d to %d, labels %c %c, weight %f\n", j, q, ptr2->label[0], v, ptr2->weight);
		                  if (ptr2->label[0] == z &&
			             (label2 == v || 
                                     (label2 == '4' && v == 'e'))) {
			              weight[ptr1->into][ptr2->into] += 
                                          ptr1->weight*ptr2->weight;
				      if (debug == 2)
					fprintf (fd, "weight[%d][%d] = %f\n",
						 ptr1->into, ptr2->into,
			              		 weight[ptr1->into][ptr2->into]); 
		                  }
			      }
	                    }
	   	    }

		    /* add new edges from (i,j) to (p, q) on labels
		     * label1 and label 2
		     */

	   	     for (p=0; p<aut1->states; p++) {
                        for (q=0; q<a2.states; q++) {
			    /* weight on epsilon loops should be 0 */
		            if (label1 == '4' && label2 == '4' && 
                                i == p && j == q) weight[p][q] = 0;

			  if (weight[p][q]) {
			     if (debug == 2) 
				fprintf (fd, "p = %d, q = %d, weight = %f\n",
					 p, q, weight[p][q]);
			     newptr = (edgetype *) malloc (sizeof (edgetype));
			     if (newptr == NULL) {
				printf ("Error in allocating a new edge\n");
				exit(1);
			     }
			     if (label1 == '4')
		                 newptr->label[0] = 'e';
                             else
		                 newptr->label[0] = (unsigned char) label1;

			     if (label2 == '4')
		                 newptr->label[1] = 'e';
                             else
		                 newptr->label[1] = (unsigned char) label2;
		             
		             newptr->into = q + p * (a2.states);
			     newptr->weight = weight[p][q]; 
			     newptr->next = autom.state[new].edges;
			     autom.state[new].edges = newptr;
			  }
		      } 
                    }
                 }    /* end for (label1) */
              }       /* end for (label2) */
            }         /* end if (ntapes == 2) */
    
	   if (ntapes == 1) {
		for (label = '0'; label < '5'; label++) {
                     if (debug == 2)
		        fprintf (fd, "Now considering label %c:\n", label);
	             /* initialze weights */

	   	     for (p=0; p<aut1->states; p++)
                          for (q=0; q<a2.states; q++)
                               weight[p][q] = 0;

	             /* find edges */

	             for (ptr1 = aut1->state[i].edges; 
                          ptr1 != NULL; ptr1 = ptr1->next) {

		          z = ptr1->label[0];
                          if (debug == 2)
			     fprintf (fd, "aut1: edge from %d to %d, label %c, weight %f\n", i, p, z, ptr1->weight);
	                  for (ptr2 = a2.state[j].edges; 
                               ptr2 != NULL; ptr2 = ptr2->next) {
			       v = ptr2->label[1];
                               if (debug == 2)
			          fprintf (fd, "aut2: edge from %d to %d, labels %c %c, weight %f\n", j, q, ptr2->label[0], v, ptr2->weight);
		               if (ptr2->label[0] == z && 
                                  (label == v || (label == '4' && v == 'e'))) {
			          weight[ptr1->into][ptr2->into] += 
                                      ptr1->weight*ptr2->weight;
				  if (debug == 2)
                                     fprintf (fd, "weight[%d][%d] = %f\n",
					ptr1->into, ptr2->into,
			              	weight[ptr1->into][ptr2->into]); 
		               }
			   }
	   	     }

		    /* add new edges from (i,j) to (p, q) on 
                     * considered label
		     */

	   	     for (p=0; p<aut1->states; p++) {
                        for (q=0; q<a2.states; q++) {
			    if (label != '4') {
			       if (weight[p][q]) {
			          if (debug == 2) 
				     fprintf (fd, "p = %d, q = %d, weight = %f\n",
					      p, q, weight[p][q]);
			          newptr = (edgetype *) malloc (sizeof (edgetype));
			          if (newptr == NULL) {
				     printf ("Error in allocating a new edge\n");
				     exit(1);
			          }
		                  newptr->label[0] = (unsigned char) label;
		                  newptr->into = q + p * (a2.states);
			          newptr->weight = weight[p][q]; 
			          newptr->next = autom.state[new].edges;
			          autom.state[new].edges = newptr;
			       }
			    } else {
				if (i==p && j==q) weight[p][q] = 0;
			        if (weight[p][q]) {
			          if (debug == 2) 
				     fprintf (fd, "p = %d, q = %d, weight = %f\n",
					      p, q, weight[p][q]);
			          newptr = (edgetype *) malloc (sizeof (edgetype));
			          if (newptr == NULL) {
				     printf ("Error in allocating a new edge\n");
				     exit(1);
			          }
		                  newptr->label[0] = 'e';
		                  newptr->into = q + p * (a2.states);
			          newptr->weight = weight[p][q]; 
			          newptr->next = autom.state[new].edges;
			          autom.state[new].edges = newptr;
				}
			    }    /* end if */
		        }        
		      }		/* end of the (p,q) - loop */
               } 		/* end for (label) */
          }			/* end if (ntapes == 1) */
      }
    }				/* end of the states loop */


    /* free weights */

    for (k=0; k < aut1->states; k++)  free (weight[k]);
    free (weight);
    free_automaton (&a2);

    /*  Find the initial and final distribution */

    nstates1 = aut1->states;
    nstates2 = aut2->states;
    autom.states = nstates;
    autom.ntapes = ntapes;

    for (i = 0; i < nstates1; i++) {
	for (j = 0; j < nstates2; j++) {
	     k = j+i*nstates2;
	     autom.state[k].initial = aut1->state[i].initial *
				      aut2->state[j].initial;
	     autom.state[k].final = aut1->state[i].final *
				    aut2->state[j].final;
	   }
      }


       free_automaton (aut1);
       for (k=0; k < nstates; k++) {
	    aut1->state[k].initial = autom.state[k].initial;
	    aut1->state[k].final = autom.state[k].final;
	    aut1->state[k].edges = autom.state[k].edges;
	}


      /* reset the tapes and tables */

      printf ("\nNumber of states: %d\n", nstates);
       aut1->states = nstates;
       aut1->ntapes = ntapes;

      for (k=0; k<nstates; k++) {
	   aut1->state[k].old = NULL;
	   aut1->state[k].new = NULL;
	   aut1->state[k].old_tape = NULL;
	   aut1->state[k].new_tape = NULL;
      }
}

/* End of compose */


int concatenate (aut1, aut2)
automaton	*aut1, *aut2;
{
   int	i, j, nstates1, nstates2, nstates, first, last, count;
   int  state_num[MAX_STATES], label1, label2, aut2_ok;
   int  final_state[MAX_STATES], label, first_final, ntapes;
   edgetype *ptr, *newptr;
   float weights[5][5], weight[5], initial, final, average;

   nstates1 = aut1->states; nstates2 = aut2->states;
   ntapes = aut1->ntapes;


   /* check the first automaton */

   for (i = 0; i < nstates1; i++) {
	if (aut1->state[i].final != 0) {
	   final_state[i] = TRUE;
	   if ((ptr=aut1->state[i].edges ) != NULL ) {

		if (ntapes == 1) 
		    for (label = 0; label <= 4; label++) weight[label] = 0;
		else {
		    for (label1 = 0; label1 <= 4; label1++) 
		         for (label2 = 0; label2 <= 4; label2++) 
			      weights[label1][label2] = 0;
		}

                for (;ptr != NULL; ptr=ptr->next) {
		    if (ptr->into != i) return (FALSE);
		    else if (ptr->weight != 1) return (FALSE);
		    else {
			if (ntapes == 1) {
			    if (ptr->label[0] == 'e') label = 4;
			    else label = ptr->label[0] - '0';
			    weight[label] = ptr->weight;
			} else {
			    if (ptr->label[0] == 'e') label1 = 4;
			    else label1 = ptr->label[0] - '0';
			    if (ptr->label[1] == 'e') label2 = 4;
			    else label2 = ptr->label[1] - '0';
			    weights[label1][label2] = ptr->weight;
			}
		      }
		  }
		  if (ntapes == 1) {
		     for (label = 0; label <= 3; label++)
			 if (weight[label] != 1) return (FALSE);
		  } else {
		     for (label1 = 0; label1 <= 4; label1++)
		         for (label2 = 0; label2 <= 4; label2++)
			      if (weights[label1][label2] != 1) return (FALSE);
		  }
	      }
	} else
	   final_state[i] = FALSE;
   }


   /* second automaton must have only one initial state */

   for (aut2_ok = TRUE, count = 0, i = 0; i < nstates2 && aut2_ok; i++) {
	if ((initial = aut2->state[i].initial) != 0) {
           if (initial == 1 && count == 0)  first = i; 
	   count++;
	   if (count > 1 || initial != 1) aut2_ok = FALSE;
	}
   }

   /* if the second automaton  is not in the normal form, transform it */
   if (!aut2_ok) {
      
      /* compute average greyness */
      for (average = 0, i = 0; i < nstates2; i++) 
	   average += aut2->state[i].initial * aut2->state[i].final;
	
      /* create a new initial state  */
      aut2->state[nstates2].edges = NULL;
      aut2->state[nstates2].initial = 1;
      aut2->state[nstates2].final = average;

      for (i = 0; i < nstates2; i++) {

        if (ntapes == 1) {
	  for (label = 0; label <= 4; label++) weight[label] = 0;
          for (j = 0; j < nstates2; j++) {
	      if ((initial = aut2->state[j].initial) != 0) {
	          for (ptr = aut2->state[j].edges; ptr != NULL; 
		       ptr = ptr->next) {
		       if (ptr->into == i) {
			   if (ptr->label[0] == 'e') label = 4;
		           else label = ptr->label[0] - '0';
		           weight[label] += initial * ptr->weight;
			}
		    }
		}
	     }

	     for (label = 0; label <= 4; label++) {
		  if (weight[label] != 0) {
	             newptr = (edgetype *) malloc (sizeof (edgetype));
	      	     if (newptr == NULL) {
		  	printf ("Error in allocation of the new edge!\n");
	          	exit (1);
	      	     }
	             newptr->into = i;
	             newptr->weight = weight[label];
		     if (label < 4) 
	                 newptr->label[0] = (unsigned char) label + '0';
		     else
			 newptr->label[0] = 'e';
	             newptr->next = aut2->state[nstates2].edges;
	             aut2->state[nstates2].edges =  newptr;
		  }
	      }
        }

        if (ntapes == 2) {
	   for (label1 = 0; label1 <= 4; label1++) 
	        for (label2 = 0; label2 <= 4; label2++) 
		     weights[label1][label2] = 0;

          for (j = 0; j < nstates2; j++) {
	      if ((initial = aut2->state[j].initial) != 0) {
	          for (ptr = aut2->state[j].edges; ptr != NULL; 
		       ptr = ptr->next) {
		       if (ptr->into == i) {
		    	  if (ptr->label[0] == 'e') 
			     label1 = 'e';
		    	  else
		             label1 = ptr->label[0] - '0';

		          if (ptr->label[1] == 'e') 
			      label2 = 'e';
		          else
		              label2 = ptr->label[1] - '0';

		           weights[label1][label2] += initial * ptr->weight;
			}
		    }
		}
	     }

	     for (label1 = 0; label1 <= 4; label1++) {
	        for (label2 = 0; label2 <= 4; label2++) {
		  if (weights[label1][label2] != 0) {
	             newptr = (edgetype *) malloc (sizeof (edgetype));
	      	     if (newptr == NULL) {
		  	printf ("Error in allocation of the new edge!\n");
	          	exit (1);
	      	     }
	             newptr->into = i;
		     if (label1 == 4)
			newptr->label[0] = 'e';
		     else
	      	        newptr->label[0] = (unsigned char) label1 + '0';
		     if (label2 == 4)
			newptr->label[1] = 'e';
		     else
	      	        newptr->label[1] = (unsigned char) label2 + '0';
	             newptr->weight = weights[label1][label2];
	             newptr->next = aut2->state[nstates2].edges;
	             aut2->state[nstates2].edges =  newptr;
		  }
	      }
           }
          }
        }

        for (i=0; i<nstates2; i++) aut2->state[i].initial = 0;
        first = nstates2;
	nstates2++; aut2->states = nstates2; 
    }
    
     /* replace the final states in the first automaton by
        the initial state of the second automaton */

     for (i = 0; i < nstates1; i++) {
      	 if (aut1->state[i].final) {
             aut1->state[i].edges = aut2->state[first].edges;
             aut1->state[i].final = aut2->state[first].final;
         }
    }

     /* append the second automaton */
     for (i = 0, j = nstates1; i < aut2->states; i++) {
          if (i != first) {
	      aut1->state[j].edges = aut2->state[i].edges;
	      aut1->state[j].initial = 0;
	      aut1->state[j].final = aut2->state[i].final;
	      state_num[i] = j++;
          }  
     }
     nstates = j;
   
     /* renumber the states */
     for (i = 0, first_final = -1; i < nstates1; i++) {
	 if (final_state[i]) {
	    if (first_final == -1) first_final = i;
            for (ptr = aut1->state[i].edges; ptr != NULL; ptr = ptr->next) 
		 if (ptr->into == first) ptr->into = i;
		 else ptr->into = state_num[ptr->into];
         }
     }

     for (j = nstates1; j < nstates;j++) {
          for (ptr = aut1->state[j].edges; ptr != NULL; ptr = ptr->next) 
	       if (ptr->into == first) ptr->into = first_final;
	       else ptr->into = state_num[ptr->into];
     }

     aut1->states = nstates;
     return (TRUE);
}



void addwfa (aut1, aut2)
automaton	*aut1, *aut2;
{
   int	i, j, nstates, nstates1, nstates2, state_num[MAX_STATES];
   edgetype *ptr;

   nstates1 = aut1->states;
   nstates2 = aut2->states;

     /* append the second automaton */
     for (i = 0, j = nstates1; i < nstates2; i++, j++) {
	      aut1->state[j].edges = aut2->state[i].edges;
	      aut1->state[j].initial = aut2->state[i].initial;
	      aut1->state[j].final = aut2->state[i].final;
	      state_num[i] = j;
     }
     nstates = j;
   
     /* renumber the states */
     for (i = 0; i < nstates2;i++) 
          for (ptr = aut1->state[nstates1+i].edges;
	       ptr != NULL; ptr = ptr->next) 
	       ptr->into = state_num[ptr->into];

     aut1->states = nstates;
     /*free (aut2);*/
} 

void free_autom_values (autom, size)
automaton *autom;
int	  size;
{
  int	j, k, N;

  N = size * size;

  for (j=0; j<autom->states; j++)  {
       if (autom->state[j].depth == 0) {
          if (autom->ntapes == 1 || autom->method == SIMPLE) 
	     if (autom->state[j].old != NULL) free(autom->state[j].old);
            
	    if (autom->ntapes > 1) {
		if (autom->method == SIMPLE) 
	           if (autom->state[j].old_address != NULL) 
                      free(autom->state[j].old_address);

		if (autom->method == GENERAL) {
                    for (k=0; k<N; k++)
	                if (autom->state[j].old_tape[k] != NULL) 
                           free_list (&(autom->state[j].old_tape[k]));
                    free (autom->state[j].old_tape);
                }
	    }
          }
  }
}

void invert_wft (wft)
automaton *wft;
{
  edgetype *ptr;
  char tmp;
  int  i;

  for (i = 0; i < wft->states; i++) {
     for (ptr = wft->state[i].edges; ptr != NULL; ptr = ptr->next) {
	  /* swap labels */
	  tmp = ptr->label[0];
	  ptr->label[0] = ptr->label[1];
	  ptr->label[1]  = tmp;
	  ptr->weight = 1.0/ptr->weight;
      }
     }
}

void check_automaton (wft)
automaton *wft;
{
  /*automaton aut;*/
  edgetype *ptr, *p;
  char tmp;
  int  i, state;

  /* reverse the wft */
/*
  aut.states = wft->states;
  for (i = 0; i < aut.states; i++) aut.state[i].edges = NULL;

  for (i = 0; i < wft->states; i++) {
     for (ptr = wft->state[i].edges; ptr != NULL; ptr = ptr->next) {
	  p = (edgetype *) malloc(sizeof(edgetype));
	  p->label[0]= ptr->label[1];
	  p->label[1]= ptr->label[0];
	  p->into=i;
	  p->next=aut.state[ptr->into].edges;
	  aut.state[ptr->into].edges=p;
      }
  }
*/

  printf ("Checking the automaton...\n"); fflush (stdout);

	
  wft->method = SIMPLE;

 /* check labels */

 for (i = 0; i < wft->states; i++) {
    if (debug == 2) {
       printf ("Now processing state %d:\n", i); 
       fflush (stdout);
    }
    for (ptr = wft->state[i].edges; ptr != NULL; ptr = ptr->next) {
         if (debug == 2) {
             printf ("Edge to state %d, labels %c %c\n", ptr->into, 
		ptr->label[0], ptr->label[1]); fflush (stdout);
         }
	 for (p = ptr->next; p != NULL; p = p->next) {
	      if (debug == 2) {
                  printf ("Edge to state %d, labels %c %c\n", 
                           p->into, p->label[0], p->label[1]); 
			   fflush (stdout);
               }
	       if ((p->label[1] == ptr->label[1] && 
                    p->label[0] != ptr->label[0]))  {
		  /*|| p->label[1] == 'e' || ptr->label[1] == 'e') */
		     wft->method = GENERAL; break; }
          }
    }
  }
  
  printf ("Done\n"); fflush (stdout);
 
  switch (wft->method) {
	case SIMPLE: printf ("Method: simple\n"); break;
	case GENERAL: printf ("Method: general\n"); break;
  }
} 


void copy_automaton (aut1, aut2) 
automaton *aut1, *aut2;
{
  int 	i,j;
  edgetype *ptr, *p;

  aut2->states = aut1->states;
  aut2->ntapes = aut1->ntapes;

  for (i=0; i < aut2->states; i++) {
	aut2->state[i].initial = aut1->state[i].initial;
	aut2->state[i].final = aut1->state[i].final;
	aut2->state[i].edges = NULL;
	aut2->state[i].new = NULL;

	for (p = aut1->state[i].edges; p != NULL; p = p->next) {
	     ptr=(edgetype *) malloc(sizeof(edgetype));
	     for (j = 0; j < aut2->ntapes; j++) 
		  ptr->label[j] = p->label[j];
	     ptr->into = p->into;
	     ptr->weight = p->weight;
	     ptr->next=aut2->state[i].edges;
	     aut2->state[i].edges=ptr;
	 }
  }
}

int cart_product (wfa1, wfa2, wfa_out)
automaton *wfa1, *wfa2, *wfa_out;
{
  int nstates1, nstates2, nstates, i, j;
  int new_edge, k, wft1_dim, wft2_dim;
  char label1, label2, label;
  edgetype *p1, *p2, *p;
  float weight;

  nstates1 = wfa1->states;
  nstates2 = wfa2->states;

  wft1_dim = 1;
  for (i=0; i<nstates1; i++)
       for (p1 = wfa1->state[i].edges; p1 != NULL; p1 = p1->next) 
	    if (p1->label[0] == '2' || p1->label[0] == '3') {
		wft1_dim = 2; break; }

  wft2_dim = 1;
  for (i=0; i<nstates2; i++)
       for (p2 = wfa2->state[i].edges; p2 != NULL; p2 = p2->next) 
	    if (p2->label[0] == '2' || p2->label[0] == '3') {
		wft2_dim = 2; break; }

  if (wft1_dim != wft2_dim) return (FALSE);
            
  printf ("Doing the cartesian product..."); fflush (stdout);
  nstates = nstates1*nstates2;
  wfa_out->states = nstates;
  wfa_out->ntapes = 1;

  for (i=0; i < nstates1; i++) {
       for (j = 0; j < nstates2; j++) {
	    k = i + j*nstates1;
	    wfa_out->state[k].initial = wfa1->state[i].initial *
					wfa2->state[j].initial;
	    wfa_out->state[k].final = wfa1->state[i].final *
				      wfa2->state[j].final;
	wfa_out->state[k].edges = NULL;
	wfa_out->state[k].new = NULL;

	for (p1 = wfa1->state[i].edges; p1 != NULL; p1 = p1->next) {
	     for (p2 = wfa2->state[j].edges; p2 != NULL; p2 = p2->next) {
		  weight = p1->weight * p2->weight;
		  label1 = p1->label[0]; label2 = p2->label[0];
		  if (weight != 0) {
		      if (wft1_dim == 1 && wft1_dim == 1) {
	                  p=(edgetype *) malloc(sizeof(edgetype));
	                  p->into = p1->into + (p2->into)*nstates1;
	                  p->weight = weight;
			  if (label1 == '0' && label2 == '0') {
			      p->label[0] = '0';	
			  } else if (label1 == '0' && label2 == '1') {
			      p->label[0] = '1';	
			  } else if (label1 == '1' && label2 == '0') {
			      p->label[0] = '2';	
			  } else if (label1 == '1' && label2 == '1') {
			      p->label[0] = '3';	
                          }
	                  p->next=wfa_out->state[k].edges;
	                  wfa_out->state[k].edges=p;
		      } else if (wft2_dim == 2 && wft1_dim == 2) {
			  new_edge = FALSE;
			  if (label1 == '0' && label2 == '0') {
			      label = '0'; new_edge = TRUE;	
			  } else if (label1 == '1' && label2 == '1') {
			      label = '1'; new_edge = TRUE;	
			  } else if (label1 == '2' && label2 == '2') {
			      label = '2'; new_edge = TRUE;
			  } else if (label1 == '3' && label2 == '3') {
			      label = '3'; new_edge = TRUE;
			  }
			  if (new_edge) {
	                      p=(edgetype *) malloc(sizeof(edgetype));
	                      p->into = p1->into + (p2->into)*nstates1;
	                      p->weight = weight;
			      p->label[0] = label;	
	                      p->next=wfa_out->state[k].edges;
	                      wfa_out->state[k].edges=p;
			  }
			}
                    }
               }
          }
       }
     }
     printf ("Done\n"); fflush (stdout);
     return (TRUE);
 }

int find_coordinates (address, x, y, size)
char	*address;
int	 *x, *y, size;
{
  int	i, j, s;

  if (*address == '\0') { *x = 0; *y = 0; return(size); }

  *x = *y = 0;
  for(s=size; *address != '\0'; address++, s /= 2) {
	switch (*address) {
		case '3': *x += s/2; break;
		case '0': *y += s/2; break;
		case '2': *x += s/2; *y += s/2; break;
        }
  }
  return (s);
}

void zoom_into_square (image, address, input_size, output_size)
imgtype		***image;
char		*address;
int 		input_size;
int		*output_size;
{
   int	i,j, ix, iy, x, y, size, step;
   imgtype **output;
   
  *output_size = DSIZE; 
  size = find_coordinates (address, &x, &y, input_size);

  output = alloc_image (*output_size, *output_size);

  step = (*output_size)/size;

  for (i=0; i < size; i++) {
	for (j=0; j<size; j++) { 
             for (ix=0; ix < step; ix++)
		  for (iy = 0; iy < step; iy++)
                       output[i*step+ix][j*step+iy] = (*image)[y+i][x+j];
       }
  }

 free_image (*image, input_size);
 *image = output;
}


void compute_weights (wfa, weights)
automaton	*wfa;
float		***weights;
{
   int	i, j, label, nstates;
   edgetype *ptr;

   nstates = wfa->states;

   for (label = 0; label <= 3; label++) {
	weights[label] = alloc_sq_matrix (nstates);

	for (i = 0; i < nstates; i++) {
	    for (j = 0; j < nstates; j++) 
		weights[label][i][j] = 0;
        }
    }
    
    for (i = 0; i < nstates; i++) {
	for (ptr = wfa->state[i].edges; ptr != NULL; ptr= ptr->next) {
	     label = ptr->label[0] - '0';
	     weights[label][i][ptr->into] = ptr->weight;
        }
    }
}

void wfa_zoom_into_square(wfa, address, init_distr)
automaton *wfa;
char	  *address;
float	  **init_distr;
{
   int  i, j, label, nstates;
   float  **weights[4], *result;
   edgetype *ptr;

   nstates = wfa->states;

   compute_weights (wfa, weights);

    for (i = 0; i < nstates; i++) 
        (*init_distr)[i] = wfa->state[i].initial;
    

    while(*address != '\0') {
        label = *address - '0';

	result = alloc_vect (nstates);

        vec_matrix_mult (*init_distr, weights[label], 
			  result, nstates, LEFT);

  	free(*init_distr);
  	*init_distr = result;
	address++;

  }
	
   for (label = 0; label <= 3; label++) free_matrix (weights[label], nstates);
   
}



int eps_free (aut)
automaton *aut;
{
  edgetype *ptr;
  int 	   i;

  for (i = 0; i < aut->states; i++) {
       for (ptr = aut->state[i].edges; ptr != NULL; ptr = ptr->next) {
	    if (ptr->label[0] == 'e') return (FALSE);
	    if (aut->ntapes == 2 && ptr->label[1] == 'e') 
		return (FALSE);
       }
  }

  return (TRUE);
}




/*****************************************************************
 * FUNCTION: new_edge()
 *
 * create a new edge from state p to state q
 * on labels l1, l2 with given weight
 *****************************************************************/

void new_edge (autom, p, q, l1, l2, weight)
automaton *autom;
int	  p,q ;
unsigned char      l1, l2;
float	  weight;
{
   	edgetype *newptr;

	newptr = (edgetype *) malloc (sizeof (edgetype));
	if (newptr == NULL) {
	    printf ("Cannot allocate memory for a new edge\n");
	    exit(1);
	}

	newptr->into = q;
	newptr->label[0] = l1;

	if (autom->ntapes == 2)
	    newptr->label[1] = l2;

	newptr->weight = weight;
	newptr->next = autom->state[p].edges;
	autom->state[p].edges = newptr;

        if (debug) {
	   if (autom->ntapes == 2) {
	       fprintf (fd, "new_edge(): New edge from state %d to state %d on labels %c %c, weight %f\n",
		        p, newptr->into, newptr->label[0], newptr->label[1], 
		        newptr->weight);
	    } else {
	       fprintf (fd, "new_edge(): New edge from state %d to state %d on label %c, weight %f\n",
		        p, newptr->into, newptr->label[0],  newptr->weight);
	    }
         }
}


/********************************************************************
 * WFA Minimalization
 *
 * The outline of the algorithm is given in the thesis, Chapter 6
 *
 ********************************************************************/

 int minimize_wfa (wfa, ortho_error, edge_error)
 automaton	*wfa;
 float		ortho_error, edge_error;
 {
   automaton	min_wfa;
   float	norm, **a, *b, d, max;
   float	**weights[4], ***base_vector, *new_vector, *remainder;
   float	**ortho_vector_init, *ortho_norm_init, 
	        **oc, *oc_norm, *std_basev, w;
   float	**ortho_vector_final, *ortho_norm_final; 
   edgetype	*ptr, **incoming_edges;
   int		i, j, k, l, n, label, nstates, *indx, *state_idx, 
		count, nremoved, nbase_init, nbase_final, nbase,
		nedges, *num_edges, orig_edges, choice; 

   edgetype     **find_incoming_edges ();
   void		find_ortho_basis (), count_edges ();

   nstates = wfa->states;

   /************** Preprocessing **************/

   /* Compute weight matrices */

   compute_weights (wfa, weights);

   if (debug) {
       for (i=0; i < 4; i++) {
            fprintf (fd,"Matrix W[%d]:\n", i);
	    print_matrix (fd,weights[i], nstates, nstates);
	    fprintf (fd,"---------------------------\n");
       }
   }

   /*
    * Allocate the base vectors, corresponding othogonal vectors 
    * and norms.
    */

    if ((base_vector = (float ***) malloc (nstates*sizeof (float **))) == NULL) {
	printf ("Memory allocation error\n"); exit (-1);
    }

    base_vector[0] = alloc_matrix (1, nstates);


    /**********************************************************
     * STEP 1 - Find the basis of the space 
     *		F = span ( W_w . F | w is a word over {0,1,2,3})
     *  	by generating the vectors in F and maintaing the
     *		base of orthogonal vectors found so far.
     *
     **********************************************************/

    if (!quiet_mode) {
       printf ("Computing the orthogonal basis ...\n"); fflush (stdout);
    }

    /* start with the space F - multiply from the right ...*/

    if ((ortho_vector_final = (float **) 
                               malloc (nstates*sizeof (float *))) == NULL) {
	printf ("Memory allocation error\n"); exit (-1);
    }

    ortho_norm_final = alloc_vect (nstates);
    ortho_vector_final[0] = alloc_vect (nstates);

    /* copy final distribution into the first base vector */
    for (i = 0; i < nstates; i++) {
	base_vector[0][0][i] = wfa->state[i].final;
	ortho_vector_final[0][i] = base_vector [0][0][i];
    }
    ortho_norm_final[0] = compute_norm (ortho_vector_final[0], nstates);
    nbase_final = 1;

    find_ortho_basis (ortho_vector_final, ortho_norm_final, base_vector, weights, 
		nstates, &nbase_final, RIGHT, ortho_error);

     if (!quiet_mode)
        printf ("Starting from final distribution - found %d base vectors\n", 
             nbase_final);

     if (debug)
         fprintf (fd, "Starting from final distribution - found %d base vectors\n", 
                 nbase_final);

       /*the same as above but for initial distr.*/

       base_vector[0] = alloc_matrix (1, nstates);
       if ((ortho_vector_init = (float **) 
                  malloc (nstates*sizeof (float *))) == NULL) {
	    printf ("Memory allocation error\n"); exit (-1);
       }

       ortho_norm_init = alloc_vect (nstates);
       ortho_vector_init[0] = alloc_vect (nstates);

       for (i = 0; i < nstates; i++) {
	   base_vector[0][0][i] = wfa->state[i].initial;
	   ortho_vector_init[0][i] = base_vector [0][0][i];
       }
       ortho_norm_init[0] = compute_norm (ortho_vector_init[0], nstates);
       nbase_init = 1;
    
       find_ortho_basis (ortho_vector_init, ortho_norm_init, base_vector, 
			weights, nstates, &nbase_init, LEFT, ortho_error);

     if (!quiet_mode)
     	printf ("Starting from initial distribution - found %d base vectors\n", 
             	nbase_init);

     if (debug) 
         fprintf (fd, "Starting from initial distribution - found %d base vectors\n", nbase_init);

      if (nbase_init < nbase_final) {
	  nbase = nbase_init;
	  choice = INITIAL;
          if (!quiet_mode) 
	     printf ("Choosing 'initial' base\n");
      } else {
	  nbase = nbase_final;
	  choice = FINAL;
          if (!quiet_mode) 
             printf ("Choosing 'final' base\n");
      }

    if (debug) 
        fprintf (fd, "Done. Number of base vectors = %d.\n", nbase); 

    if (!quiet_mode) {
        printf ("Done. Number of base vectors = %d.\n", nbase); fflush (stdout);
    }

    for (label = 0; label <= 3; label++) free_matrix (weights[label], nstates);

    /******************************************************
     * STEP 2 - Find the orthogonal complement of F in R^n 
     *		by applying the orthogonalization procedure
     *		to standard base vectors of R^n
     *		(n = nstates)
     ****************************************************/

     if (nbase < nstates) {
	
	/* generate the standard base of R^n */

	std_basev = alloc_vect (nstates);
        nremoved = nstates - nbase;
	oc = alloc_matrix(nremoved, nstates);
        oc_norm = alloc_vect (nremoved);
	new_vector = alloc_vect (nstates);
	remainder = alloc_vect (nstates);

        /* stores indices of columns that correspond to removed states */
	state_idx = alloc_int_array (nstates);
/*
        printf ("Number of states to remove : %d\n", nremoved);
        printf ("Generating the orthogonal complement...\n");
*/
	
	if (debug) {
            fprintf (fd,"Number of states to remove : %d\n", nremoved);
            fprintf (fd, "Generating the orthogonal complement...\n");
	}

	for (i = 0, n = 0, count = 0; i < nstates; i++) {

	     /* initialize the corresponding standard base vector */

	     for (j = 0; j < nstates; j++) 
		 std_basev[j] = (i == j)? 1 : 0;
		

	     if (debug) 
		fprintf (fd, "Currently considering base vector %d\n", i);

             switch (choice) {
		case INITIAL:
	             orthogonalize (ortho_vector_init, ortho_norm_init,
				    std_basev, new_vector, 
				    nstates, nbase);
		     break;

		case FINAL:
	             orthogonalize (ortho_vector_final, ortho_norm_final,
				    std_basev, new_vector, 
				    nstates, nbase);
		     break;
	       }

	     norm = compute_norm (new_vector, nstates);

	     if (debug) {
                fprintf (fd, "New vector : \n");
	        print_vector (fd, new_vector, nstates);
		fprintf (fd, "Norm : %f\n", norm);
	     }

	     if (sqrt((double) norm) > ortho_error && n < nremoved) {

		/* new vector is in the orthogonal complement */
		

                if (n == 0) { /* First vector */
		    copy_vect (new_vector, oc[n], nstates);
		    oc_norm[n] = norm;
		    state_idx[i] = -1;
		    if (debug) {
		          fprintf (fd, "i = %d : New vector in the orthogonal complement:\n", i);
		          print_vector (fd, oc[n], nstates);
			  fprintf (fd, "Norm : %f\n", oc_norm[n]);
		    }
                    n++;

		} else {   /* n > 0 */

		   orthogonalize (oc, oc_norm, new_vector, remainder,
				  nstates, n);
	           norm = compute_norm (remainder, nstates);

	           if (debug == 2) {
                	fprintf (fd, "Remainder : \n");
	        	print_vector (fd, remainder, nstates);
			fprintf (fd, "Norm : %f\n", norm);
		   }

                   if ( sqrt((double) norm) > ortho_error) {

		       if (debug) {
		          fprintf (fd, "New vector in the orthogonal complement:\n");
		          print_vector (fd, remainder, nstates);
		       }

		       copy_vect (remainder, oc[n], nstates);
		       oc_norm[n] = norm;
                       n++;
		       state_idx[i] = -1;

                    } else { /* already there */
		       state_idx[i] = count++;
		    }
		  }

              } else {  /* new vector not in OC */
		     state_idx[i] = count++;
              }
	   }

	     /* cleanup */

	   free_matrix (ortho_vector_init, nbase_init);
	   free (ortho_vector_init);
	   free (ortho_norm_init);
	   free_matrix (ortho_vector_final, nbase_final);
	   free (ortho_vector_final);
	   free (ortho_norm_final);
	   free (remainder);
	   free (new_vector);
	   free (std_basev);

	   if (!quiet_mode) {
              printf ("Done. Number of vectors in OC: %d\n", n);
	      fflush (stdout);
	   }

	   if (debug) {
                fprintf (fd, "Done. Number of vectors in OC: %d\n", n);
		fprintf (fd, "State index:\n");
		for (i=0; i<nstates; i++) 
		     fprintf (fd, "i = %d: %d\n", i, state_idx[i]);
           }

            /* Solve the linear system to find the coefficients of
	     * the linear combination. The linear system is solved
	     * by LU decomposition. The algorithm and the code 
	     * for LU decomposition is
	     * taken from "Numerical recipes in C". 
	     */
		
	     /* create an auxiliary matrix to store LHS */
             a = alloc_sq_matrix (nremoved+1);
	     indx = alloc_int_array (nremoved+1);
	     b = alloc_vect (nremoved+1);

	     if (!quiet_mode) {	
	        printf("Solving the linear system to find the lin. combination ...");
	        fflush (stdout);
	     }

	     if (debug)
	        fprintf(fd,"Solving the linear system to find the lin. combination ..."); 
	     /* Find max for each column. If |max| > 0, copy 
	      * column into the matrix a. Array state_idx[]
	      * keeps track of new state numbers and removed
	      * states.
	      */

	     for (j = 0, l=1; j < nstates; j++) {
		  if (state_idx[j] == -1) {
	              for (i = 1; i <= nremoved; i++) 
			   a[i][l] = oc[i-1][j];
		      l++;
		  }
	     }

	     if (debug) {
		fprintf (fd, "LHS matrix:\n");
		print_matrix (fd, a, nremoved+1, nremoved+1);
                fprintf (fd, "--------------------------\n");
             }
		
	     /* LU decmposition */
	     if (!ludcmp (a, nremoved, indx, &d)) {
	         free (b);
	         free_matrix (a, nremoved+1);
                 free (indx);
                 free_matrix (oc, nremoved);
		 return (FALSE);
	     }
		

	     /* solve for each base vector by backsubstitution */

	     for (i = 0; i < nstates; i++) {
		 if (state_idx[i] >= 0) {
		    /* copy ith column */
		    for (j = 1; j <= nremoved; j++) 
		         b[j] = (-1.0) * oc[j-1][i];

		    /* backsubstitution */
		    lubksb (a, nremoved, indx, b);

		    /* put it back */
		    for (j = 1; j <= nremoved; j++) 
		         oc[j-1][i] = b[j];
		 }
	     }
		 

	     /* cleanup */
	     free (b);
	     free_matrix (a, nremoved+1);
             free (indx);

	     if (!quiet_mode) printf ("Done\n");

	/******************************************************
	 * STEP 3 - Remove the states that do not belong to the
         *	    minimal WFA
         ******************************************************/

        /* initialize the minimal wfa */

        nedges = 0;
	min_wfa.states = nbase;
	min_wfa.ntapes = 1;

	for (i = 0; i < nbase; i++) { 
	    min_wfa.state[i].edges =  NULL;
	    min_wfa.state[i].new = NULL;
	}

	for (i = 0; i < nstates; i++) {
	     if ((k = state_idx[i]) >= 0) {
	        min_wfa.state[k].initial = wfa->state[i].initial;
	        min_wfa.state[k].final = wfa->state[i].final;
		for (ptr = wfa->state[i].edges;
		     ptr != NULL; ptr = ptr->next)  {
		     if (state_idx[ptr->into] >= 0) {
                        if (debug) {
                            printf ("Retained edge : from state %d to state %d, label %c, weight %f\n", k, state_idx[ptr->into], ptr->label[0], ptr->weight);
                            fprintf (fd, "Retained edge : from state %d to state %d, label %c, weight %f\n", k, state_idx[ptr->into], ptr->label[0], ptr->weight);
			}
			 
		         new_edge (&min_wfa, k, state_idx[ptr->into],
			           ptr->label[0], NULL, ptr->weight);
		         nedges++;
		     }
		}
	      }
        }

	if (!quiet_mode) {
           printf ("Minimal WFA starts with : %d edges\n", nedges);
           printf ("Number of states to remove: %d\n", nremoved); fflush (stdout);
	}

	if (debug) {
           fprintf (fd, "Minimal WFA starts with : %d edges\n", nedges);
           fprintf (fd, "Number of states to remove: %d\n", nremoved); fflush (stdout);
	}

	num_edges = alloc_int_array (nstates);

        if (choice == FINAL) {
    	   /* Find the list of incoming edges for every state */

   	   incoming_edges = find_incoming_edges (wfa, num_edges);

	   if (debug) 
               for (i = 0; i < nstates; i++) {
	            fprintf (fd, "State %d - number of incoming edges %d:\n", i, num_edges[i]);
	            for (ptr = incoming_edges[i]; ptr != NULL; ptr = ptr->next)
		         fprintf (fd, "edge from state %d to state %d, label %c, weight %f\n", ptr->from, i, ptr->label[0], ptr->weight);
	        }

        } else {
	    count_edges (wfa, num_edges);
        }
	     
	
	orig_edges = 0;
        for (i = 0; i < nstates; i++)  orig_edges += num_edges[i];

	for (i = 0, k=0;  i < nstates; i++) {
	     if (state_idx[i] < 0) {
/**
		printf ("Removing the state no. %d\n", i); 
		fflush (stdout);
		printf ("Number of incoming edges: %d\n", num_edges[i]);
**/		
		if (debug) {
		    fprintf (fd,"Removing the state no. %d\n", i); 
		    fprintf (fd,"Number of incoming edges: %d\n", num_edges[i]);
		}

                if (choice == FINAL) {

        	    /* adjust the incoming edges */

	            n = 0;
		    for (ptr = incoming_edges[i]; ptr != NULL; 
		         ptr = ptr->next) {
		         for (j = 0; j < nstates; j++) {
			     if (state_idx[j] >= 0 && 
			         state_idx[ptr->from] >= 0) {
				 w = ptr->weight * oc[k][j]; 
			         if (fabs(w) > edge_error) {
			             new_edge (&min_wfa, 
					       state_idx[ptr->from], 
					       state_idx[j], 
					       ptr->label[0], 
					       NULL, w);

				     if (debug) {
					printf ("Adding a new edge from state %d to state %d, label %c, weight %f\n", state_idx[ptr->from], state_idx[j], 
					ptr->label[0], w);
					fprintf (fd,"Adding a new edge from state %d to state %d, label %c, weight %f\n", state_idx[ptr->from], state_idx[j], 
					ptr->label[0], w);
				     }
		     		     nedges++; n++;
				  }
			   }
		      }
		  }
/***
	        printf ("Added %d edges, currently %d edges\n", n, nedges);
		fflush (stdout);
**/
		if (debug)
	            fprintf (fd, "Added %d edges, currently %d edges\n", n, nedges);

		/* adjust initial distribution */

		for (j = 0; j < nstates ; j++) {
                     if (state_idx[j] >= 0)
	    	        min_wfa.state[state_idx[j]].initial += 
				oc[k][j] * wfa->state[i].initial;
		 }

             } else {

                /* start with outgoing edges */

	            n = 0;
		    for (ptr = wfa->state[i].edges; ptr != NULL; 
		         ptr = ptr->next) {
		         for (j = 0; j < nstates; j++) {
			     if (state_idx[j] >= 0 && 
			         state_idx[ptr->into] >= 0) {
				 w = ptr->weight * oc[k][j]; 
			         if (fabs(w) > edge_error) {
			             new_edge (&min_wfa, 
					       state_idx[j], 
					       state_idx[ptr->into], 
					       ptr->label[0], 
					       NULL, w);

				     if (debug) {
					printf ("Adding a new edge from state %d to state %d, label %c, weight %f\n", state_idx[j], state_idx[ptr->into], 
					ptr->label[0], w);
					fprintf (fd,"Adding a new edge from state %d to state %d, label %c, weight %f\n", state_idx[j], state_idx[ptr->into], 
					ptr->label[0], w);
				     }
		     		     nedges++; n++;
				  }
			   }
		      }
		  }
/****
	        printf ("Added %d edges, currently %d edges\n", n, nedges);
		fflush (stdout);
****/
		if (debug)
	            fprintf (fd, "Added %d edges, currently %d edges\n", n, nedges);

		/* adjust initial distribution */

		for (j = 0; j < nstates ; j++) {
                     if (state_idx[j] >= 0)
	    	        min_wfa.state[state_idx[j]].final += 
				oc[k][j] * wfa->state[i].final;
		 }
               }
	       k++;
               if (!quiet_mode) printf ("Done\n");
	    }
	  }

          free_matrix (oc, nremoved);
          free (oc_norm);
          free_automaton (wfa);
          copy_automaton (&min_wfa, wfa);
	  printf("Original WFA : %d states, %d edges.\n"
		 "Minimal WFA : %d states, %d edges.\n", 
		  nstates, orig_edges, nbase, nedges);
	  if (debug)
	      fprintf(fd, "Original WFA : %d states, %d edges. \n minimal WFA : %d states, %d edges.\n", 
		  nstates, orig_edges, nbase, nedges);
      } else {
	  printf ("WFA is minimal. No states have been removed.\n");
	  if (debug)
              fprintf (fd, "WFA is minimal. No states have been removed.\n");
      }
      return (TRUE);
}


void find_ortho_basis (ortho_vector, ortho_norm, base_vector, weights, 
		       nstates, nbase, mult_type, epsilon)
float	**ortho_vector, *ortho_norm, ***base_vector, ***weights;
int	nstates, *nbase, mult_type;
float	epsilon;
{
	int  	j, depth, size, i, nadded, k, label, new_size;
	float	*new_vector, *remainder, norm;


    	new_vector = alloc_vect (nstates);
    	remainder = alloc_vect (nstates);
    	for (depth = 1, size = 1; depth < nstates; 
	     depth++, size *= 4) {

		new_size = size*4;
		if (debug)
        	    fprintf (fd, "Current size = %d, depth = %d.\n", new_size, depth); fflush (stdout);

		/* allocate base vectors for the given size */
		if ((base_vector[depth] = (float **) malloc (new_size * 
					  sizeof (float *))) == NULL) {
	  		printf ("Memory allocation error\n"); exit(-1);
		}

		for (i = 0; i < new_size; i++) 
	     	    base_vector[depth][i] = NULL;

		/*
	 	* Generate the vectors of F for the new  depth and check
	 	* for each vector if it belongs to the base.
         	* A vector W_aw.F is a candidate for the base only if the
	 	* vector W_w.F is in the base, where |w| = depth - 1.
         	*/ 

		for (i = 0, nadded = 0; i < size; i++) {
	     		if (base_vector[depth-1][i] != NULL) {
			for (label = 0; label < 4; label++) {
		   
		    	/* generate a new vector */

		    	vec_matrix_mult (base_vector[depth-1][i],
				     	weights[label], new_vector,
				     	nstates, mult_type);


			if (debug) {
		    	    fprintf (fd, "New vector:\n");
		    	    print_vector (fd, new_vector, nstates);
			}

		    	/* check if the new vector belongs to the base */

		    	orthogonalize (ortho_vector, ortho_norm, new_vector,
				   	remainder, nstates, *nbase);
		    	norm = compute_norm (remainder, nstates);

			if (debug) {
		    	    fprintf (fd,"Remainder:\n");
		    	    print_vector (fd,remainder, nstates);
		    	    fprintf (fd,"Norm = %f\n", norm);
			}

/*****************************************************************************
		    	if (sqrt((double) norm) > 0 && *nbase < nstates) {
			     k = 4*i + label;	
			     base_vector[depth][k] = alloc_vect (nstates);
			     copy_vect (new_vector, base_vector[depth][k], nstates);
			     if (debug) {
		    	         fprintf (fd, "kept non-base vector :\n"); 
		    	         print_vector (fd, base_vector[depth][k], nstates);
			     }
		             nadded++;
                        }
*************************************************************************/

		    	if (sqrt((double) norm) > epsilon && *nbase < nstates) {

		            /* add the vector to the orthogonal base */

			     k = 4*i + label;	

			     if (debug)
			         fprintf(fd, "New vector added to the orthogonal base: depth = %d, position = %d\n", depth, k); fflush (stdout);
                        
			     ortho_vector[*nbase] =  alloc_vect (nstates);
			     base_vector[depth][k] = alloc_vect (nstates);
			     copy_vect (new_vector, base_vector[depth][k], nstates);
			     copy_vect (remainder, ortho_vector[*nbase], nstates);
			     ortho_norm[*nbase] = norm;

			     if (debug) {
			         for (j = 0; j < *nbase; j++) {
			             fprintf (fd,"Ortho-base vector %d:\n", j);	
			             print_vector(fd, ortho_vector[j], nstates);
			             fprintf (fd,"--------------------------\n");
			         }

		                 fprintf (fd, "New ortho base vector: \n");
			         print_vector (fd, ortho_vector[*nbase], nstates);
			         fprintf (fd, "norm = %f\n", ortho_norm[*nbase]);
		    	         fprintf (fd, "Corresponding Base vector :\n"); 
		    	         print_vector (fd, base_vector[depth][k], nstates);
			     }
			     (*nbase)++;
		             nadded++;
		         }

                         if(*nbase == nstates) {
                            if (nadded > 0) {
                                for (j = 0; j < new_size; j++) 
	                            if (base_vector[depth][j] != NULL)
			                free (base_vector[depth][j]);
	                            free (base_vector[depth]);
                             }
                             for (j = 0; j < size; j++) 
	                          if (base_vector[depth-1][i] != NULL)
		                    free (base_vector[depth-1][i]);
	                     free (base_vector[depth-1]);
	                     free (new_vector);
	                     free (remainder);
	                     return;
                          }
		}
		free (base_vector[depth-1][i]);
             }
	}
	if (debug)
            fprintf (fd,"Number of base vectors = %d. (%d vectors added in the last step)\n", *nbase, nadded);

/***
        printf ("Number of base vectors = %d. (%d vectors added in the last step)\n", *nbase, nadded);
***/
	free (base_vector[depth-1]);

	/* check if we need to exit */
        if (nadded == 0) {
            free (base_vector[depth]);
	    free (new_vector);
	    free (remainder);
	    return;
        }
    }


     /* cleanup -- if the loop is exited regularly */

     for (i = 0; i < size; i++) 
	  if (base_vector[depth-1][i] != NULL)
		free (base_vector[depth-1][i]);
     
     free (base_vector[depth-1]);
     free (new_vector);
     free (remainder);
}


edgetype **find_incoming_edges (wfa, num_edges)
automaton	*wfa;
int		*num_edges;
{
  edgetype	*ptr, *newptr, **incoming_edges;
  int		i, nstates;

  nstates = wfa->states;

    incoming_edges = (edgetype **) malloc (nstates*sizeof(edgetype*));
    if (incoming_edges == NULL) {
	printf ("Memory allocation error\n"); exit (-1);
   }
   for (i = 0; i < nstates; i++) {
	incoming_edges[i] = NULL;
 	num_edges[i] = 0;
   }

   for (i = 0; i < nstates; i++) {
   	for (ptr = wfa->state[i].edges; ptr != NULL; ptr = ptr->next) {

	     newptr = (edgetype *) malloc (sizeof (edgetype));
	     if ( newptr == NULL) {
		printf ("Memory allocation error\n"); exit (-1);
   	     }
	     
	     newptr->from = i; 
	     newptr->weight = ptr->weight;
	     newptr->label[0] = ptr->label[0];
	     newptr->next = incoming_edges[ptr->into];
	     incoming_edges[ptr->into] = newptr;
	     num_edges[ptr->into]++;
         }
   }
   return (incoming_edges);
}

void count_edges (wfa, num_edges)
automaton	*wfa;
int		*num_edges;
{
  edgetype	*ptr;
  int		i;

   for (i = 0, num_edges[i] = 0; i < wfa->states; i++) 
   	for (ptr = wfa->state[i].edges; ptr != NULL; 
	     num_edges[i]++, ptr = ptr->next); 
}

