/************************************************************************
 * FILE:  read_image.c                                                	*
 *        A generic function to read a color image in PPM format      	*
 *	  or a grey-scale image in the PGM or raw format              	*
 *								      	*
 * Authors: Peter Rajcani and Mike Hilton				*
 ************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>

#define MAXSIZE	   512*512
#define RAW 0
#define PBM 1   /* portable bit map ID */
#define PGM 2   /* portable gray map ID */
#define PPM 3   /* portable pixel map ID */

typedef  float  imgtype;
extern imgtype	**alloc_image ();

/***********************************************
 * Function: read_grey_image                   *
 * 	     Read the input grey scale image   *
 ***********************************************/
	 
void read_grey_image (name, input_vector, image, rows, columns, 
		      size, depth, type)
char *name;
unsigned char  *input_vector;
imgtype ***image;
int  *rows, *columns, *size, *depth, type;
{
 unsigned int	m;
 int		i, j, k, N;
 FILE		*infile;
 int	        c;
 float		sz;
 unsigned char  pixel;

 sz = (double) ((*rows >= *columns)? *rows : *columns);
 *depth = ceil (log(sz)/ log ((double) 2.0));
 for (*size = 1, i = 0; i < *depth; i++, *size *= 2);
 N = (*rows)*(*columns);

 *image = alloc_image (*size, *size);

 for(i = k = 0; i < *size; i++) {
    for(j = 0; j < *size; j++) {
	if (i < *rows && j < *columns) {
           pixel = input_vector[k++];
	   (*image) [i][j] = (imgtype) pixel;
           /*printf ("input_vector : %d, image[%d][%d] = %f\n", 
                   pixel, i, j, (*image)[i][j]);*/
	} else
	   (*image) [i][j] = (imgtype) 0;
    }
 }
	  
 printf("Size of image = %d (%d x %d) bytes\n",N, *rows,*columns);
 printf("Depth = %d\n",*depth);
 return;
}


/***********************************************
 * Function: read_color_image                   *
 * 	     Read the input grey scale image   *
 ***********************************************/
	 
void read_color_image (input_vector, image_red, image_green, image_blue, 
		      rows, columns, size, depth)
unsigned char  *input_vector;
imgtype ***image_red, ***image_green, ***image_blue;
int  rows, columns, *size, *depth;
{
 int		i, j, k, N;
 float		sz;

 sz = (double) ((rows >= columns)? rows : columns);
 *depth = ceil (log(sz)/ log ((double) 2.0));
 for (*size = 1, i = 0; i < *depth; i++, *size *= 2);
 N = 3*rows*columns;

 *image_red = alloc_image (*size, *size);
 *image_green = alloc_image (*size, *size);
 *image_blue = alloc_image (*size, *size);

 for(i = k = 0; i < *size; i++) {
    for(j = 0; j < *size; j++) {
	if (i < rows && j < columns) {
	   (*image_red) [i][j] = (imgtype) input_vector[k++];
	   (*image_green) [i][j] = (imgtype) input_vector[k++];
	   (*image_blue) [i][j] = (imgtype) input_vector[k++];
	} else {
	   (*image_red) [i][j] = (imgtype) 0;
	   (*image_green) [i][j] = (imgtype) 0;
	   (*image_blue) [i][j] = (imgtype) 0;
        }
    }
 }
	  
 printf("Size of image = %d x %d, depth = %d\n", rows,columns, *depth);
 return;
}


/* PNM file read and write utilities
/*
/* Mike Hilton, 28 Jan 1993
/*
/* This file contains functions for reading and writing portable
/* any map (PNM) image files.  For more information on PNM image
/* files and formats, see the PBM-Plus utilities documentation.
/*
/* YOU MUST LINK THIS FILE WITH THE MATH LIBRARY.  USE -lm IN YOUR
/* CC COMMAND LINE!
*/

/* READ_PNM_FILE
/* Given a FILENAME, open the file and, if the file is a PNM file,
/* read its contents into a malloc'ed array of type unsigned char.
/* The type and size of the image are returned in the parameters 
/* TYPE, ROWS, and COLS, respectively.  Currently, three file types
/* are supported:
/*   PBM, portable bit maps (TYPE == 1)
/*   PGM, portable grayscale maps (TYPE == 2)
/*   PPM, portable pixel maps (TYPE == 3)
/*
/* For PBM files, the returned array stores the bitmaps in row major
/* format, each byte holding 8 pixels.  The leftmost pixel is in the
/* most significant bit of each byte.
/*
/* For PGM files, the returned array stores the 8-bit grayscale values
/* in row major order.
/*
/* For PPM files, the returned array stores each pixel as three 
/* consecutive bytes, representing the pixel's 8-bit red, blue, and green 
/* components, respectively.  The pixels are stored in row major order.
/*
/*
/* If there is an error reading the file, the function calls EXIT.
*/


unsigned char *read_pnm_file(char *filename, int *type,
			     int *rows, int *cols)
{
  char  pmode, c, firstchar;
  int   msize, size, N, x;
  int red, green, blue;
  float bytes_per_pixel;
  FILE  *fp;
  char unsigned *raw, *ptr;
  size_t bytes_read;


  /* try to open the file */
  if ((fp = fopen(filename, "rw")) == NULL) {
    fprintf(stderr, "read_pnm_file: error loading image from %s.\n", 
	    filename);
    perror("read_pnm_file");
    return (NULL);
  }

  /* Check for the correct magic number */
  fscanf(fp, "%c\n", &firstchar);
  if (firstchar != 'P') {
    fputc (firstchar, fp);
    *type = RAW;
    /* gray scale file is without header */
	if ((raw = (unsigned char *) malloc (MAXSIZE)) == NULL) { 
		printf ("Error in allocating image array\n"); exit(-1); }
	for (N = 0; (x = fgetc (fp)) != EOF && N < MAXSIZE; N++) 
             raw[N] = (unsigned char) x;
	fclose (fp);
        *rows = *cols = sqrt ((double) N);
    return (raw);
  }

  fscanf(fp, "%c\n", &pmode);
  switch (pmode) {
  	case '4': *type = PBM; bytes_per_pixel = 0.125; break;
  	case '5': *type = PGM; bytes_per_pixel = 1; break;
  	case '6': *type = PPM; bytes_per_pixel = 3; break;
  	default:
    		fprintf(stderr, "read_pnm_file: unrecognized magic number: %c\n", pmode);
		return (NULL);
  }

  /* throw out any comments */
  while ((c = getc(fp)) == '#') 
    while (!feof(fp) && ((c = getc(fp)) != '\n'));
  ungetc(c, fp);
  if (feof(fp)) {
    fprintf(stderr, "read_pnm_file: unexpected end of file in %s!\n", 
	    filename);
    return (NULL);
  }
  
  /* read the header information */
  fscanf(fp,"%d %d%c", cols, rows, &c);
  if (*type != PBM) fscanf(fp, "%d%c", &msize, &c);

  /* read in data */
  size = ceil((double)((*rows) * (*cols) * bytes_per_pixel));
  raw = (unsigned char *) malloc(size);
  if (raw == NULL) {
    fprintf(stderr, "read_pnm_file: unable to malloc memory for image %s!\n",
	    filename);
    exit(1);
  }
  bytes_read = fread(raw, sizeof(unsigned char), size, fp);
  if (bytes_read != size) {
    fprintf(stderr, 
	    "\nread_pnm_file: Warning: file %s not of expected size!\n", 
	    filename);
  }

  return(raw);
}

