/**************************************************************************
 *  Main driver of the image manipulation system based on weighted finite *
 *  transducers. Implementation in X-windows with OSF/Motif graphical 	  *
 *  user interface. The program handles both grayscale and color images.  *
 *  For more information, see the accompanying README file.		  *
 *									  *
 * Usage:                                                                 *
 *	   wftx  [<input image file>] 		       		     	  *
 *									  *
 * Author: Peter Rajcani, March 1995.					  *
 *									  *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <math.h>

#include "xm.h"
#include "wft.h"
#include "const.h"
#include "resource.h"

/* EXTERNAL FUNCTIONS */

/* defined in autom.c */

extern void	write_automaton (),
		write_color_aut (),
		read_grey_image(),
		read_color_image(),
		output_image (),
		output_image_id (),
		decode_automaton (),
		copy_automaton (),
		compose (),
		addwfa (),
		free_automaton (),
		invert_wft (),
		free_autom_values (),
       		zoom_into_square (),
       		wfa_zoom_into_square (),
		display_aut ();

extern int	read_automaton(), 
		read_color_automaton (),
		concatenate (), eps_free (),
		minimize_wfa (),
		cart_product ();

extern unsigned char *read_pnm_file ();
extern imgtype  **alloc_image ();
extern void	free_image(), copy_image ();
extern float	*alloc_vect ();

/* defined in xutils.c */

 extern void 	xs_create_menu_buttons (),
		display_file_sb (),
		display_dialog (),
		set_label (),
  		quit (), 
		error (), crop_name (),
		create_dialog_box (),
		flush_Xbuffer(); 

extern Widget   create_button_dialog (), editor [MAXFIELDS];


/* Menu functions */

void	load_image(), load_single_wfa(), load_wfa_color(), load_3wfa (),
        load_rgb_wfa (), save_image(), save_wfa(), save_pgm (), 
        save_raw (), print_image (), unroll_wfa (),
        shrink (), change_init_distr (),
	load_wft(), apply(), original_image (), cproduct (),
        composition(), save_wft(), change_contrast (),
        fliph (), flipv (), rotate (), zoom (), crop (), revert (), 
	inverse_wft (), iter_wft (), scale_values (),
	uncrop (), concat_images (), redisplay (), undo (), 
	load_red_wft (), load_green_wft (), load_blue_wft (), 
	load_orig_wfa (), load_3wft (), load3 (), sum (), 
	mchange (), mchange_red(), mchange_blue (), mchange_green (),
	zoom_square (), concatenation (), set_mode_im (),
	set_mode_wfa (), set_mode_wft (), undo_op (), change_final_distr (),
	free_crop_image (), rot_image (), scale (), resize_img (), 
	min_wfa (), find_wft_matrix ();

/* Other functions */

void	clear_wfa_labels (), clear_wft_labels (), adjust_image ();
int	power_of_2 ();
int	save_img ();
void 	expose (), resize (), build_grey_table (); 
void	create_grey_image ();
void 	mouse_click (), mouse_drag (), mouse_release ();
int	create_color_image ();
void	xv_display (), find_size ();
unsigned char    imgtype_to_char ();



/* Create list of menu items */

static xs_menu_struct ImageMenu[] = {
	{"LOAD IMAGE", load_image, "LOAD IMAGE", NULL, 0, NULL},
	{"LOAD COLOR IMAGE (3 FILES)", load3, 
         "LOAD COLOR IMAGE (3 FILES)", NULL, 0, NULL},
	/*{"DISPLAY IMAGE", redisplay, "DISPLAY", NULL, 0, NULL},*/
	{"ORIGINAL IMAGE", original_image, "ORIGINAL IMAGE", NULL, 0, NULL},
	{"REVERT COLORS", revert, "REVERT COLORS", NULL, 0, NULL},
	{"SAVE IMAGE", save_image, "SAVE IMAGE", NULL, 0, NULL},
	{"QUIT", quit, "QUIT", NULL, 0, NULL}
};


static xs_menu_struct TransformMenu[] = {
	{"FLIP HORIZONTAL", fliph, "FLIP HORIZONTAL", NULL, 0, NULL},
	{"FLIP VERTICAL", flipv, "FLIP VERTICAL", NULL, 0, NULL},
	{"ROTATE (90 DEG.)", rotate, "ROTATE (90 DEG.)", NULL, 0, NULL},
	{"ROTATE", rot_image, "ROTATE", NULL, 0, NULL},
	{"SCALE", scale, "SCALE", NULL, 0, NULL},
	{"RESIZE", resize_img, "RESIZE", NULL, 0, NULL},
	{"CONTRAST", change_contrast, "CONTRAST", NULL, 0, NULL},
	{"CROP", crop, "CROP", NULL, 0, NULL},
	{"UNCROP", uncrop, "UNCROP", NULL, 0, NULL},
	{"DOUBLE SIZE", zoom, "DOUBLE SIZE", NULL, 0, NULL},
	{"HALF SIZE", shrink, "HALF SIZE", NULL, 0, NULL},
	{"UNDO", undo, "UNDO", NULL, 0, NULL}
};


static xs_menu_struct WFALoadMenu[] = {
	{"RGB FILES", load_rgb_wfa, "RGB FILES", NULL, 0, NULL},
	{"INDEPENDENT FILES", load_3wfa, "INDEPENDENT FILES", NULL, 0, NULL}
};

static xs_menu_struct WFAmenu[] = {
	{"LOAD WFA (1 FILE)", load_single_wfa, "LOAD WFA", NULL, 0, NULL},
	{"LOAD WFA (3 FILES)", NULL, "LOAD WFA (3 FILES)", WFALoadMenu,
	  XtNumber (WFALoadMenu), NULL},
	{"DECODE WFA", unroll_wfa, "DECODE WFA", NULL, 0, NULL},
	{"MINIMIZE WFA", min_wfa, "MINIMIZE WFA", NULL, 0, NULL},
	{"CARTESIAN PRODUCT", cproduct, "CART. PRODUCT", NULL, 0, NULL},
	{"ORIGINAL WFA", load_orig_wfa, "ORIGINAL WFA", NULL, 0, NULL},
	{"SAVE WFA", save_wfa, "SAVE", NULL, 0, NULL}
};

static xs_menu_struct OperationsMenu[] = {
	{"APPLY WFT", apply, "APPLY WFT", NULL, 0, NULL},
	{"SUM", sum, "SUM", NULL, 0, NULL},
	{"CONCATENATION", concatenation, "CONCATENATION", NULL, 0, NULL},
	{"ZOOM", zoom_square, "ZOOM", NULL, 0, NULL},
	{"UNDO", undo_op, "UNDO", NULL, 0, NULL}
};

static xs_menu_struct WFTLoadMenu[] = {
	{"LOAD RED WFT", load_red_wft, "LOAD RED WFT", NULL, 0, NULL},
	{"LOAD GREEN WFT", load_green_wft, "LOAD GREEN WFT", NULL, 0, NULL},
	{"LOAD BLUE WFT", load_blue_wft, "LOAD BLUE WFT", NULL, 0, NULL},
	{"LOAD THREE WFT", load_3wft, "LOAD THREE WFT", NULL, 0, NULL}
};


static xs_menu_struct MethodMenu[] = {
	{"SINGLE WFT", mchange, "SINGLE WFT", NULL, 0, NULL},
	{"RED WFT", mchange_red, "RED WFT", NULL, 0, NULL},
	{"GREEN WFT", mchange_green, "GREEN WFT", NULL, 0, NULL},
	{"BLUE WFT", mchange_blue, "BLUE WFT", NULL, 0, NULL}
};


static xs_menu_struct WFTmenu[] = {
	{"LOAD WFT (SINGLE)", load_wft, "LOAD WFT (SINGLE)", NULL, 0, NULL},
	{"LOAD WFT (MULTIPLE)", NULL, "LOAD WFT (MULTIPLE)", 
	  WFTLoadMenu, XtNumber (WFTLoadMenu), NULL},
	{"COMPOSITION", composition, "COMPOSITION", NULL, 0, NULL},
	{"ITERATE WFT", iter_wft, "ITERATE WFT", NULL, 0, NULL},
	{"CHANGE METHOD", NULL, "CHANGE METHOD", MethodMenu, 
	  XtNumber (MethodMenu), NULL},
	{"INVERSE WFT", inverse_wft, "INVERSE WFT", NULL, 0, NULL},
	{"CHANGE INITIAL DISTR.", change_init_distr, 
	"CHANGE INITIAL DISTR.", NULL, 0, NULL},
	{"CHANGE FINAL DISTR.", change_final_distr, 
	"CHANGE FINAL DISTR.", NULL, 0, NULL},
	{"SAVE WFT", save_wft, "SAVE", NULL, 0, NULL}
};


static xs_menu_struct ModeMenu[] = {
	{"IMAGE", set_mode_im, "IMAGE", NULL, 0, NULL},
	{"WFA", set_mode_wfa, "WFA", NULL, 0, NULL},
	{"WFT", set_mode_wft, "WFT", NULL, 0, NULL}
};


/* Describe the menu bar */

static xs_menu_struct PulldownData [] = {
	{"IMAGE", NULL, NULL, ImageMenu, XtNumber (ImageMenu), NULL},
	{"WFA", NULL, NULL, WFAmenu, XtNumber (WFAmenu), NULL},
	{"WFT", NULL, NULL, WFTmenu, XtNumber (WFTmenu), NULL},
	{"MODE", NULL, NULL, ModeMenu, XtNumber (ModeMenu), NULL},
	{"OPERATION", NULL, NULL, OperationsMenu, 
         XtNumber (OperationsMenu), NULL},
	{"TRANSFORMATION", NULL, NULL, TransformMenu, 
	 XtNumber (TransformMenu), NULL}
};


typedef struct {
	int	x, y;
	int	width, height;
} rectangle;

typedef struct {
	unsigned short	red, green, blue;
} color_pixels;

/* GLOBAL VARIABLES */

Display		*theDisplay;
Colormap	theColormap;
GC		theGC;
int		theScreen;
Widget		top_level, main_window, pop_window,
                save_dialog, form, canvas, menubar, board,
		wfa_label, wft_label, red_label, blue_label, green_label,
		red_wfa_label, blue_wfa_label, green_wfa_label,
		crop_label, value_label, mlabel, mlabel_red, 
		mlabel_green, mlabel_blue, mode_label;
XImage		*image;
rectangle	subimg;

automaton      	wfa, wft, wft_aux, wfa_red, wfa_green, wfa_blue, 
		orig_wfa, wft_red, wft_green, wft_blue, wfa1_aux,
		orig_wfa_red, orig_wfa_green, orig_wfa_blue;
imgtype  	**raw_image = NULL, **crop_image = NULL, **orig_image = NULL; 
imgtype  	**raw_image_red = NULL, **crop_image_red = NULL, 
		**orig_image_red = NULL; 
imgtype  	**raw_image_green = NULL, **crop_image_green = NULL, 
		**orig_image_green = NULL; 
imgtype  	**raw_image_blue = NULL, **crop_image_blue = NULL, 
		**orig_image_blue = NULL; 
imgtype   	**quant_image_red = NULL, **quant_image_green = NULL, 				**quant_image_blue = NULL;
imgtype   	**prev_image = NULL, **prev_image_red = NULL, 
		**prev_image_green = NULL, **prev_image_blue = NULL;
char            *conv_image, gtable[256]; 
int		rows, columns, size, depth, xtop, ytop;
int		orig_rows, orig_columns, orig_size, orig_depth;
int		quant_rows, quant_columns, quant_size;
int		owidth, oheight, odepth, osize, wft_depth, wfa_depth,
		usize;
int		rrows, rcolumns, rsize, rdepth;
int		grows, gcolumns, gsize, gdepth;
int		brows, bcolumns, bsize, bdepth;
int		wfa_selected = FALSE;
int		color_wfa_selected = FALSE;
int		red_wfa_selected = FALSE;
int		green_wfa_selected = FALSE;
int		blue_wfa_selected = FALSE;
int		wft_selected = FALSE, wft_unrolled = FALSE;
int		red_wft_selected = FALSE, red_wft_unrolled = FALSE;
int		green_wft_selected = FALSE, green_wft_unrolled = FALSE;
int		blue_wft_selected = FALSE, blue_wft_unrolled = FALSE;
int		wft_started = FALSE;
int		wfa_load_type, wft_load_type, nloaded = 0, action_type;
int		image_flag = FALSE, orig_flag;
int		crop_flag = FALSE, rect_selected = FALSE;
int		syn_flag = FALSE, adj_flag = FALSE;
int		debug = FALSE, debug2 = FALSE, show_values = FALSE;
float		ScaleV, ScaleH;
FILE		*fd;
unsigned long   pixels[PIXELS];
unsigned long   plane_masks[PLANES];
unsigned int    nplanes = 0;
unsigned int    ncolors;
int		color_flag = FALSE, color_aut = FALSE, 
		scaling_flag = FALSE, resize_flag = FALSE;
char		*orig_color_wfa_name, *orig_wfa_name, *orig_red_wfa_name;
char		*orig_green_wfa_name, *orig_blue_wfa_name;
int		prows, pcolumns, psize, pdepth;
int		transf_done, mode = IMAGE, test_img_loaded = FALSE;
int		resize_activated = FALSE;
int		xv_flag = FALSE, xv_disp_flag = FALSE, 
		resource_file = FALSE, cutoff_flag = FALSE;

/****** main routine *****************/

main(argc, argv) 
int argc; char *argv[]; 
{
 char 		*imagefile;
 unsigned char  *input_vector;
 int  		type, i, j, n, top, step;
 Arg  		wargs[10];
 Status		result;
 int 		sw, img_ok, success, nlabels;
 Dimension	width, height, labelh;
 XmString	xmstr;

 char *switches[] = {"-x", "-r", "-v", "-rs"};

 /* 
  * Initialize and setup the X-window interface
  */
  

  top_level = XtInitialize (argv[0], "wftx", NULL, 0, &argc, argv);
  main_window = XmCreateMainWindow (top_level, "Main_window", NULL, 0);

 /*
  * Parse the command line
  */

rows = columns = 512;
img_ok = FALSE;
image_flag = FALSE;

  if (argc > 1) {
/*************************
	for (i = 1; i<argc; i++) {
	    for (sw = 0; sw < XtNumber(switches); sw++)
		 if (!strcmp (argv[i], switches[sw])) break; 
	    switch (sw) {
		   case 0:	xv_flag = TRUE; break;
		   case 1:	resize_activated = TRUE; break;
		   case 2:	debug = TRUE; break;
		   case 3:	resource_file = TRUE; break;
		   default:	imagefile = argv[i];  
				img_ok = TRUE; break;
             }
         }
*********************/
   imagefile = argv[1];  
   img_ok = TRUE; 
  } 

 if (debug) {
    if((fd = fopen ("wft.debug", "w")) == NULL) {
       printf ("Cannot open the wft.debug file ! Exiting the debug mode.\n");
       debug = FALSE;
    }
  }


 /*
  * read the input image
  */

 if (img_ok) {
     if ((input_vector = read_pnm_file (imagefile, &type, &orig_rows, 
					&orig_columns)) == NULL) {
          img_ok = FALSE;
	  image_flag = FALSE;
	  rows = columns = 512;

     } else {

     	if (type == RAW || type == PGM) {
	 	color_flag = FALSE;
	 	orig_flag = GREY;
         	read_grey_image (imagefile, input_vector, &orig_image, 
                          	&orig_rows, &orig_columns, 
                          	&orig_size, &orig_depth, type);

		raw_image = alloc_image (orig_size, orig_size);
		copy_image (orig_image, raw_image, orig_size, orig_size);

     	}

     	if (type == PPM) {
		color_flag = TRUE;
        	orig_flag = COLOR;
		read_color_image (input_vector, &orig_image_red, &orig_image_green,
			  &orig_image_blue, orig_rows, orig_columns, &orig_size,			  &orig_depth);

		raw_image_red = alloc_image (orig_size, orig_size);
		copy_image (orig_image_red, raw_image_red, orig_size, orig_size);
		raw_image_blue = alloc_image (orig_size, orig_size);
		copy_image (orig_image_blue, raw_image_blue, orig_size, orig_size);
		raw_image_green = alloc_image (orig_size, orig_size);
		copy_image (orig_image_green, raw_image_green, orig_size, orig_size);
      }

      rows = orig_rows; columns = orig_columns; 
      size = orig_size; depth = orig_depth;
      xtop = ytop = 0;
      free (input_vector);
    }
  }


  /**************************
   * Setup widgets
   **************************/

  nlabels = (xv_flag) ? 9 : 11;

  /* Bulletin board manager widget */


   board = XtCreateManagedWidget ("board", xmBulletinBoardWidgetClass, 
				     main_window, NULL, 0);

   if (resource_file) {
   	n = 0;
   	XtSetArg ( wargs[n], XmNwidth, &width); n++;
   	XtSetArg ( wargs[n], XmNheight, &height); n++;
   	XtGetValues (board, wargs, n);

        printf ("Using a resource file: width = %d, height = %d\n", width, height);

   	if (width == 0 || height == 0) {
      	    printf ("Warning : Resource file not set properly. Using default values.\n");
      	    resource_file = FALSE;
        }
    }
   
   if (!resource_file) {
      n = 0;
      XtSetArg ( wargs[n], XmNwidth, BOARD_WIDTH); n++;
      XtSetArg ( wargs[n], XmNheight, (LABEL_HEIGHT+BORDER_WIDTH)*nlabels); n++;
      XtSetValues (board, wargs, n);
      /*printf ("Default: width = %d, height = %d\n", BOARD_WIDTH, nlabels*(LABEL_HEIGHT+BORDER_WIDTH));*/
   }

  /* Labels */

  mode_label = XtCreateManagedWidget ("mode_label", 
                 xmLabelWidgetClass, board, NULL, 0); 

  if (resource_file) {
       n = 0;
       XtSetArg ( wargs[n], XmNheight, &labelh); n++;
       XtGetValues (mode_label, wargs, n);
  } else {
       n = 0;
/*
       XtSetArg ( wargs[n], XmNwidth, BOARD_WIDTH-MSG_WIDTH); n++;
       XtSetArg ( wargs[n], XmNheight, LABEL_HEIGHT); n++;
*/
       XtSetArg ( wargs[n], XmNy, 0); n++;
       XtSetValues (mode_label, wargs, n);
       labelh = LABEL_HEIGHT;
       /*printf ("labelh = %d\n", labelh); fflush (stdout);*/
  }

  top = step = labelh + BORDER_WIDTH;

  for (i = 1; i < nlabels; i++, top += step) {
       /* create labels */
       n = 0;
/*
       if (!resource_file) {
           XtSetArg ( wargs[n], XmNwidth, BOARD_WIDTH-MSG_WIDTH); n++;
           XtSetArg ( wargs[n], XmNheight, LABEL_HEIGHT); n++;
       }
*/
       /*printf ("top = %d \n", top); fflush (stdout);*/
       XtSetArg ( wargs[n], XmNy, top); n++;
       switch (i) {
	      case 1:
                     wfa_label = XtCreateManagedWidget ("wfa_label", 
                                      xmLabelWidgetClass,  
                                      board, wargs, n); break;
	      case 2:
  		     red_wfa_label = XtCreateManagedWidget ("red_wfa_label", 
					xmLabelWidgetClass,  
                                      	board, wargs, n); break;

	      case 3:
  		     green_wfa_label = XtCreateManagedWidget ("green_wfa_label",
					xmLabelWidgetClass,  
                                      	board, wargs, n); break;

	      case 4:
  		     blue_wfa_label = XtCreateManagedWidget ("blue_wfa_label", 
					xmLabelWidgetClass,  
                                      	board, wargs, n); break;


	      case 5:
  		     wft_label = XtCreateManagedWidget ("wft_label", 
					xmLabelWidgetClass,  
                                      	board, wargs, n); break;

	      case 6:
  		     red_label = XtCreateManagedWidget ("red_label", 
					xmLabelWidgetClass,  
                                      	board, wargs, n); break;

	      case 7:
  		     green_label = XtCreateManagedWidget ("green_label",
					xmLabelWidgetClass,  
                                      	board, wargs, n); break;

	      case 8:
  		     blue_label = XtCreateManagedWidget ("blue_label", 
					xmLabelWidgetClass,  
                                      	board, wargs, n); break;

	      case 9:
  		     crop_label = XtCreateManagedWidget ("crop_label", 
					xmLabelWidgetClass,  
                                        board, wargs, n); break;
	      case 10:
                     value_label = XtCreateManagedWidget ("value_label",
				 	xmLabelWidgetClass,  
                                        board, wargs, n); break;
	}
  }

  for (top = 5*step, i = 1; i <= 4; i++, top += step) {
       /* create labels */
       n = 0;
/*
       XtSetArg ( wargs[n], XmNwidth, MSG_WIDTH); n++;
       if (!resource_file)
           XtSetArg ( wargs[n], XmNheight, LABEL_HEIGHT); n++;
*/
       XtSetArg ( wargs[n], XmNy, top); n++;
       XtSetArg ( wargs[n], XmNx, BOARD_WIDTH-MSG_WIDTH); n++;
       switch (i) {
	      case 1:
                     mlabel = XtCreateManagedWidget ("mlabel", 
                                      xmLabelWidgetClass,  
                                      board, wargs, n); break;

	      case 2:
                     mlabel_red = XtCreateManagedWidget ("mlabel_red", 
                                      xmLabelWidgetClass,  
                                      board, wargs, n); break;

	      case 3:
                     mlabel_green = XtCreateManagedWidget ("mlabel_green", 
                                      xmLabelWidgetClass,  
                                      board, wargs, n); break;

	      case 4:
                     mlabel_blue = XtCreateManagedWidget ("mlabel_blue", 
                                      xmLabelWidgetClass,  
                                      board, wargs, n); break;


	}
  }

  set_label (mode_label, "IMAGE MODE");
  set_label (wfa_label, "Loaded WFA: ");
  set_label (wft_label, "Loaded WFT: ");
  set_label (red_label, "");
  set_label (green_label, "");
  set_label (blue_label, "");
  set_label (red_wfa_label, "");
  set_label (green_wfa_label, "");
  set_label (blue_wfa_label, "");
  set_label (mlabel, "");
  set_label (mlabel_red, "");
  set_label (mlabel_green, "");
  set_label (mlabel_blue, "");

  if (!xv_flag) {
      set_label (crop_label, "");
      set_label (value_label, "");
  }

  /* Create the menus */
  menubar = XmCreateMenuBar (main_window, "menubar", NULL, 0);
  XtManageChild (menubar);
  xs_create_menu_buttons (NULL, menubar, PulldownData, 
                          XtNumber (PulldownData));

  /* Attach menu and drawing area to main */
  XmMainWindowSetAreas (main_window, menubar, NULL, NULL, NULL, board);
  XtManageChild (main_window);

  if (!xv_flag) {
      /* create a popup window for image display */
      n = 0;
      XtSetArg ( wargs[n], XmNwidth, columns); n++;
      XtSetArg ( wargs[n], XmNheight, rows); n++;
      xmstr = XmStringCreate ("image", XmSTRING_DEFAULT_CHARSET);
      XtSetArg (wargs[n], XmNdialogTitle, xmstr); n++;

      pop_window = XmCreateFormDialog (board, "img_window", wargs, n);
      free (xmstr);

      /* canvas for image display */

      canvas = XtCreateWidget ("image", xmDrawingAreaWidgetClass,
				  pop_window, NULL, 0);
      n = 0;
      XtSetArg ( wargs[n], XmNwidth, columns); n++;
      XtSetArg ( wargs[n], XmNheight, rows); n++;
      XtSetArg( wargs[n], XmNresizePolicy, XmRESIZE_ANY); n++;
      XtSetArg( wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
      XtSetArg( wargs[n], XmNbottomAttachment, XmATTACH_FORM); n++;
      XtSetArg( wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
      XtSetArg( wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
      XtSetValues (canvas, wargs, n);

      /* Add event handlers for expose and resize */

      XtAddEventHandler (canvas, ExposureMask, FALSE, expose, &image);
      XtAddEventHandler (canvas, StructureNotifyMask, FALSE, resize, NULL);
      XtAddEventHandler (canvas, ButtonPressMask, FALSE, mouse_click, &subimg);
      XtAddEventHandler (canvas, ButtonMotionMask, FALSE, mouse_drag, &subimg);
      XtAddEventHandler (canvas, ButtonReleaseMask, FALSE, mouse_release, &subimg);

      /*
       * Create the Graphics Context (GC)
       */

      theDisplay = XtDisplay (canvas);
      theScreen = DefaultScreen (theDisplay);
      theColormap = DefaultColormap (theDisplay, theScreen);
      theGC = XCreateGC (theDisplay, DefaultRootWindow (theDisplay), 0, NULL);

      /*
       * Allocate colors
       */

      ncolors = 240;
      result = XAllocColorCells (theDisplay, theColormap, FALSE, 
			     plane_masks, nplanes, pixels, ncolors);

      while (result == FALSE) {
         ncolors -= 4;
         if (result = XAllocColorCells (theDisplay, theColormap, FALSE, 
			        plane_masks, nplanes, pixels, ncolors))
             printf ("Succesfully allocated %d colors\n", ncolors);
      }	
  }

  XtRealizeWidget (top_level);

  if (img_ok) {
  
      /*
       * Display image
       */

      if (xv_flag) {
	 xv_display ();
      } else {
         if (type == PGM || type == RAW) {
             build_grey_table (gtable);
             create_grey_image (gtable, raw_image, rows, columns);
	     success = TRUE;
         } else if (type == PPM) {
	     success = create_color_image (raw_image_red, raw_image_green, 
			      raw_image_blue, rows, columns);
         }

         if (success) {
            XtManageChild (pop_window);
            XtManageChild (canvas);
            XPutImage (XtDisplay (canvas), XtWindow (canvas), theGC, image,
	               0, 0, 0, 0, columns, rows);
	    if (xv_disp_flag) xv_disp_flag = FALSE;
         } 
       }
       transf_done = FALSE;
       image_flag = TRUE; 
  }

   XtMainLoop ();
}

/******** End of the main program ****************/


/********************************************************************
 * Image Handling functions                                         *
 * Original code by M. Hilton, modified by P. Rajcani               *
 ********************************************************************/

/* build a color translation table for grayscale pictures */

void build_grey_table (gtable)
unsigned char *gtable;
{
   int 	i, j, k, n, first, last;
   int  step;
   XColor color;

   step = ceil((double) 256.0/ncolors);
   if (step <= 2) {
      first = 256 - ncolors;
      last  = 256 - first;
   
      for (i=0, k=0; i < 256;) {
	   color.red = color.blue = color.green = i*255;
	   color.flags = DoRed | DoGreen | DoBlue;
           color.pixel = pixels[k];
	   XStoreColor (theDisplay, theColormap, &color);
           if (i >= first && i < last) {
	      gtable[i++] = color.pixel; k++;
	   } else {
	      gtable[i++] = color.pixel;
	      gtable[i++] = color.pixel;
	      k++;
           }
       }
    } else {
      for (i=0, k=0; i < 256; i += step, k++) {
	   color.red = color.blue = color.green = i*255;
	   color.flags = DoRed | DoGreen | DoBlue;
           color.pixel = pixels[k];
	   XStoreColor (theDisplay, theColormap, &color);
	   for (j = 0; j < step; j++)
	      gtable[i+j] = color.pixel;
       }
   }
}

/* create image on the screen */

void create_grey_image (gtable, raw_img, rows, columns)
char *gtable;
imgtype **raw_img;
int	rows, columns;
{
  int i, j, k;

/*
  printf ("rows = %d, columns = %d, size = %d, xtop = %d, ytop = %d\n",
	   rows, columns, size, xtop, ytop);
*/

  if (image_flag) {
      free(image);
      free(conv_image);
  }

  if ((conv_image = (char *) malloc (rows*columns*
                			     sizeof (unsigned char))) == NULL) {
	printf ("Error in allocating converted image array\n");
	exit (1);
  }

  for (k=0, i = ytop; i < ytop + rows; i++) 
      for (j = xtop; j < xtop + columns; j++)
           conv_image[k++] =  gtable[imgtype_to_char(raw_img[i][j])];

  image = XCreateImage (theDisplay,
			 DefaultVisual (theDisplay, theScreen),
			 DefaultDepth (theDisplay, theScreen),
			 ZPixmap, 0, conv_image, columns, rows, 8, 0);

}

/* build a color translation table for color pictures */

void build_color_table (conv_image, raw_img_red, raw_img_green, 
			raw_img_blue, rows, columns)
char *conv_image;
imgtype **raw_img_red, **raw_img_green, **raw_img_blue;
int	rows, columns;
{
   int 	i, j, k, n, last, found;
   XColor color;
   unsigned short red, green, blue;
   color_pixels     ctable[256];

   found = FALSE; last = 0;
   for (n=0, i=0; i < rows; i++) {
	for (j = 0; j < columns; j++) {
	     red = imgtype_to_char(raw_img_red[i][j])*255;
	     green = imgtype_to_char(raw_img_green[i][j])*255;
             blue = imgtype_to_char(raw_img_blue[i][j])*255;
	     found = FALSE;
             for (k=0; k<last; k++)
		 if (ctable[k].red == red && 
		     ctable[k].blue == blue && 
		     ctable[k].green == green) { 
			found = TRUE; 
			conv_image[n++] = pixels[k];
			break; 
                  }

	     if (!found) {
		 color.red = ctable[k].red = red; 
		 color.blue = ctable[k].blue = blue; 
		 color.green = ctable[k].green = green; 
		 color.pixel = pixels[k];
              	 color.flags = DoRed | DoGreen | DoBlue;
		 XStoreColor (theDisplay, theColormap, &color);
	    	 conv_image[n++] = color.pixel;
		 last = k+1;
	     }
             
        }
   }
}


/* create color image on the screen */

int create_color_image (raw_img_red, raw_img_green, raw_img_blue, 
			 rows, columns)
imgtype **raw_img_red, **raw_img_green, **raw_img_blue;
int	rows, columns;
{
  int 		i, j, k, type;
  int		quant_depth;
  unsigned char *input_vector;
  FILE		*tmp;
  char		cmd[50];
  void		rescale_colors ();
  
  if (image_flag) {
      free(image);
      free(conv_image);
  }

  if (color_flag) {
     for (i = 0; i < quant_size; i++) {
	 free (quant_image_red[i]);
	 free (quant_image_green[i]);
	 free (quant_image_blue[i]);
     }
     free (quant_image_red); 
     free (quant_image_green); 
     free (quant_image_blue); 
  }

/*
  printf ("rows = %d, columns = %d, size = %d, xtop = %d, ytop = %d\n",
	   rows, columns, size, xtop, ytop);
*/

  if ((tmp = fopen ("tmp.ppm", "w")) == NULL) {
	printf ("Cannot display the color image!\n"); 
	printf ("Please check the file tmp.ppm.\n");
	exit(-1);
  }

  fprintf (tmp, "P6\n");
  fprintf (tmp, "%d %d\n", columns, rows);
  fprintf (tmp, "255\n");

  for (i = ytop; i < ytop + rows; i++) {
      for (j = xtop; j < xtop + columns; j++) {
           fputc (imgtype_to_char(raw_img_red[i][j]), tmp);
           fputc (imgtype_to_char(raw_img_green[i][j]), tmp);
           fputc (imgtype_to_char(raw_img_blue[i][j]), tmp);
      }
  }
  fclose (tmp);

  sprintf (cmd, "ppmquant %d tmp.ppm > tmp_quant", ncolors);
  system (cmd);

  if ((input_vector = read_pnm_file ("tmp_quant", &type, 
			&quant_rows, &quant_columns)) == NULL) {

      printf ("Warning: Could not open the tmp_quant file.\n");
      system ("rm tmp.ppm");
      xv_display ();
      return (FALSE);
  }

  read_color_image (input_vector, &quant_image_red, &quant_image_green,
		    &quant_image_blue, quant_rows, quant_columns, 
                    &quant_size, &quant_depth);

  rescale_colors ( quant_image_red, quant_image_green,
		   quant_image_blue, quant_rows, quant_columns);

  system ("rm tmp.ppm tmp_quant");
  if ((conv_image = (char *) malloc (quant_rows*quant_columns*
                     sizeof (unsigned char))) == NULL) {
	printf ("Error in allocating converted image array\n");
	exit (1);
  }

  printf ("Building the color translation table..."); fflush (stdout);
  build_color_table (conv_image, quant_image_red, quant_image_green,
		     quant_image_blue, quant_rows, quant_columns);
  printf ("Done\n"); fflush (stdout);


  image = XCreateImage (theDisplay,
			 DefaultVisual (theDisplay, theScreen),
			 DefaultDepth (theDisplay, theScreen),
			 ZPixmap, 0, conv_image, columns, 
			 rows, 8, 0);
  return (TRUE);
}

/**********************************************************
 * Function: rescale_colors
 *
 *********************************************************/

 void rescale_colors ( quant_image_red, quant_image_green,
		       quant_image_blue, qrows, qcolumns)
 imgtype **quant_image_red, **quant_image_green, 
	       **quant_image_blue;
 int		qrows, qcolumns;
 {
   int	i, j;
   int  max;
   float x, factor;

   max = 0;
   for (i=0; i<qrows; i++) {
	for (j = 0; j < qcolumns; j++) {
	     if (quant_image_red[i][j] > max) 
		max = quant_image_red[i][j];
	     if (quant_image_green[i][j] > max) 
		max = quant_image_green[i][j];
	     if (quant_image_blue[i][j] > max) 
		max = quant_image_blue[i][j];
        }
    }

    printf ("Max color value = %d.\n", max);

    if (max < 255) {
       /* rescale values */
       factor = (float) 255/ (float) max;

	if (factor > 1.5) {
            printf ("Rescaling colors ...."); fflush(stdout);
	    for (i = 0; i < qrows; i++) {
		for (j = 0; j<qcolumns; j++) {
		     quant_image_red[i][j] *= factor;
		     quant_image_green[i][j] *= factor;
		     quant_image_blue[i][j] *= factor;
		}
	     }
             printf ("Done\n"); fflush(stdout);
	}
    }
 }

/************************************************************
 * FUNCTION: resize_drawing_area                            *
 * 	     Set new dimensions for the image -             *
 *	     canvas drawing area                            *
 ************************************************************/

void resize_drawing_area (rows, columns)
int	rows, columns;
{
  int	n;
  Arg   wargs[2];

  n = 0;
  XtSetArg ( wargs[n], XmNwidth, columns); n++;
  XtSetArg ( wargs[n], XmNheight, rows); n++;
  XtSetValues (canvas, wargs, n);
  XtSetValues (pop_window, wargs, n);
  XtManageChild (pop_window);
  XtManageChild (canvas);
}

/* cut the black rim around the image if needed */

void get_subimage ( image, s) 
imgtype **image;
int	  s;
{
    int	i, j;
    int	  xmin, xmax, ymin, ymax;
    double sz;
    void  save_grey ();

    xmax = 0; ymin = 9999; ymax = 0; xmin = 9999;
    for (i=0; i<s; i++) {
         for (j=0; j<s; j++) { 
		if (image[i][j] > 0) {
		    if (j > xmax) xmax = j;
		    if (j < xmin) xmin = j;
		    if (i > ymax) ymax = i;
		    if (i < ymin) ymin = i;
		}
	    }
     }
     if (ymin == 9999 || xmin == 9999) {
        rows = columns = s;
        xtop = ytop = 0;
     } else {
        columns = xmax-xmin+1; rows = ymax-ymin+1;
        xtop = xmin; ytop = ymin;


	if (xtop > 0 || ytop > 0) {
	   /* shift the image */
	   for (i=0; i < rows; i++) 
		for (j=0; j< columns; j++) 
		     image[i][j] = image[i+ytop][j+xtop];
	   xtop = ytop = 0;
        }


   	sz = (double) (rows > columns) ? rows : columns;
   	depth = ceil (log (sz) / log ((double) 2.0));
   	for (size = 1, i = 0; i < depth; size *= 2, i++); 

       /*
       printf ("Cut image: rows = %d, columns = %d, size = %d, depth = %d.\n", rows, columns, size, depth);
	*/
     }
     /*save_grey (image, size);*/
}


/* cut the black rim around the image if needed */

void get_subimage_color ( image_red, image_green, image_blue, s)
imgtype **image_red, **image_green, **image_blue;
int	  s;
{
    int	i, j;
    int	  xmin, xmax, ymin, ymax;
    double sz;

    xmax = 0; ymin = 9999; ymax = 0; xmin = 9999;
    for (i=0; i<s; i++) {
         for (j=0; j<s; j++) { 
		if (image_red[i][j] > 0 || image_green[i][j] > 0 ||
		    image_blue[i][j] > 0) {
		    if (j > xmax) xmax = j;
		    if (j < xmin) xmin = j;
		    if (i > ymax) ymax = i;
		    if (i < ymin) ymin = i;
		}
	    }
     }
     if (ymin == 9999 || xmin == 9999) {
        rows = columns = s;
        xtop = ytop = 0;
     } else {
        columns = xmax-xmin+1; rows = ymax-ymin+1;
        xtop = xmin; ytop = ymin;

	if (xtop > 0 || ytop > 0) {
	   /* shift the image */
	   for (i=0; i < rows; i++) {
		for (j=0; j< columns; j++) {
		     image_red[i][j] = image_red[i+ytop][j+xtop];
		     image_blue[i][j] = image_blue[i+ytop][j+xtop];
		     image_green[i][j] = image_green[i+ytop][j+xtop];
                }
           }
	   xtop = ytop = 0;
	}

                      
   	sz = (double) (rows > columns) ? rows : columns;
   	depth = ceil (log (sz) / log ((double) 2.0));
   	for (size = 1, i = 0; i < depth; size *= 2, i++); 

    /*  printf ("xmin = %d, xmax = %d, ymin = %d, ymax = %d\n", 
                 xmin, xmax, ymin, ymax);*/
/*
       printf ("Cut image: rows = %d, columns = %d, size = %d, depth = %d.\n", rows, columns, size, depth);
*/
     }
}

/*** Display an image using xv ***/
void xv_display ()
{
  	int  type;

	printf ("Displaying the image using xv ...\n"); fflush (stdout);
        type = (color_flag) ? PPM : PGM;

        if (save_img ("tmp", type)) {
           /* kill the previous xv process  */
	   if (xv_disp_flag) 
		system ("xvkill");
	   system ("xv tmp &");
           xv_disp_flag = TRUE;
	 } else {
	    error ("Cannot display the image. Please check the 'tmp' file");
        }
}

void display_image (cut_flag)
int cut_flag;
{
   int success;

   /* display the output image */

    if (cut_flag) {
        if (color_flag) {
	     if (crop_flag) {
                get_subimage_color (crop_image_red, crop_image_green,
			            crop_image_blue, size);
             } else {
            	get_subimage_color (raw_image_red, raw_image_green,
			            raw_image_blue, size);
	     }
	  } else {
		if (crop_flag) {
                   get_subimage (crop_image, size);
	        } else {
            	    get_subimage (raw_image, size);
	        }
          }
   }

   if (xv_flag) {
       xv_display ();
       return;
   }
        

   if (color_flag) {
	if (crop_flag) {
               success = create_color_image (crop_image_red, crop_image_green,
			        crop_image_blue , rows, columns);
	} else {
  	    success = create_color_image (raw_image_red, raw_image_green,
			        raw_image_blue , rows, columns);
	}
   } else {

        success = TRUE;
	if (crop_flag) {
  	    create_grey_image (gtable, crop_image, rows, columns);
	} else {
          if (cut_flag)
            get_subimage (raw_image, size);
  	    create_grey_image (gtable, raw_image, rows, columns);
	}
    }

    if (success) {
        resize_drawing_area (rows, columns);
        XPutImage (XtDisplay (canvas), XtWindow (canvas), theGC, image,
	           0, 0, 0, 0, columns, rows);
	if (xv_disp_flag) xv_disp_flag = FALSE;
    }
}

/****************************************************************
 * Event Handlers                                               *
 ****************************************************************/

void expose (w, img, event)
Widget	w;
XImage  **img;
XEvent  *event;
{
  if (event->xexpose.count == 0) {
      if (image_flag) {
          XPutImage (XtDisplay (w), XtWindow (w), theGC, *img,
	        	0, 0, 0, 0, columns, rows);
      }
   }
}

void redisplay (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
    int	result;

    if (image_flag) {
	if (xv_flag) { xv_display (); return; }

        if (color_flag) {
	  if (crop_flag) 
  	    result = create_color_image (crop_image_red, crop_image_green,
			        crop_image_blue , rows, columns);
	  else 
  	    result = create_color_image (raw_image_red, raw_image_green,
			        raw_image_blue , rows, columns);
        } else {
           result = TRUE;
	   if (crop_flag) 
  	      create_grey_image (gtable, crop_image, rows, columns);
	  else 
  	      create_grey_image (gtable, raw_image, rows, columns);
        }

       if (result) {
          XPutImage (XtDisplay (canvas), XtWindow (canvas), theGC, image,
	             0, 0, 0, 0, columns, rows);
	  xv_disp_flag = FALSE;
       } 
     } else
	error ("No image loaded\n");
}


/********************************************************
 * resize event handler - can be switched on using the
 * '/r' option in the command line
 *******************************************************/

void resize (w, client_data, event)
Widget	w;
caddr_t *client_data;
XEvent  *event;
{

     Dimension 	height, width;
     Cardinal	argcount;
     Arg	args[10];
     float	scaleV, scaleH;
     void	scale_img ();

     if (image_flag && resize_activated) {
	    argcount = 0;
       	    XtSetArg(args[argcount], XmNheight, &height);
            argcount++;
            XtSetArg(args[argcount], XmNwidth, &width);
            argcount++;
            XtGetValues(canvas, args, argcount);

	    /* only redraw if actually changed */
	    if (height != rows || width != columns)
	    {

	  	scaleV = (float) height / (float) rows;
	  	scaleH = (float) width / (float) columns;

          	printf ("height: %d, width: %d, scaleH = %f, scaleV = %f\n", 
			height, width, scaleH, scaleV);

	  	scale_img (scaleV, scaleH);

   		if (color_flag) {
	    	   if (crop_flag) 
  	              	create_color_image (crop_image_red, crop_image_green,
			            	crop_image_blue , rows, columns);
	    	   else 
  	        	create_color_image (raw_image_red, raw_image_green,
			            raw_image_blue , rows, columns);
        	} else {
	    	   if (crop_flag)
  	       		create_grey_image (gtable, crop_image, 
					   rows, columns);
	    	   else 
  	       	        create_grey_image (gtable, raw_image, rows, columns);
		}
                XPutImage (XtDisplay (canvas), XtWindow (canvas), 
			theGC, image, 0, 0, 0, 0, columns, rows);
	      }
     }
}


/* first mouse click - pick the coordinates */
void mouse_click (w, subimg, event)
Widget	w;
rectangle  *subimg;
XEvent  *event;
{
  int  x, y;
  imgtype greyness, red, green, blue;
  char label[35];
  
  if (event->xbutton.button == Button1) {
      x = event->xbutton.x;
      y = event->xbutton.y;
      if (x + xtop < size && y + ytop < size) {
         if (color_flag) {
	    red = quant_image_red[y][x];
	    green = quant_image_green[y][x];
	    blue = quant_image_blue[y][x];
            sprintf (label, "x = %d, y = %d, red: %5.1f, green: %5.1f, blue: %5.1f", x, y, red, green, blue);
      } else {
         if (crop_flag)
	    greyness = crop_image[y+ytop][x+xtop];
         else
	    greyness = raw_image[y+ytop][x+xtop];

         sprintf (label, "x = %d, y = %d, greyness = %5.1f", x, y, greyness);
      }
      set_label (value_label, label);
      show_values = TRUE;
     }

  } else {
      
     if (!rect_selected) {
         subimg->x = event->xbutton.x;
         subimg->y = event->xbutton.y;
     }
  }
}


void mouse_drag (w, subimg, event)
Widget	  w;
rectangle *subimg;
XEvent    *event;
{
   int	width, height, x, y;
   imgtype  red, green, blue, greyness;
   char label[35];

   if (show_values) {
      x = event->xmotion.x;
      y = event->xmotion.y;
      if (x >= 0 && x < columns && y >= 0 && y < rows) {
         if (color_flag) {
	    red = quant_image_red[y][x];
	    green = quant_image_green[y][x];
	    blue = quant_image_blue[y][x];
            sprintf (label, "x = %d, y = %d, red: %5.1f, green: %5.1f, blue: %5.1f", x, y, red, green, blue);
         } else {
            if (crop_flag)
	       greyness = crop_image[y+ytop][x+xtop];
            else
	       greyness = raw_image[y+ytop][x+xtop];

            sprintf (label, "x = %d, y = %d, greyness = %5.1f", x, y, greyness);
         }
         set_label (value_label, label);
      }
   } else {
      width = event->xmotion.x - subimg->x;
      height = event->xmotion.y - subimg->y;
   
      sprintf (label, "Cropped image: %d rows x %d columns", height, width);
      set_label (crop_label, label);
   }
}


/* button release - get the size of the cropping rectangle */

void mouse_release (w, subimg, event)
Widget	w;
rectangle  *subimg;
XEvent  *event;
{
  int	x, y;
  char label[35];
  XColor color;
  long   black_pixel;

  if (event->xbutton.button == Button1) {
      set_label (value_label, "");
      show_values = FALSE;
  } else {
      if (!rect_selected) {
         x = event->xbutton.x;
         y = event->xbutton.y;
         if (x > subimg->x && y > subimg->y) {
            subimg->width = x - subimg->x;
            subimg->height = y - subimg->y;
            color.red = color.green = color.blue = 0;
            if (XAllocColor (theDisplay, theColormap, &color))
                black_pixel = color.pixel;
            else
	        black_pixel = 0;
            XSetForeground (XtDisplay (canvas), theGC, black_pixel);
            XDrawRectangle (XtDisplay (canvas), XtWindow (canvas), theGC,
		            subimg->x, subimg->y, 
                            subimg->width, subimg->height);
            rect_selected = TRUE;
            sprintf (label, "Cropped image: %d rows x %d columns", 
                      subimg->height, subimg->width);
            set_label (crop_label, label);
         } else {
	    rect_selected = FALSE;
            set_label (crop_label, "");
         }
    }
  }
}

/*-------------------------------------------------------------*/


/***********************************************************
 * I/O CALLBACKS
 **********************************************************/

void load_image (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
   void	  image_ok();

      display_file_sb ("Load image", IMAGE_MASK, image_ok);
}


void load3 (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
   void	  red_image_ok();

      display_file_sb ("Load red image", IMAGE_MASK, red_image_ok);
}


void red_image_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmFileSelectionBoxCallbackStruct *call_data;
{
  char *imagefile;
  int  type, i, j;
  unsigned char *input_vector;
  void green_image_ok ();

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &imagefile));
  XtUnmanageChild (dialog_box);

  if (imagefile != NULL) {

     input_vector = read_pnm_file (imagefile, &type, 
				   &rrows, &rcolumns);
     if (input_vector == NULL) {
	error ("Cannot read the red component!");
	return;
     }

	free_crop_image ();

        if (raw_image_red != NULL) {
           for (i=0; i < size; i++) 
               free ((raw_image_red)[i]); 
           free (raw_image_red); 
           raw_image_red = NULL;
        }

        if (orig_image_red != NULL) {
           for (i=0; i < orig_size; i++) 
               free ((orig_image_red)[i]); 
           free (orig_image_red); 
           orig_image_red = NULL; 
        }


     read_grey_image (imagefile, input_vector, &orig_image_red, 
                          &rrows, &rcolumns, 
                          &rsize, &rdepth, type);


      free (input_vector);
      display_file_sb ("Load green image", IMAGE_MASK, green_image_ok);
 }
}


void green_image_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmFileSelectionBoxCallbackStruct *call_data;
{
  char *imagefile;
  int  type, i, j;
  unsigned char *input_vector;
  void blue_image_ok ();

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &imagefile));
  XtUnmanageChild (dialog_box);

  if (imagefile != NULL) {
     input_vector = read_pnm_file (imagefile, &type, 
				   &grows, &gcolumns);
     if (input_vector == NULL) {
	error ("Cannot read the green component!");
	return;
     }


        if (raw_image_green != NULL) {
           for (i=0; i < size; i++) 
               free ((raw_image_green)[i]); 
           free (raw_image_green); 
           raw_image_green = NULL;
        }

        if (orig_image_green != NULL) {
           for (i=0; i < orig_size; i++) 
               free ((orig_image_green)[i]); 
           free (orig_image_green); 
           orig_image_green = NULL; 
        }


     read_grey_image (imagefile, input_vector, &orig_image_green, 
                          &grows, &gcolumns, 
                          &gsize, &gdepth, type);

      free (input_vector);
      display_file_sb ("Load blue image", IMAGE_MASK, blue_image_ok);
 }
}


void blue_image_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmFileSelectionBoxCallbackStruct *call_data;
{
  char *imagefile;
  int  type, i, j, max ();
  unsigned char *input_vector;
  void zoom_image ();

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &imagefile));
  XtUnmanageChild (dialog_box);

  if (imagefile != NULL) {

     input_vector = read_pnm_file (imagefile, &type, 
				   &brows, &bcolumns);
     if (input_vector == NULL) {
	error ("Cannot read the blue component!");
	return;
     }


        if (raw_image_blue != NULL) {
           free_image (raw_image_blue, size); 
           raw_image_blue = NULL;
        }

        if (orig_image_blue != NULL) {
           free_image (orig_image_blue, orig_size); 
           orig_image_blue = NULL; 
        }


     color_flag = TRUE;
     orig_flag = COLOR;
     read_grey_image (imagefile, input_vector, &orig_image_blue, 
                          &brows, &bcolumns, 
                          &bsize, &bdepth, type);
      free (input_vector);


     /* adjust the image sizes */

     rows = orig_rows = max (rrows, grows, brows);
     columns = orig_columns = max (rcolumns, gcolumns, bcolumns);
     size = orig_size = max (rsize, gsize, bsize);
     depth = orig_depth = max (rdepth, gdepth, bdepth);

     if (size > rsize) zoom_image (&orig_image_red, rsize, size);
     if (size > gsize) zoom_image (&orig_image_green, gsize, size);
     if (size > bsize) zoom_image (&orig_image_blue, bsize, size);

     raw_image_red = alloc_image (orig_size, orig_size);
     copy_image (orig_image_red, raw_image_red, orig_size, orig_size);

     raw_image_green = alloc_image (orig_size, orig_size);
     copy_image (orig_image_green, raw_image_green, orig_size, orig_size);

     raw_image_blue = alloc_image (orig_size, orig_size);
     copy_image (orig_image_blue, raw_image_blue, orig_size, orig_size);

      xtop = ytop = 0;

      if (xv_flag) {
	 xv_display ();
      } else {
         if (image_flag) XtUnmanageChild (pop_window);
         if (create_color_image (raw_image_red, raw_image_green, 
				raw_image_blue, rows, columns)) {
             resize_drawing_area (rows, columns);
             XPutImage (XtDisplay (canvas), XtWindow (canvas), 
			theGC, image, 0, 0, 0, 0, columns, rows);
             xv_disp_flag = FALSE;
         }
      }
      image_flag = TRUE;
 }
}


void inverse_wft (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  if (wft_selected) {
     if (wft_unrolled) free_autom_values (&wft, usize);
     invert_wft (&wft);
     wft_selected = TRUE;
     wft_started = FALSE;
     wft_unrolled = FALSE;
     usize = 0; wft_depth = 0;
     printf ("The automaton inverted successfully\n"); fflush (stdout);
     set_label(wft_label, "Loaded WFT:inverted WFT"); 
     switch (wft.method) {
	case SIMPLE:
             set_label (mlabel,"method: simple"); break;
	case GENERAL:
             set_label(mlabel, "method: general"); break;
     }
   } else 
	error ("Please load the WFT first");
}


/******************************************************************
 * Function: concatenation                                        *
 * 	     Concatenate two images , WFA or WFTs                 *
 ******************************************************************/

void concatenation (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
   void	  concat_images();
   void   load_wfa2(), load_wft2 ();

   switch (mode) {
	case IMAGE:
             if (image_flag)
                display_file_sb ("Load second image", IMAGE_MASK, 
				 concat_images);
             else
                error ("Please read the first image before concatenation");
	     break;

	case WFA:
	     if (color_aut) {
		error ("This operation cannot be performed in color mode");
		break;
	     }
  	     if (wfa_selected) {
  	        action_type = CONCAT;
                display_file_sb ("Load the second automaton", WFA_MASK, 
                                 load_wfa2);
             } else
                 error ("Please load the first wfa before concatenation");
	     break;

	case WFT:
	     if (red_wft_selected || green_wft_selected || 
		blue_wft_selected) {
		   error ("This operation cannot be performed in color mode");
		   break;
	     }
  	     if (wft_selected) {
  	         action_type = CONCAT;
      	         display_file_sb ("Load the second WFT", WFT_MASK, 
				  load_wft2);
  	      } else
     		   error ("Please load the first wft before concatenation");
	       break;
      }
}

void cproduct (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void 	load_first();

  action_type = CPROD;
  display_file_sb ("Load the first automaton", WFA_MASK, load_first);
}

void sum (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void 	load_wfa2(), load_wft2(), add_images ();

  switch (mode) {
	case IMAGE:
             if (image_flag)
                display_file_sb ("Load second image", IMAGE_MASK, 
				 add_images);
   	     else
                error ("Please read the first image before addition");
	     break;

	case WFA:
	     if (color_aut) {
		error ("This operation cannot be performed in color mode");
		break;
	     }
  	     if (wfa_selected) {
  	        action_type = SUM;
                display_file_sb ("Load the second automaton", WFA_MASK, 
                                 load_wfa2);
             } else
                error ("Please load the first wfa before addition");
	     break;

	case WFT:
	     if (red_wft_selected || green_wft_selected || 
		blue_wft_selected) {
		   error ("This operation cannot be performed in color mode");
		   break;
	     }
  	     if (wft_selected) {
  	         action_type = SUM;
      	         display_file_sb ("Load the second WFT", WFT_MASK, 
				  load_wft2);
  	      } else
                 error ("Please load the first WFT before addition");
	      break;
	}
}


void load_wft (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
   void	  	wft_ok();

   wft_load_type = SINGLE;
   display_file_sb ("Load WFT", WFT_MASK, wft_ok);
}

void load_red_wft (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
   void	  	wft_ok();

   if (color_flag) {
      wft_load_type = RED;
      display_file_sb ("Load red WFT", WFT_MASK, wft_ok);
   } else
      error ("This option can be used only with color images");
}


void load_green_wft (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
   void	  	wft_ok();

   if (color_flag) {
      wft_load_type = GREEN;
      display_file_sb ("Load green WFT", WFT_MASK, wft_ok);
   } else
      error ("This option can be used only with color images");
}

void load_blue_wft (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
   void	  	wft_ok();

   if (color_flag) {
      wft_load_type = BLUE;
      display_file_sb ("Load blue WFT", WFT_MASK, wft_ok);
   } else
      error ("This option can be used only with color images");
}

void load_single_wfa (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
   void	  	wfa_ok();

   display_file_sb ("Load automaton", WFA_MASK, wfa_ok);
}


void load_3wfa (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
   void	  	red_wfa_ok();

   display_file_sb ("Load red WFA", WFA_MASK, red_wfa_ok);
}

void load_rgb_wfa (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
   void	  	rgb_wfa_ok();

   display_dialog ("Please enter the filename without extension:", 
		    rgb_wfa_ok);
}

void load_3wft (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
   void	  	red_wft_ok();

   if (color_flag) {
      display_file_sb ("Load red WFT", WFT_MASK, red_wft_ok);
   } else
      error ("This option can be used only with color images");
}


void save_wft (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void		write_wft ();

  display_dialog ("Please select the name of the output file", write_wft);
}

void zoom_square (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void		zoom_ok ();
  void		wfa_zoom_ok ();

  switch (mode) {
	case IMAGE:
             if (image_flag) 
                 display_dialog ("Please enter the address of the subsquare", 
                                 zoom_ok);
             else
     		error ("Please read the image first");
	     break;

	case WFA:
             if (wfa_selected || red_wfa_selected || 
                 green_wfa_selected || blue_wfa_selected) {
		 action_type = ZOOM;
                 display_dialog ("Please enter the address of the subsquare", 
                   wfa_zoom_ok);
             } else
                  error ("Please read the WFA first");
	     break;

	case WFT:
		error ("This operation cannot be performed on WFT");
		break;
        }
}

void save_wfa (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void		write_wfa ();

  display_dialog ("Please select the name of the output file", write_wfa);
}

void save_image (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void	save_ok ();

  display_dialog ("Please select the name of the output file", save_ok);
}


/*********** Returns to the original image ***************/

void original_image (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  int	i, j, success;

  if (image_flag) {

     if (orig_flag == COLOR) {

	free_crop_image ();

       if (raw_image_red != NULL) {
          free_image (raw_image_red, size); 
          raw_image_red = NULL;
       }

       if (raw_image_green != NULL) {
          free_image (raw_image_green, size); 
          raw_image_green = NULL;
       }

       if (raw_image_blue != NULL) {
          free_image(raw_image_blue, size); 
          raw_image_blue = NULL;
       }

	raw_image_red = alloc_image (orig_size, orig_size);
	copy_image (orig_image_red, raw_image_red, orig_size, orig_size);
	raw_image_blue = alloc_image (orig_size, orig_size);
	copy_image (orig_image_blue, raw_image_blue, orig_size, orig_size);
	raw_image_green = alloc_image (orig_size, orig_size);
	copy_image (orig_image_green, raw_image_green, orig_size, orig_size);
         color_flag = TRUE;

     } else if (orig_flag == GREY) {

	free_crop_image ();

         if (raw_image != NULL) {
            free_image (raw_image, size); 
            raw_image = NULL;
         }
     
	raw_image = alloc_image (orig_size, orig_size);
	copy_image (orig_image, raw_image, orig_size, orig_size);
        color_flag = FALSE;
      }
      rows = orig_rows; columns = orig_columns; 
      size = orig_size; depth = orig_depth;
      xtop = ytop = 0;
      transf_done = FALSE;

      /* display the output image */
      if (xv_flag) {
	 xv_display ();
      } else {
          XtUnmanageChild (pop_window);
          if (color_flag)
	     success = create_color_image (raw_image_red, raw_image_green,
			     raw_image_blue, rows, columns);
          else {
	     success = TRUE;
             create_grey_image (gtable, raw_image, rows, columns);
          }

          if (success) {
             resize_drawing_area (rows, columns);
             XPutImage (XtDisplay (canvas), XtWindow (canvas), theGC, image,
	       	    0, 0, 0, 0, columns, rows);
             xv_disp_flag = FALSE;
          }
      }
   } else 
      error ("Please read the image first");
}

reload_color_wfa ()
{
  char str[50];

  free_automaton (&wfa_red);
  copy_automaton (&orig_wfa_red, &wfa_red);

  free_automaton (&wfa_blue);
  copy_automaton (&orig_wfa_blue, &wfa_blue);

   free_automaton (&wfa_green);
   copy_automaton (&orig_wfa_green, &wfa_green);
   if (wfa_load_type == MULTIPLE) {
      sprintf (str, "Red component: %s", orig_red_wfa_name);
      set_label(red_wfa_label, str);
      sprintf (str, "Loaded WFA: %s", orig_green_wfa_name);
      set_label(green_wfa_label, str);
      sprintf (str, "Blue component: %s", orig_blue_wfa_name);
      set_label(blue_wfa_label, str);
      set_label(wfa_label, "Loaded WFA: ");
    } else {
      sprintf (str, "Loaded WFA: %s", orig_color_wfa_name);
      set_label(wfa_label, str);
    }
}

void reload_wfa ()
{
  char str[50];

   free_automaton (&wfa);
   copy_automaton (&orig_wfa, &wfa);
   sprintf (str, "Loaded WFA: %s", orig_wfa_name);
   set_label(wfa_label, str);
   wfa_depth = 0;
}

void undo_op (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
   void undo_transf (), decode_aut ();

   switch (mode) {
	case IMAGE: 
  	        if (wft_unrolled || red_wft_unrolled || 
		    green_wft_unrolled || blue_wft_unrolled ||  
		    transf_done) 
                    undo_transf (); 
                else 
	            error ("Please do the operation first");
                break;
	case WFA:   
	        if (action_type == ZOOM) {
		   decode_aut (DDEPTH);
		} else {
		   if (color_aut && red_wfa_selected && 
                       green_wfa_selected && blue_wfa_selected) {
			reload_color_wfa ();
                   } else  if (wfa_selected) {
		     reload_wfa ();
		   } else 
	             error ("Please load the WFA first");
                 }
		 break;

	case WFT:   
		error ("Cannot undo a WFT operation");
    }
}


void undo (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void undo_transf ();

   if (transf_done) { 
      undo_transf ();
   } else 
      error ("Please apply wft or a transformation first");
}

void undo_transf ()
{
  int	i, j, cut_flag;

     if (color_flag) {
        if (crop_flag) {
           free_image(crop_image_red, size); crop_image_red = NULL;
           free_image(crop_image_green, size); crop_image_green = NULL;
           free_image(crop_image_blue, size); crop_image_blue = NULL;

	   crop_image_red = alloc_image (psize, psize);
	   copy_image (prev_image_red, crop_image_red, psize, psize);

	   crop_image_blue = alloc_image (psize, psize);
	   copy_image (prev_image_blue, crop_image_blue, psize, psize);

	   crop_image_green = alloc_image (psize, psize);
	   copy_image (prev_image_green, crop_image_green, psize, psize);

       } else {

       if (raw_image_red != NULL) {
          free_image (raw_image_red, size); raw_image_red = NULL;
       }

       if (raw_image_green != NULL) {
          free_image (raw_image_green, size); raw_image_green = NULL;
       }

       if (raw_image_blue != NULL) {
          free_image(raw_image_blue, size); raw_image_blue = NULL;
       }

     
	   raw_image_red = alloc_image (psize, psize);
	   copy_image (prev_image_red, raw_image_red, psize, psize);

	   raw_image_blue = alloc_image (psize, psize);
	   copy_image (prev_image_blue, raw_image_blue, psize, psize);

	   raw_image_green = alloc_image (psize, psize);
	   copy_image (prev_image_green, raw_image_green, psize, psize);

       }

     } else  {
        if (crop_flag) {
            free_image (crop_image,size); crop_image = NULL;

	   crop_image = alloc_image (psize, psize);
	   copy_image (prev_image, crop_image, psize, psize);

       } else {

         if (raw_image != NULL) {
            free_image (raw_image, size); raw_image = NULL;
         }
     
	   raw_image = alloc_image (psize, psize);
	   copy_image (prev_image, raw_image, psize, psize);
      }
     }
      rows = prows; columns = pcolumns; 
      size = psize; depth = pdepth;
      if (!xv_flag) XtUnmanageChild (pop_window);

      if (transf_done) {
	 if (rows < size || columns < size) cut_flag = TRUE;
	 else cut_flag = FALSE;
      } else cut_flag = FALSE;

      /* display the output image */
      display_image (cut_flag);
      transf_done = FALSE;
}


void revert (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  int	i,j;
  if (image_flag) {
     if (!color_flag) {
        if (!crop_flag) {
	    for (i = ytop; i < ytop+rows; i++)
	        for (j = xtop; j < xtop+columns; j++)
		    raw_image[i][j] = 255 - raw_image[i][j];
	} else {
	    for (i = ytop; i < ytop+rows; i++)
	        for (j = xtop; j < xtop+columns; j++)
		    crop_image[i][j] = 255 - crop_image[i][j];
        }
     } else {
        if (!crop_flag) {
	    for (i = ytop; i < ytop+rows; i++)
	        for (j = xtop; j < xtop+columns; j++)
		    raw_image_red[i][j] = 255 - raw_image_red[i][j];

	    for (i = ytop; i < ytop+rows; i++)
	        for (j = xtop; j < xtop+columns; j++)
		    raw_image_green[i][j] = 255 - raw_image_green[i][j];

	    for (i = ytop; i < ytop+rows; i++)
	        for (j = xtop; j < xtop+columns; j++)
		    raw_image_blue[i][j] = 255 - raw_image_blue[i][j];
	} else {
	    for (i = ytop; i < ytop+rows; i++)
	        for (j = xtop; j < xtop+columns; j++)
		    crop_image_red[i][j] = 255 - crop_image_red[i][j];

	    for (i = ytop; i < ytop+rows; i++)
	        for (j = xtop; j < xtop+columns; j++)
		    crop_image_green[i][j] = 255 - crop_image_green[i][j];

	    for (i = ytop; i < ytop+rows; i++)
	        for (j = xtop; j < xtop+columns; j++)
		    crop_image_blue[i][j] = 255 - crop_image_blue[i][j];

       }		
     }
     display_image (FALSE);

   } else 
	error ("Please load the image first");
}


/*
void print_image (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{

  if (! color_flag) {
     save_img ("tmp", RAW);
     if (size == 256) 
         system ("/fac/culik/bin/primg tmp");
     else if (size == 512)
         system ("/fac/culik/bin/primg512 tmp");
     else
         error ("This image cannot be printed - save it and print it using xv");
  } else 
	  error ("Cannot print a color image");
}
*/


/* User selected ok to read the image */

void image_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmFileSelectionBoxCallbackStruct *call_data;
{
  char *imagefile;
  int  type, i, j, disp_ok;
  unsigned char *input_vector;

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &imagefile));
  XtUnmanageChild (dialog_box);

  if (imagefile != NULL) {
     input_vector = read_pnm_file (imagefile, &type, 
				   &orig_rows, &orig_columns);
     if (input_vector == NULL) {
	error ("Error in reading of the input image!");
	return;
     }

    if (color_flag) {

     	free_crop_image ();

        if (raw_image_red != NULL) {
           free_image (raw_image_red, size); raw_image_red = NULL;
        }

        if (raw_image_blue != NULL) {
           free_image (raw_image_blue, size); raw_image_blue = NULL;
        }

        if (raw_image_green != NULL) {
           free_image (raw_image_green, size); raw_image_green = NULL;
        }


        if (orig_image_red != NULL) {
           free_image (orig_image_red, orig_size); 
           orig_image_red = NULL; 
        }

        if (orig_image_blue != NULL) {
           free_image (orig_image_blue, orig_size); 
           orig_image_blue = NULL; 
        }

        if (orig_image_green != NULL) {
           free_image (orig_image_green, orig_size); 
	   orig_image_green = NULL; 
        }

     } else {


	free_crop_image ();

        if (raw_image != NULL) {
           free_image (raw_image, size); raw_image = NULL; 
        }

        if (orig_image != NULL) {
           free_image (orig_image, orig_size); orig_image = NULL; 
        }
     }


     if (type == RAW || type == PGM) {

	 color_flag = FALSE;
	 orig_flag = GREY;
         read_grey_image (imagefile, input_vector, &orig_image, 
                          &orig_rows, &orig_columns, 
                          &orig_size, &orig_depth, type);

	raw_image = alloc_image (orig_size, orig_size);
	copy_image (orig_image, raw_image, orig_size, orig_size);
     }

     if (type == PPM) {
        color_flag = TRUE;
        orig_flag = COLOR;
	read_color_image (input_vector, &orig_image_red, &orig_image_green,
			  &orig_image_blue, orig_rows, orig_columns, &orig_size,			  &orig_depth);

	raw_image_red = alloc_image (orig_size, orig_size);
	copy_image (orig_image_red, raw_image_red, orig_size, orig_size);
	raw_image_blue = alloc_image (orig_size, orig_size);
	copy_image (orig_image_blue, raw_image_blue, orig_size, orig_size);
	raw_image_green = alloc_image (orig_size, orig_size);
	copy_image (orig_image_green, raw_image_green, orig_size, orig_size);
      }

      rows = orig_rows; columns = orig_columns; 
      size = orig_size; depth = orig_depth;
      xtop = ytop = 0;
      free (input_vector);
	
      transf_done = FALSE;
      mode = IMAGE;
      set_label (mode_label, "IMAGE MODE");

      if (xv_flag) { 
	  xv_display (); 
      } else {
          if (image_flag) XtUnmanageChild (pop_window);

          if (type == PGM || type == RAW) {
              build_grey_table (gtable);
              create_grey_image (gtable, raw_image, rows, columns);
	      disp_ok = TRUE;
          } else if (type == PPM) {
	     disp_ok = create_color_image (raw_image_red, raw_image_green, 
				       raw_image_blue, rows, columns);
	      color_flag = TRUE;
          }

          if (disp_ok) {
              resize_drawing_area (rows, columns);
              XPutImage (XtDisplay (canvas), XtWindow (canvas), 
		         theGC, image, 0, 0, 0, 0, columns, rows);
             xv_disp_flag = FALSE;
          }
       }
       image_flag = TRUE;
   }
}


void concat_images (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmFileSelectionBoxCallbackStruct *call_data;
{
  char *imagefile;
  unsigned char *input_vector;
  imgtype **second_image, **concat_image;
  imgtype **second_image_red, **concat_image_red;
  imgtype **second_image_green, **concat_image_green;
  imgtype **second_image_blue, **concat_image_blue;
  int  type, i, j;
  int  srows, scolumns, ssize, sdepth;
  int  nrows, ncolumns, nsize, ndepth;
  float factor, sz;
  void  save_prev_image ();

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &imagefile));
  XtUnmanageChild (dialog_box);

  if (imagefile != NULL) {
     save_prev_image ();
     input_vector = read_pnm_file (imagefile, &type, &srows, &scolumns);

     if (input_vector == NULL) {
	error ("Error in reading of the second image - aborting concatenation");
	return;
     }

     if (color_flag && type == PPM) {
	read_color_image (input_vector, &second_image_red, 
			  &second_image_green,
			  &second_image_blue, srows, scolumns, 
			  &ssize, &sdepth);

    } else if (type == RAW || type == PGM) {
        read_grey_image(imagefile, input_vector, &second_image, 
                         &srows, &scolumns, &ssize, &sdepth, type);
    } else {
	error ("Cannot read both color and greyscale image!");
        free (input_vector);
        return;
    }

    nrows = (rows > srows) ? rows : srows;
    ncolumns = scolumns + columns;
    sz = (double) ((nrows >= ncolumns)? nrows : ncolumns);
    ndepth = ceil (log(sz)/ log ((double) 2.0));
    for (nsize = 1, i = 0; i < ndepth; i++, nsize *= 2);

    printf ("First image: rows = %d, columns = %d, size = %d\n", rows, columns, size);
    printf ("Second image: rows = %d, columns = %d, size = %d\n", srows, scolumns, ssize);
    printf ("Concatenated image: rows = %d, columns = %d, size = %d\n", nrows, ncolumns, nsize);

    if (color_flag) {
	concat_image_red = alloc_image (nsize, nsize);
	concat_image_blue = alloc_image (nsize, nsize);
	concat_image_green = alloc_image (nsize, nsize);

             for (i = 0; i < nsize; i++) {
	         for (j = 0; j < nsize; j++) {
		      if (i < ytop + rows && j < xtop + columns) {
		          if (crop_flag) {
		            concat_image_red[i][j] = crop_image_red[i][j];
		            concat_image_green[i][j] = crop_image_green[i][j];
		            concat_image_blue[i][j] = crop_image_blue[i][j];
		          } else {
		            concat_image_red[i][j] = raw_image_red[i][j];
		            concat_image_green[i][j] = raw_image_green[i][j];
		            concat_image_blue[i][j] = raw_image_blue[i][j];
                         }
		      } else if (i < ytop+srows && j <= xtop + columns+scolumns) {
			    concat_image_red[i][j] = second_image_red[i][j-(xtop+columns)];
			    concat_image_green[i][j] = second_image_green[i][j-(xtop+columns)];
			    concat_image_blue[i][j] = second_image_blue[i][j-(xtop+columns)];
		      } else {
			    concat_image_red[i][j] = 0;
			    concat_image_green[i][j] = 0;
			    concat_image_blue[i][j] = 0;
		      }
                  }
              }
		       
             if (crop_flag) { 
                free_image(crop_image_red, size); 
                free_image(crop_image_green, size); 
                free_image(crop_image_blue, size); 
	        crop_image_red = concat_image_red;
	        crop_image_green = concat_image_green;
	        crop_image_blue = concat_image_blue;
             } else {
                free_image (raw_image_red, size); 
                free_image (raw_image_green, size); 
                free_image (raw_image_blue, size); 
	        raw_image_red = concat_image_red;
	        raw_image_green = concat_image_green;
	        raw_image_blue = concat_image_blue;
             }

             free_image (second_image_red,ssize); 
             free_image (second_image_green,ssize); 
             free_image (second_image_blue,ssize); 

      } else {

	     concat_image = alloc_image (nsize, nsize);

             for (i = 0; i < nsize; i++) {
	         for (j = 0; j < nsize; j++) {
		      if (i < ytop+rows && j < xtop+columns) {
		          if (crop_flag)
		            concat_image[i][j] = crop_image[i][j];
		          else
		            concat_image[i][j] = raw_image[i][j];
		      } else if (i < ytop+srows && j <= xtop + columns+scolumns) {
			    concat_image[i][j] = second_image[i][j-(xtop+columns)];
		      } else {
			    concat_image[i][j] = 0;
		      }
                  }
              }
		       
             if (crop_flag) { 
                free_image(crop_image, size); 
	        crop_image = concat_image;
             } else {
                free_image (raw_image, size); 
	        raw_image = concat_image;
             }

             free_image (second_image,ssize); 
         }
         rows = nrows; columns = ncolumns; 
         size = nsize; depth = ndepth;
     	 transf_done = TRUE; 
         image_flag = TRUE;

         if (!xv_flag) XtUnmanageChild (pop_window);
	 display_image (FALSE);
   }
}


void add_images (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmFileSelectionBoxCallbackStruct *call_data;
{
  char *imagefile;
  unsigned char *input_vector;
  imgtype **second_image;
  imgtype **second_image_red;
  imgtype **second_image_green;
  imgtype **second_image_blue;
  int  type, i, j, p, q, res;
  int  srows, scolumns, ssize, sdepth;
  int  nrows, ncolumns;
  int  imin, imax, jmin, jmax;
  float factor;
  void shrink_image (), zoom_image (), save_prev_image ();

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &imagefile));
  XtUnmanageChild (dialog_box);

  if (imagefile != NULL) {
     save_prev_image ();
     input_vector = read_pnm_file (imagefile, &type, 
				   &srows, &scolumns);

     if (input_vector == NULL) {
	error ("Error in reading of the second image - aborting addition");
	return;
     }

     if (color_flag && type == PPM) {
	read_color_image (input_vector, &second_image_red, 
			  &second_image_green,
			  &second_image_blue, srows, scolumns, 
			  &ssize, &sdepth);

    } else if (type == RAW || type == PGM) {
        read_grey_image(imagefile, input_vector, &second_image, 
                         &srows, &scolumns, &ssize, &sdepth, type);
    } else {
	error ("Cannot read both color and greyscale image!");
        free (input_vector);
        return;
    }

        printf ("First image: rows = %d, columns = %d, size = %d\n", rows, columns, size); fflush (stdout);
        printf ("Second image: rows = %d, columns = %d, size = %d\n", srows, scolumns, ssize); fflush (stdout);

    factor = ((float) size)/ssize;
        printf ("factor = %f\n", factor);
    if (ssize < size) {
	 if (color_flag) {
	     zoom_image (&second_image_red, ssize, size);
	     zoom_image (&second_image_green, ssize, size);
	     zoom_image (&second_image_blue, ssize, size);
	} else {
	     zoom_image (&second_image, ssize, size);
        }
	srows *= factor; scolumns *= factor; ssize = size;
        printf ("Second image (adjusted) : rows = %d, columns = %d, size = %d\n", srows, scolumns, ssize); fflush (stdout);
    }
		
    if (ssize > size) {
	 if (color_flag) {
	     shrink_image (&second_image_red, ssize, size);
	     shrink_image (&second_image_green, ssize, size);
	     shrink_image (&second_image_blue, ssize, size);
	} else {
	     shrink_image (&second_image, ssize, size);
        }
	srows *= factor; scolumns *= factor; ssize = size;
        printf ("Second image (adjusted) : rows = %d, columns = %d, size = %d\n", srows, scolumns, ssize); fflush (stdout);
    }

        imax = nrows = (rows > srows) ? rows : srows;
        jmax = ncolumns = (columns > scolumns) ? columns : scolumns;
        printf ("Result: rows = %d, columns = %d, size = %d\n", nrows, ncolumns, size); fflush (stdout);

	imin = (rows <= srows) ? rows : srows;
	jmin = (columns <= scolumns) ? columns : scolumns;

        if (color_flag) {
             for (i = 0; i < size; i++) {
	         for (j = 0; j < size; j++) {
		      if (i < imin && j < jmin) {
		          if (crop_flag) {
			    p = 255 - crop_image_red[i+ytop][j+xtop];
			    q = 255 - second_image_red[i][j];
			    res = (p + q > 255) ? 255: p+q;
		            crop_image_red[i][j+xtop] = 255 - res;

			    p = 255 - crop_image_green[i+ytop][j+xtop];
			    q = 255 - second_image_green[i][j];
			    res = (p + q > 255) ? 255: p+q;
		            crop_image_green[i][j+xtop] = 255 - res;

			    p = 255 - crop_image_blue[i+ytop][j+xtop];
			    q = 255 - second_image_blue[i][j];
			    res = (p + q > 255) ? 255: p+q;
		            crop_image_blue[i][j+xtop] = 255 - res;
		          } else {
			    p = 255 - raw_image_red[i+ytop][j+xtop];
			    q = 255 - second_image_red[i][j];
			    res = (p + q > 255) ? 255: p+q;
		            raw_image_red[i][j] = 255 - res;

			    p = 255 - raw_image_green[i+ytop][j+xtop];
			    q = 255 - second_image_green[i][j];
			    res = (p + q > 255) ? 255: p+q;
		            raw_image_green[i][j] = 255 - res;

			    p = 255 - raw_image_blue[i+ytop][j+xtop];
			    q = 255 - second_image_blue[i][j];
			    res = (p + q > 255) ? 255: p+q;
		            raw_image_blue[i][j] = 255 - res;
                         }
                    } else {
		      if (srows <= rows && scolumns <= columns) {
                         if (i < imax && j < jmax) {
		            if (crop_flag) {
			       crop_image_red[i][j] = crop_image_red[i+ytop][j+xtop];
			       crop_image_green[i][j] = crop_image_green[i+ytop][j+xtop];
			       crop_image_blue[i][j] = crop_image_blue[i+ytop][j+xtop];
		            } else {
			       raw_image_red[i][j] = raw_image_red[i+ytop][j+xtop];
			       raw_image_green[i][j] = raw_image_green[i+ytop][j+xtop];
			       raw_image_blue[i][j] = raw_image_blue[i+ytop][j+xtop];
		            }
			} else {
		            if (crop_flag) {
			       crop_image_red[i][j] = 0;
			       crop_image_green[i][j] = 0;
			       crop_image_blue[i][j] = 0;
		            } else {
			       raw_image_red[i][j] = 0;
			       raw_image_green[i][j] = 0;
			       raw_image_blue[i][j] = 0;
		            }
			}
		
                    } else if (srows > rows && scolumns <= columns) {
			   if (i < imax && j < scolumns) {
		              if (crop_flag) {
			          crop_image_red[i][j] = second_image_red[i][j];
			          crop_image_green[i][j] = second_image_green[i][j];
			          crop_image_blue[i][j] = second_image_blue[i][j];
		              } else {
			         raw_image_red[i][j] = second_image_red[i][j];
			         raw_image_green[i][j] = second_image_green[i][j];
			         raw_image_blue[i][j] = second_image_blue[i][j];
		              }
			
                           } else if (i < rows && j < jmax) {
		            if (crop_flag) {
			       crop_image_red[i][j] = crop_image_red[i+ytop][j+xtop];
			       crop_image_green[i][j] = crop_image_green[i+ytop][j+xtop];
			       crop_image_blue[i][j] = crop_image_blue[i+ytop][j+xtop];
		            } else {
			       raw_image_red[i][j] = raw_image_red[i+ytop][j+xtop];
			       raw_image_green[i][j] = raw_image_green[i+ytop][j+xtop];
			       raw_image_blue[i][j] = raw_image_blue[i+ytop][j+xtop];
		            }
			} else {
		            if (crop_flag) {
			       crop_image_red[i][j] = 0;
			       crop_image_green[i][j] = 0;
			       crop_image_blue[i][j] = 0;
		            } else {
			       raw_image_red[i][j] = 0;
			       raw_image_green[i][j] = 0;
			       raw_image_blue[i][j] = 0;
		            }
                        }
                    } else if (srows <= rows && scolumns > columns) {
			   if (i < srows && j < jmax) {
		              if (crop_flag) {
			          crop_image_red[i][j] = second_image_red[i][j];
			          crop_image_green[i][j] = second_image_green[i][j];
			          crop_image_blue[i][j] = second_image_blue[i][j];
		              } else {
			         raw_image_red[i][j] = second_image_red[i][j];
			         raw_image_green[i][j] = second_image_green[i][j];
			         raw_image_blue[i][j] = second_image_blue[i][j];
		              }
			
                           } else if (i < imax && j < columns) {
		            if (crop_flag) {
			       crop_image_red[i][j] = crop_image_red[i+ytop][j+xtop];
			       crop_image_green[i][j] = crop_image_green[i+ytop][j+xtop];
			       crop_image_blue[i][j] = crop_image_blue[i+ytop][j+xtop];
		            } else {
			       raw_image_red[i][j] = raw_image_red[i+ytop][j+xtop];
			       raw_image_green[i][j] = raw_image_green[i+ytop][j+xtop];
			       raw_image_blue[i][j] = raw_image_blue[i+ytop][j+xtop];
		            }
			} else {
		            if (crop_flag) {
			       crop_image_red[i][j] = 0;
			       crop_image_green[i][j] = 0;
			       crop_image_blue[i][j] = 0;
		            } else {
			       raw_image_red[i][j] = 0;
			       raw_image_green[i][j] = 0;
			       raw_image_blue[i][j] = 0;
		            }
                        }
			
		      } else {
                         if (i < imax && j < jmax) {
		            if (crop_flag) {
			       crop_image_red[i][j] = second_image_red[i][j];
			       crop_image_green[i][j] = second_image_green[i][j];
			       crop_image_blue[i][j] = second_image_blue[i][j];
		            } else {
			       raw_image_red[i][j] = second_image_red[i][j];
			       raw_image_green[i][j] = second_image_green[i][j];
			       raw_image_blue[i][j] = second_image_blue[i][j];
		            }
			} else {
		            if (crop_flag) {
			       crop_image_red[i][j] = 0;
			       crop_image_green[i][j] = 0;
			       crop_image_blue[i][j] = 0;
		            } else {
			       raw_image_red[i][j] = 0;
			       raw_image_green[i][j] = 0;
			       raw_image_blue[i][j] = 0;
		            }
			}
                     }
                 }
			
               }
             }

             for (i=0; i < ssize; i++) 
                 free ((second_image_red)[i]); 
             for (i=0; i < ssize; i++) 
                 free ((second_image_green)[i]); 
             for (i=0; i < ssize; i++) 
                 free ((second_image_blue)[i]); 
             free (second_image_red); 
             free (second_image_green); 
             free (second_image_blue); 

      } else {

             for (i = 0; i < size; i++) {
	         for (j = 0; j < size; j++) {
		      if (i < imin && j < jmin) {
		          if (crop_flag) {
			    p = 255 - crop_image[i+ytop][j+xtop];
			    q = 255 - second_image[i][j];
			    res = (p + q > 255) ? 255: p+q;
		            crop_image[i][j] = 255 - res;
		          } else {
			    p = 255 - raw_image[i+ytop][j+xtop];
			    q = 255 - second_image[i][j];
			    res = (p + q > 255) ? 255: p+q;
		            raw_image[i][j] = 255 - res;
			   }
		      } else {  
		         if (srows <= rows && scolumns <= columns) {
                            if (i < imax && j < jmax) {
		               if (crop_flag) 
			          crop_image[i][j] = crop_image[i+ytop][j+xtop];
		               else 
			          raw_image[i][j] = raw_image[i+ytop][j+xtop];
			    } else {
		               if (crop_flag) 
			          crop_image[i][j] = 0;
		               else 
			          raw_image[i][j] = 0;
		            }
		
                    } else if (srows > rows && scolumns <= columns) {
			   if (i < imax && j < scolumns) {
		              if (crop_flag) 
			          crop_image[i][j] = second_image[i][j];
		               else 
			         raw_image[i][j] = second_image[i][j];
			
                           } else if (i < rows && j < jmax) {
		            if (crop_flag) 
			       crop_image[i][j] = crop_image[i+ytop][j+xtop];
		            else 
			       raw_image_red[i][j] = raw_image_red[i][j];
			} else {
		            if (crop_flag) 
			       crop_image[i][j] = 0;
		            else 
			       raw_image[i][j] = 0;
                        }
                    } else if (srows <= rows && scolumns > columns) {
			   if (i < srows && j < jmax) {
		              if (crop_flag) {
			          crop_image[i][j] = second_image[i][j];
		              } else {
			         raw_image[i][j] = second_image[i][j];
		              }
			
                           } else if (i < imax && j < columns) {
		            if (crop_flag) {
			       crop_image[i][j] = crop_image[i+ytop][j+xtop];
		            } else {
			       raw_image[i][j] = raw_image[i+ytop][j+xtop];
		            }
			} else {
		            if (crop_flag) {
			       crop_image[i][j] = 0;
		            } else {
			       raw_image[i][j] = 0;
		            }
                        }
			
		      } else {
                         if (i < imax && j < jmax) {
		            if (crop_flag) {
			       crop_image[i][j] = second_image[i][j];
		            } else {
			       raw_image[i][j] = second_image[i][j];
		            }
			} else {
		            if (crop_flag) {
			       crop_image[i][j] = 0;
		            } else {
			       raw_image[i][j] = 0;
		            }
			}
                     }
                 }
               }
              }
		       
             for (i=0; i < ssize; i++) 
                 free ((second_image)[i]); 
             free (second_image); 
         }
         image_flag = TRUE;
         transf_done = TRUE; 
	rows = nrows; columns = ncolumns;
	xtop = 0; ytop = 0;

         if (!xv_flag) XtUnmanageChild (pop_window);
	 display_image (FALSE);
   }
}



/* User selected ok to read the wft */

void wft_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmFileSelectionBoxCallbackStruct *call_data;
{
  char *name, str[50];
  FILE *infile;
  int  ntapes = 2, read_ok;
  void	clear_wft_labels ();

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &name)) read_ok = TRUE;
  else read_ok = FALSE;
  XtUnmanageChild (dialog_box);
  
  if (read_ok) {


	if (wft_selected && wft_load_type == SINGLE) {
            printf ("Deleting the old automaton..."); fflush(stdout);
	    if (wft_unrolled) 
               free_autom_values (&wft, usize);
	    free_automaton (&wft);
            printf ("Done\n"); fflush (stdout);
        }
	if (red_wft_selected && wft_load_type == RED) { 
            printf ("Deleting the old automaton..."); fflush(stdout);
	    if (red_wft_unrolled) 
               free_autom_values (&wft_red, usize);
	    free_automaton (&wft_red);
            printf ("Done\n"); fflush (stdout);
        }
	if (green_wft_selected && wft_load_type == GREEN) {
            printf ("Deleting the old automaton..."); fflush(stdout);
	    if (green_wft_unrolled) free_autom_values (&wft_green, usize);
	    free_automaton (&wft_green);
            printf ("Done\n"); fflush (stdout);
        }
	if (blue_wft_selected && wft_load_type == BLUE) {
            printf ("Deleting the old automaton..."); fflush(stdout);
	    if (blue_wft_unrolled) free_autom_values (&wft_blue, usize);
	    free_automaton (&wft_blue);
            printf ("Done\n"); fflush (stdout);
        }
	infile = fopen (name, "r");
	if (infile == NULL) { 
	    error ("Can't open input wft file"); return;
        }

        crop_name (&name);

	switch (wft_load_type) {
                case SINGLE: if (!read_automaton (infile, &wft, ntapes)) {
				error ("Error in reading the input WFT file");
				fclose (infile); return; 
			      }
  		     	     wft_selected = TRUE;
			     wft_unrolled = FALSE;
			     clear_wft_labels ();
			     sprintf (str, "Loaded WFT: %s", name); 
			     set_label (wft_label, str);
     			     switch (wft.method) {
			       case SIMPLE:
             			  set_label (mlabel,"method: simple");
				  break;
			       case GENERAL:
             			  set_label(mlabel, "method: general");
				  break;
     			     }
			     break;

		case RED:    if (!read_automaton (infile, &wft_red, ntapes)) {
				error ("Error in reading the input WFT file");
				fclose (infile); return; 
			      }
			     if (!red_wft_selected) nloaded++;
  		             red_wft_selected = TRUE;
			     red_wft_unrolled = FALSE;
			     set_label (wft_label, "Loaded WFT:");
			     set_label (mlabel, "");
			     sprintf (str, "Red component: %s", name); 
     			     switch (wft_red.method) {
			       case SIMPLE:
             			  set_label (mlabel_red,"method: simple");
				  break;
			       case GENERAL:
             			  set_label(mlabel_red, "method: general");
				  break;
			     }
			     break;

		case GREEN:  if (!read_automaton (infile, &wft_green, ntapes)) {
				error ("Error in reading the input WFT file");
				fclose (infile); return; 
			     }
			     if (!green_wft_selected) nloaded++;
  		             green_wft_selected = TRUE;
			     green_wft_unrolled = FALSE;
			     set_label (wft_label, "Loaded WFT:");
			     set_label (mlabel, "");
			     sprintf (str, "Green component: %s", name); 
     			     switch (wft_green.method) {
			       case SIMPLE:
             			  set_label (mlabel_green,"method: simple");
				  break;
			       case GENERAL:
             			  set_label(mlabel_green, "method: general");
				  break;
			     }
			     break;

		case BLUE:   if (!read_automaton (infile, &wft_blue, ntapes)) {
				error ("Error in reading the input WFT file");
				fclose (infile); return; 
			     }
			     if (!blue_wft_selected) nloaded++;
  		             blue_wft_selected = TRUE;
			     blue_wft_unrolled = FALSE;
			     set_label (wft_label, "Loaded WFT:");
			     set_label (mlabel, "");
			     sprintf (str, "Blue component: %s", name); 
     			     switch (wft_blue.method) {
			       case SIMPLE:
             			  set_label (mlabel_blue,"method: simple");
				  break;
			       case GENERAL:
             			  set_label(mlabel_blue, "method: general");
				  break;
			     }
			     break;
        }
	fclose (infile);
	wft_started = FALSE;
	usize = 0; wft_depth = 0;
	switch (wft_load_type) {
		case RED: set_label (red_label, str); break;
		case GREEN: set_label (green_label, str); break;
		case BLUE: set_label (blue_label, str); break;
        }
	/*printf ("The automaton %s has been read successfully.\n", name);*/
  }
}



void red_wft_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmFileSelectionBoxCallbackStruct *call_data;
{
  char *name, str[50];
  FILE *infile;
  int  ntapes = 2, read_ok;
  void green_wft_ok ();

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &name)) read_ok = TRUE;
  else read_ok = FALSE;
  XtUnmanageChild (dialog_box);
  
  if (read_ok) {
	if (red_wft_selected) { 
            printf ("Deleting the old automaton..."); fflush(stdout);
	    if (red_wft_unrolled) free_autom_values (&wft_red, usize);
	    free_automaton (&wft_red);
            printf ("Done\n"); fflush (stdout);
        }
	infile = fopen (name, "r");
	if (infile == NULL) { 
	    error ("Can't open input wft file"); return;
        }
	if (!read_automaton (infile, &wft_red, ntapes)) {
	  error ("Error in reading the input WFT file");
	  fclose (infile); return; 
	}
	crop_name (&name);
  	red_wft_selected = TRUE;
        red_wft_unrolled = FALSE;
	fclose (infile);
	wft_started = FALSE;
	usize = 0; wft_depth = 0;
	set_label (wft_label, "Loaded WFT:");
	set_label (mlabel, "");
	printf ("The automaton %s has been read successfully.\n", name);
        sprintf (str, "Red component: %s", name); 
        switch (wft_red.method) {
	       case SIMPLE:
          		set_label (mlabel_red,"method: simple");
			break;
	       case GENERAL:
             	        set_label(mlabel_red, "method: general");
			break;
        }
	set_label (red_label, str);
        display_file_sb ("Load green WFT", WFT_MASK, green_wft_ok);
  }
}

void green_wft_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmFileSelectionBoxCallbackStruct *call_data;
{
  char *name, str[50];
  FILE *infile;
  int  ntapes = 2, read_ok;
  void blue_wft_ok ();

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &name)) read_ok = TRUE;
  else read_ok = FALSE;
  XtUnmanageChild (dialog_box);
  
  if (read_ok) {
	if (green_wft_selected) { 
            printf ("Deleting the old automaton..."); fflush(stdout);
	    if (green_wft_unrolled) free_autom_values (&wft_green, usize);
	    free_automaton (&wft_green);
            printf ("Done\n"); fflush (stdout);
        }
	infile = fopen (name, "r");
	if (infile == NULL) { 
	    error ("Can't open input wft file"); return;
        }
	if (!read_automaton (infile, &wft_green, ntapes)) {
	  error ("Error in reading the input WFT file");
	  fclose (infile); return; 
	}
	crop_name (&name);
  	green_wft_selected = TRUE;
        green_wft_unrolled = FALSE;
	fclose (infile);
	wft_started = FALSE;
	usize = 0; wft_depth = 0;
	printf ("The automaton %s has been read successfully.\n", name);
        sprintf (str, "Green component: %s", name); 
        switch (wft_green.method) {
	       case SIMPLE:
          		set_label (mlabel_green,"method: simple");
			break;
	       case GENERAL:
             	        set_label(mlabel_green, "method: general");
			break;
        }
	set_label (green_label, str);
        display_file_sb ("Load blue WFT", WFT_MASK, blue_wft_ok);
  }
}


void blue_wft_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmFileSelectionBoxCallbackStruct *call_data;
{
  char *name, str[50];
  FILE *infile;
  int  ntapes = 2, read_ok;

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &name)) read_ok = TRUE;
  else read_ok = FALSE;
  XtUnmanageChild (dialog_box);
  
  if (read_ok) {
	if (blue_wft_selected) { 
            printf ("Deleting the old automaton..."); fflush(stdout);
	    if (blue_wft_unrolled) free_autom_values (&wft_blue, usize);
	    free_automaton (&wft_blue);
            printf ("Done\n"); fflush (stdout);
        }
	infile = fopen (name, "r");
	if (infile == NULL) { 
	    error ("Can't open input wft file"); return;
        }
        crop_name (&name);
	if (!read_automaton (infile, &wft_blue, ntapes)) {
	  error ("Error in reading the input WFT file");
	  fclose (infile); return; 
	}
  	blue_wft_selected = TRUE;
        blue_wft_unrolled = FALSE;
	fclose (infile);
	wft_started = FALSE;
	wft_load_type = MULTIPLE;
	usize = 0; wft_depth = 0;
	printf ("The automaton %s has been read successfully.\n", name);
        sprintf (str, "Blue component: %s", name); 
        switch (wft_blue.method) {
	       case SIMPLE:
          		set_label (mlabel_blue,"method: simple");
			break;
	       case GENERAL:
             	        set_label(mlabel_blue, "method: general");
			break;
        }
	set_label (blue_label, str);
  }
}

void red_wfa_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmFileSelectionBoxCallbackStruct *call_data;
{
  char *name, str[50];
  FILE *infile;
  int  ntapes = 1, read_ok;
  void green_wfa_ok ();

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &name)) read_ok = TRUE;
  else read_ok = FALSE;
  XtUnmanageChild (dialog_box);
  
  if (read_ok) {
	syn_flag = FALSE;
	adj_flag = FALSE;
	if (red_wfa_selected) { 
            printf ("Deleting the old automaton..."); fflush(stdout);
	    free_automaton (&wfa_red);
            printf ("Done\n"); fflush (stdout);
        }
	infile = fopen (name, "r");
	if (infile == NULL) { 
	    error ("Can't open input wft file"); return;
        }
        crop_name (&name);
	if (!read_automaton (infile, &wfa_red, ntapes)) {
	  error ("Error in reading the input WFA file");
	  fclose (infile); return; 
	}
	set_label (wfa_label, "Loaded WFA:");
	fclose (infile);
	printf ("The automaton %s has been read successfully.\n", name);
        sprintf (str, "Red component: %s", name); 
	set_label (red_wfa_label, str);
	color_aut = TRUE;
	if (red_wfa_selected) free_automaton (&orig_wfa_red);
	orig_red_wfa_name = name;
	copy_automaton (&wfa_red, &orig_wfa_red);
  	red_wfa_selected = TRUE;
	scaling_flag = FALSE;
	cutoff_flag = FALSE;
	wfa_depth = 0;
        display_file_sb ("Load green WFA", WFA_MASK, green_wfa_ok);
  }
}

void green_wfa_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmFileSelectionBoxCallbackStruct *call_data;
{
  char *name, str[50];
  FILE *infile;
  int  ntapes = 1, read_ok;
  void blue_wfa_ok ();

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &name)) read_ok = TRUE;
  else read_ok = FALSE;
  XtUnmanageChild (dialog_box);
  
  if (read_ok) {
	syn_flag = FALSE;
	adj_flag = FALSE;
	if (green_wfa_selected)  
	    free_automaton (&wfa_green);
        
	infile = fopen (name, "r");
	if (infile == NULL) { 
	    error ("Can't open input wfa file"); return;
        }
        crop_name (&name);
	if (!read_automaton (infile, &wfa_green, ntapes)) {
	  error ("Error in reading the input WFA file");
	  fclose (infile); return; 
	}
	fclose (infile);
	if (green_wfa_selected) free_automaton (&orig_wfa_green);
	orig_green_wfa_name = name;
	copy_automaton (&wfa_green, &orig_wfa_green);
  	green_wfa_selected = TRUE;
	printf ("The automaton %s has been read successfully.\n", name);
        sprintf (str, "Green component: %s", name); 
	set_label (green_wfa_label, str);
	wfa_depth = 0;
        display_file_sb ("Load blue WFA", WFA_MASK, blue_wfa_ok);
  }
}


void blue_wfa_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmFileSelectionBoxCallbackStruct *call_data;
{
  char *name, str[50];
  FILE *infile;
  int  ntapes = 1, read_ok;

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &name)) read_ok = TRUE;
  else read_ok = FALSE;
  XtUnmanageChild (dialog_box);
  
  if (read_ok) {
	syn_flag = FALSE;
	adj_flag = FALSE;
	if (blue_wfa_selected)  
	    free_automaton (&wft_blue);
        
	infile = fopen (name, "r");
	if (infile == NULL) { 
	    error ("Can't open input wft file"); return;
        }
	if (!read_automaton (infile, &wfa_blue, ntapes)) {
	  error ("Error in reading the input WFA file");
	  fclose (infile); return; 
	}
	crop_name (&name);
	fclose (infile);
	if (blue_wfa_selected) free_automaton (&orig_wfa_blue);
	orig_blue_wfa_name = name;
	copy_automaton (&wfa_blue, &orig_wfa_blue);
  	blue_wfa_selected = TRUE;
	wfa_load_type = MULTIPLE;
	color_aut = TRUE;
	printf ("The automaton %s has been read successfully.\n", name);
        sprintf (str, "Blue component: %s", name); 
	set_label (blue_wfa_label, str);
	wfa_depth = 0;
	mode = WFA;
	set_label (mode_label, "WFA MODE");
  }
}

/* User selected ok to read the wfa */


void wfa_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmFileSelectionBoxCallbackStruct *call_data;
{
  char *name, str[50];
  FILE *infile;
  int  ntapes = 1, read_ok;
  char c;

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &name)) read_ok = TRUE;
  else read_ok = FALSE;
  XtUnmanageChild (dialog_box);
  
  if (read_ok) {
	infile = fopen (name, "r");
	if (infile == NULL) { 
            error ("Can't open input wfa file"); return;
        }

	crop_name (&name);
	syn_flag = FALSE;
	adj_flag = FALSE;


        if ((c = fgetc(infile)) == 'c') {
	if ((c = fgetc (infile)) != '\n') ungetc (c, infile);
	    color_aut = TRUE;
	    wfa_load_type = SINGLE;
	    if (color_wfa_selected) {
		free_automaton (&wfa_red);
		free_automaton (&wfa_green);
		free_automaton (&wfa_blue);
	    }
            if (!read_color_automaton (infile, &wfa_red, &wfa_green,
				       &wfa_blue, ntapes)) {
		error ("Error in reading the input wfa file"); 
		fclose (infile); return;
	    }
	    fclose (infile);
	    if (red_wfa_selected) free_automaton (&orig_wfa_red);
	    if (green_wfa_selected) free_automaton (&orig_wfa_green);
	    if (blue_wfa_selected) free_automaton (&orig_wfa_blue);
	    orig_color_wfa_name = name;
	    copy_automaton (&wfa_red, &orig_wfa_red);
	    copy_automaton (&wfa_green, &orig_wfa_green);
	    copy_automaton (&wfa_blue, &orig_wfa_blue);
  	    red_wfa_selected = TRUE;
  	    green_wfa_selected = TRUE;
  	    blue_wfa_selected = TRUE;
            color_wfa_selected = TRUE;
        } else {
	    ungetc (c,infile);
	    color_aut = FALSE;
	    wfa_load_type = SINGLE;
	    if (wfa_selected) free_automaton (&wfa);
            if (!read_automaton (infile, &wfa, ntapes)) { 
		error ("Error in reading the input WFA file");
		fclose (infile);
		return;
	    }
	    fclose (infile);
	    if (wfa_selected) free_automaton (&orig_wfa);
	    orig_wfa_name = name;
	    copy_automaton (&wfa, &orig_wfa);
  	    wfa_selected = TRUE;
         }

	clear_wfa_labels ();
	printf ("The automaton %s has been read successfully.\n", name);
	sprintf (str, "Loaded WFA: %s", name);
	set_label (wfa_label, str);
	mode = WFA;
	scaling_flag = FALSE;
	cutoff_flag = FALSE;
	wfa_depth = 0;
	set_label (mode_label, "WFA MODE");
  }
}



void load_orig_wfa (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  char	str[50];

   if (color_aut && red_wfa_selected && 
       green_wfa_selected && blue_wfa_selected) {
	reload_color_wfa ();
   } else  if (wfa_selected) {
     reload_wfa ();
   } else 
      error ("Please load the WFA first");
}


void rgb_wfa_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmSelectionBoxCallbackStruct *call_data;
{
  char *name, str[50], filenamer[50], filenameg[50], filenameb[50];
  char str1[50], str2[50], str3[50];
  FILE *infiler, *infileg, *infileb;
  int  ntapes = 1, read_ok;

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &name)) read_ok = TRUE;
  else read_ok = FALSE;
  XtUnmanageChild (dialog_box);
  
  if (read_ok) {
	syn_flag = FALSE;
	adj_flag = FALSE;
	if (red_wfa_selected) free_automaton (&wfa_red);
	if (green_wfa_selected) free_automaton (&wfa_green);
	if (blue_wfa_selected) free_automaton (&wfa_blue);
	sprintf (filenamer, "%s.r", name);
	sprintf (filenameg, "%s.g", name);
	sprintf (filenameb, "%s.b", name);
	if (((infiler = fopen (filenamer, "r")) == NULL) ||
	    ((infileg = fopen (filenameg, "r")) == NULL) ||
	    ((infileb = fopen (filenameb, "r")) == NULL)) 
        {
            error ("Can't open input wfa file"); return;
        }

        if (!read_automaton (infiler, &wfa_red, ntapes)) { 
	     error ("Error in reading the input red WFA file");
	     fclose (infiler); return;
        }

        if (!read_automaton (infileg, &wfa_green, ntapes)) { 
	     error ("Error in reading the input green WFA file");
	     fclose (infileg); return;
        }

        if (!read_automaton (infileb, &wfa_blue, ntapes)) { 
	     error ("Error in reading the input blue WFA file");
	     fclose (infileb); return;
	}

	fclose (infiler);
	fclose (infileg);
	fclose (infileb);

	/* save the WFA files for retrieving */

	if (red_wfa_selected) free_automaton (&orig_wfa_red);
	orig_red_wfa_name = filenamer;
	copy_automaton (&wfa_red, &orig_wfa_red);

	if (green_wfa_selected) free_automaton (&orig_wfa_green);
	orig_green_wfa_name = filenameg;
	copy_automaton (&wfa_green, &orig_wfa_green);

	if (blue_wfa_selected) free_automaton (&orig_wfa_blue);
	orig_blue_wfa_name = filenameb;
	copy_automaton (&wfa_blue, &orig_wfa_blue);

	/* set flags and labels */

	scaling_flag = FALSE;
	cutoff_flag = FALSE;
	color_aut = TRUE;
  	red_wfa_selected = TRUE;
  	green_wfa_selected = TRUE;
  	blue_wfa_selected = TRUE;
	sprintf (str, "Loaded WFA:");
	set_label (wfa_label, str);
	sprintf (str1, "Red component: %s", filenamer);
	sprintf (str2, "Green component: %s", filenameg);
	sprintf (str3, "Blue component: %s", filenameb);
	set_label (red_wfa_label, str1);
	set_label (green_wfa_label, str2);
	set_label (blue_wfa_label, str3);
	wfa_load_type = MULTIPLE;
	wfa_depth = 0;
	printf ("The automaton %s has been read successfully.\n", name);

  }
}



/***************************************************************
 * Callback for saving an image - pops up a dialog box         *
 * for the user to select a format                             *
 ***************************************************************/

void save_pgm (w, name, call_data)
Widget	w;
char    *name;
XmAnyCallbackStruct *call_data;
{

  XtUnmanageChild (save_dialog);
  save_img (name, PGM);
}



void save_raw (w, name, call_data)
Widget	w;
char    *name;
XmAnyCallbackStruct *call_data;
{

  XtUnmanageChild (save_dialog);
  save_img (name, RAW);
}


void save_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmSelectionBoxCallbackStruct *call_data;
{
  char 	    *name;
  int	    result;

/*
  button    buttons[] = {
            {"PGM format", save_pgm},
	    {"RAW format", save_raw}
  };
*/

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &name));
  XtUnmanageChild (dialog_box);
  
  if (name != NULL) {
      if (color_flag) {
          result = save_img (name, PPM);
      } else {
	  result = save_img (name, PGM);

	/* Create a dialog to find out the output file format 

	save_dialog = create_button_dialog ("Save image","Select image format", 
				             buttons, 2, name);

	XtManageChild (save_dialog);
        */
      }
      if (result) printf ("Successfully written %s.\n", name);
   }
}


int save_img (name, type)
char *name;
int  type;
{
  FILE	*outfile;
  int   i,j;

  outfile = fopen (name, "w");
  if (outfile == NULL) {
       error("Can't open output image file"); 
       return (FALSE);
  }


  if (type == PGM) {
	fprintf (outfile, "P5\n");
	fprintf (outfile, "%d %d\n", columns, rows);
	fprintf (outfile, "%d\n", 255);
  }

  if (type == PGM || type == RAW) {
      for (i = ytop; i < ytop + rows; i++)
          for (j = xtop; j < xtop + columns; j++) {
               if (crop_flag)
	          fputc (imgtype_to_char(crop_image[i][j]), outfile);
               else
	          fputc (imgtype_to_char(raw_image[i][j]), outfile);
          }
  }

  if (type == PPM) {
	fprintf (outfile, "P6\n");
	fprintf (outfile, "%d %d\n", columns, rows);
	fprintf (outfile, "%d\n", 255);
        for (i = ytop; i < ytop + rows; i++)
          for (j = xtop; j < xtop + columns; j++) {
               if (crop_flag) {
	          fputc (imgtype_to_char (crop_image_red[i][j]), 
			outfile);
	          fputc (imgtype_to_char (crop_image_green[i][j]),
			 outfile);
	          fputc (imgtype_to_char (crop_image_blue[i][j]),
			 outfile);
               } else {
	          fputc (imgtype_to_char (raw_image_red[i][j]),
			 outfile);
	          fputc (imgtype_to_char (raw_image_green[i][j]), 
			 outfile);
	          fputc (imgtype_to_char (raw_image_blue[i][j]),
			 outfile);
	       }
          }
  }

  fclose (outfile);
  return (TRUE);
}



void write_wfa (w, dialog_box, call_data)
Widget		w;
Widget		dialog_box;
XmSelectionBoxCallbackStruct *call_data;
{
  char 	    *name, filenamer[50], filenameg[50], filenameb[50];
  int  	    ntapes;
  FILE	    *outfile, *outfiler, *outfileg, *outfileb;

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &name));
  XtUnmanageChild (dialog_box);

  if (name != NULL) {
	ntapes = 1;
        if (color_aut && wfa_load_type == MULTIPLE) {
	    sprintf (filenamer, "%s.r", name);
	    sprintf (filenameg, "%s.g", name);
	    sprintf (filenameb, "%s.b", name);
     	    outfiler = fopen (filenamer, "w");
     	    outfileg = fopen (filenameg, "w");
     	    outfileb = fopen (filenameb, "w");
     	    if (outfiler == NULL || outfileb == NULL || outfiler == NULL) {
          	error("Can't open output automaton file"); 
          	return;
     	    }

	    write_automaton (outfiler, &wfa_red, ntapes);
	    write_automaton (outfileg, &wfa_green, ntapes);
	    write_automaton (outfileb, &wfa_blue, ntapes);
	    fclose (outfiler);
	    fclose (outfileg);
	    fclose (outfileb);
	    printf ("Succesfully written %s.\n", filenamer);
	    printf ("Succesfully written %s.\n", filenameg);
	    printf ("Succesfully written %s.\n", filenameb);
        } else if (color_aut && wfa_load_type == SINGLE) {
     	    outfile = fopen (name, "w");
     	    if (outfile == NULL) {
          	error("Can't open output automaton file"); 
          	return;
     	    }

	    write_color_aut (outfile, &wfa_red, &wfa_green, &wfa_blue, ntapes);
	    fclose (outfile);
	    printf ("Succesfully written %s.\n", name);
        } else {
     	    outfile = fopen (name, "w");
     	    if (outfile == NULL) {
          	error("Can't open output automaton file"); 
          	return;
     	    }

	    write_automaton (outfile, &wfa, ntapes);
	    fclose (outfile);
	    printf ("Succesfully written %s.\n", name);
      }
  }
}
  
void write_wft (w, dialog_box, call_data)
Widget		w;
Widget		dialog_box;
XmSelectionBoxCallbackStruct *call_data;
{
  char 	    *name;
  int  	    ntapes;
  FILE	    *outfile;

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &name));
  XtUnmanageChild (dialog_box);

  if (name != NULL) {
	ntapes = 2;
     	outfile = fopen (name, "w");
     	if (outfile == NULL) {
          	error("Can't open output wft file"); 
          	return;
     	}

	write_automaton (outfile, &wft, ntapes);
	fclose (outfile);
	printf ("Succesfully written %s.\n", name);
  }
}


/*********************************************************
 * CALLBACKS FOR DECODING WFA and WFT and for composition
 **********************************************************/

void apply (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void	apply_single (), apply_multiple (), compose_wfa (), load_wft2 ();
  switch (mode) {
	case IMAGE:
             if ((wft_selected || red_wft_selected || green_wft_selected || 
                  blue_wft_selected) && image_flag) {
	           wft_started = TRUE;

		   if (wft_load_type == SINGLE) 
			apply_single ();
		   else 
			apply_multiple ();
		   
             } else {
	           if (!image_flag)
	              error ("Please read the image first");
	           else
	              error ("Please read the automaton first");
             }
	     break;
	
	case WFA:	
  	     if (wft_selected) {
	         
	         if (color_aut) {
                     compose_wfa (&wfa_red, &wft); 
                     compose_wfa (&wfa_green, &wft); 
                     compose_wfa (&wfa_blue, &wft); 
		     adj_flag = FALSE; syn_flag = FALSE;
		     wfa_depth = 0;
	          } else if (wfa_selected) {
                    compose_wfa (&wfa, &wft); 
		     wfa_depth = 0;
		     adj_flag = FALSE; syn_flag = FALSE;
	          } else
           	   error ("Please load the wfa first");

                } else 
		   error ("Please load the single WFT first");
                break;

	case WFT:
		error ("This operation cannot be used in WFT mode");
		break;
	}
}

void apply_single ()
{
  void  set_scaling_flag (), change_method (), 
        apply_wft (), set_cutoff_flag ();
  char  *text[1], str[10];
  int	s, d;

  char	    *edit_labels[] = {"DEPTH:"};
  radio_button    radio_buttons[] = 
	{"Simple", change_method, SIMPLE, FALSE, 
      	 "General", change_method, GENERAL, FALSE};
  toggle_button   toggle_buttons[] = 
	          {"Scale colors", set_scaling_flag, FALSE,
		   "Cut off large values", set_cutoff_flag, FALSE};

     /* text in the edit widget */
     s = size; 
     if (!power_of_2 (rows) || !power_of_2 (columns))  {
	 d = depth;
         find_size (rows, columns, &s, &d);
     } else {
         d = ((wfa_selected || red_wfa_selected || 
               green_wfa_selected || blue_wfa_selected ) 
               && wfa_depth > 0) ? wfa_depth : depth;
     }

     sprintf (str, "%d", d);
     text[0] = str;


      /* find out which button is to be set */
      if (wft.method == SIMPLE) radio_buttons[0].set = TRUE;
      if (wft.method == GENERAL) radio_buttons[1].set = TRUE;
      toggle_buttons[0].set = scaling_flag = FALSE;
      toggle_buttons[1].set = cutoff_flag = FALSE;

      create_dialog_box ("Apply WFT", edit_labels, text,
			 XtNumber(edit_labels), "METHOD:", 
			 radio_buttons, XtNumber(radio_buttons),
			 NULL, toggle_buttons, XtNumber(toggle_buttons),
			 apply_wft);
 }


void apply_multiple ()
{
  void  set_scaling_flag (), apply_wft (), set_cutoff_flag ();
  char  *text[1], str[10];
  int	s, d;

  char	    *edit_labels[] = {"DEPTH:"};
  toggle_button   toggle_buttons[] =  
	          {"Scale colors", set_scaling_flag, FALSE,
		   "Cut off large values", set_cutoff_flag, FALSE};

    /* text in the edit widget */
     s = size; 
     if (!power_of_2 (rows) || !power_of_2 (columns))  {
         d = depth;
         find_size (rows, columns, &s, &d);
     } else {
         d = ((wfa_selected || red_wfa_selected || 
               green_wfa_selected || blue_wfa_selected ) 
               && wfa_depth > 0) ? wfa_depth : depth;
     }

     sprintf (str, "%d", d);
     text[0] = str;


      /* find out which button is to be set */
      toggle_buttons[0].set = scaling_flag = FALSE;
      toggle_buttons[1].set = cutoff_flag = FALSE;

      create_dialog_box ("Apply WFT", edit_labels, text,
			 XtNumber(edit_labels), NULL, NULL, 0, NULL, 
			 toggle_buttons, XtNumber(toggle_buttons),
			 apply_wft);
 }



void find_wft_matrix (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void  set_scaling_flag (), change_method (), unroll_wft ();

  char	    *edit_labels[] = {"DEPTH:", "OUTPUT FILE:"};
  radio_button    radio_buttons[] = 
	{"Simple", change_method, SIMPLE, FALSE, 
      	 "General", change_method, GENERAL, FALSE};

    if (wft_selected) { 
	/*wft_started = TRUE;*/

	   if (wft_load_type == SINGLE) {

               /* find out which button is to be set */
               if (wft.method == SIMPLE) radio_buttons[0].set = TRUE;
               if (wft.method == GENERAL) radio_buttons[1].set = TRUE;

      	       create_dialog_box ("Unroll WFT", edit_labels, NULL,
			     XtNumber(edit_labels), "METHOD:", 
			     radio_buttons, XtNumber(radio_buttons),
			     NULL, NULL, 0, unroll_wft);
 	   } else 
		error ("This operation cannot be performed if more than one WFT is loaded");
      } else 
	    error ("Please read the automaton first");
}

	

void change_init_distr (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void		init_distr_ok ();
  char		str[50];

  if (wft_selected) { 
	wft_started = FALSE;
        wft_unrolled = FALSE;
        usize = 0; wft_depth = 0;
	sprintf (str, "Please enter the new initial distribution (nstates = %d)", wft.states); 
	display_dialog (str, init_distr_ok);
  } else {
	   error ("Please read the single wft  first");
  }
}

void init_distr_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmSelectionBoxCallbackStruct *call_data;
{
  char 	    *input_str, *s[MAX_STATES];
  int	    i;
  float     x;

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &input_str));
  XtUnmanageChild (dialog_box);
  
  if (input_str != NULL) {
        while (*input_str == ' ') input_str++;
	s[0] = input_str;
	for (i=1; i < wft.states; i++) { 
            while (*input_str != ' ' && *input_str != ',') input_str++;
            while (*input_str == ' ' || *input_str == ',') input_str++;
	    s[i] = input_str;
        }

	for (i=0; i < wft.states; i++) { 
	     /*printf ("i = %d, input string: %s\n", i, s[i]);*/
	     sscanf (s[i], "%f", &x);
             wft.state[i].initial = x;
	     /*printf ("value: %f\n", x);*/
        }
  }
}


void change_final_distr (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void		final_distr_ok ();
  char		str[50];

  if (wft_selected) { 
	wft_started = FALSE;
        wft_unrolled = FALSE;
        usize = 0; wft_depth = 0;
	sprintf (str, "Please enter the new final distribution (nstates = %d)", wft.states); 
	display_dialog (str, final_distr_ok);
  } else {
	   error ("Please read the single wft  first");
  }
}

void final_distr_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmSelectionBoxCallbackStruct *call_data;
{
  char 	    *input_str, *s[MAX_STATES];
  int	    i;
  float     x;

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &input_str));
  XtUnmanageChild (dialog_box);
  
  if (input_str != NULL) {
        while (*input_str == ' ') input_str++;
	s[0] = input_str;
	for (i=1; i < wft.states; i++) { 
            while (*input_str != ' ' && *input_str != ',') input_str++;
            while (*input_str == ' ' || *input_str == ',') input_str++;
	    s[i] = input_str;
        }

	for (i=0; i < wft.states; i++) { 
	     /*printf ("i = %d, input string: %s\n", i, s[i]);*/
	     sscanf (s[i], "%f", &x);
             wft.state[i].final = x;
	     /*printf ("value: %f\n", x);*/
        }
  }
}


void zoom_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmSelectionBoxCallbackStruct *call_data;
{
  char 	    *address;
  int	    output_size;
  void      save_prev_image ();

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &address));
  XtUnmanageChild (dialog_box);
  
  if (address != NULL) {
     save_prev_image ();
     transf_done = TRUE; 
     printf ("Zooming on the image...");
     if (color_flag) {
         if (crop_flag) {
             zoom_into_square (&crop_image_red, address,size, &output_size);
             zoom_into_square (&crop_image_green, address,size, &output_size);
             zoom_into_square (&crop_image_blue, address,size, &output_size);
         } else {
             zoom_into_square (&raw_image_red, address,size, &output_size);
             zoom_into_square (&raw_image_green,address, size, &output_size);
             zoom_into_square (&raw_image_blue,address, size, &output_size);
         }
     } else {
         if (crop_flag)
             zoom_into_square (&crop_image,address, size, &output_size);
         else
             zoom_into_square (&raw_image, address,size, &output_size);
     }
     rows = output_size; columns = output_size;
     printf ("Done\n"); fflush(stdout);
     size = output_size; depth = DDEPTH;
     if (!xv_flag) XtUnmanageChild (pop_window);
     display_image (FALSE);
 }
}


void wfa_zoom_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmSelectionBoxCallbackStruct *call_data;
{
  char 	    *address;
  float	    *init_distr_red, *init_distr_green, *init_distr_blue;
  float     *init_distr;
  float	    *old_init_distr, *old_init_distr_red, 
	    *old_init_distr_green, *old_init_distr_blue;
  int	    i;
  void      decode_aut ();

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &address));
  XtUnmanageChild (dialog_box);
  
   if (address == NULL) return;

   if (color_aut) {
       if (!eps_free (&wfa_red) || !eps_free (&wfa_green) || 
	   !eps_free(&wfa_blue)) {
		error ("The input WFA must be epsilon free!");
		return;
       }
    } else {
       if (!eps_free (&wfa)) { 
		error ("The input WFA must be epsilon free!");
		return;
       }
    }
 

   if (color_aut) {
	init_distr_red = alloc_vect (wfa_red.states);
	init_distr_green = alloc_vect (wfa_green.states);
	init_distr_blue = alloc_vect (wfa_blue.states);
	old_init_distr_red = alloc_vect (wfa_red.states);
	old_init_distr_green = alloc_vect (wfa_green.states);
	old_init_distr_blue = alloc_vect (wfa_blue.states);

	wfa_zoom_into_square (&wfa_red, address, &init_distr_red);
	wfa_zoom_into_square(&wfa_green, address, &init_distr_green);
	wfa_zoom_into_square(&wfa_blue, address, &init_distr_blue);

	for (i = 0; i < wfa_red.states; i++) {
	     old_init_distr_red[i] = wfa_red.state[i].initial;
	     wfa_red.state[i].initial = init_distr_red[i];
	}

	for (i = 0; i < wfa_green.states; i++) {
	     old_init_distr_green[i] = wfa_green.state[i].initial;
	     wfa_green.state[i].initial = init_distr_green[i];
	}

	for (i = 0; i < wfa_green.states; i++) {
	     old_init_distr_blue[i] = wfa_blue.state[i].initial;
	     wfa_blue.state[i].initial = init_distr_blue[i];
	}
	
	free (init_distr_red);
	free (init_distr_green);
	free (init_distr_blue);

     } else {

	init_distr = alloc_vect(wfa.states); 
	old_init_distr = alloc_vect(wfa.states); 

        wfa_zoom_into_square(&wfa, address, &init_distr);

        for (i = 0; i < wfa.states; i++) {
	     old_init_distr[i] = wfa.state[i].initial;
	     wfa.state[i].initial = init_distr[i];
	}
	free (init_distr);
      }

      adj_flag = FALSE; syn_flag = FALSE;
      decode_aut (DDEPTH);

      /* restore the original values */
      if (color_aut) {
	for (i = 0; i < wfa_red.states; i++) 
	     wfa_red.state[i].initial = old_init_distr_red[i];

	for (i = 0; i < wfa_green.states; i++) 
	     wfa_green.state[i].initial = old_init_distr_green[i];

	for (i = 0; i < wfa_blue.states; i++) 
	     wfa_blue.state[i].initial = old_init_distr_blue[i];

	free (old_init_distr_red);
	free (old_init_distr_green);
	free (old_init_distr_blue);
      } else {
	for (i = 0; i < wfa.states; i++) 
	     wfa.state[i].initial = old_init_distr[i];
	free (old_init_distr);
      }
}

void mchange (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{

  if (wft_selected && wft_load_type == SINGLE) { 
      if (wft_unrolled) {
          free_autom_values (&wft, usize);
	  wft_unrolled = FALSE;
      }
	wft_depth = 0;
        if (wft.method == SIMPLE) {
            wft.method = GENERAL;
	    set_label (mlabel, "method: general");
	} else {
            wft.method = SIMPLE;
	    set_label (mlabel, "method: simple");
        }
  } else 
	error ("Please read the automaton first");
}


void mchange_red (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{

  if (red_wft_selected) { 
	    if (red_wft_unrolled) {
                free_autom_values (&wft_red, usize);
		red_wft_unrolled = FALSE;
            }
	wft_depth = 0;
        if (wft_red.method == SIMPLE) {
            wft_red.method = GENERAL; 
	    set_label (mlabel_red, "method: general");
	} else {
            wft_red.method = SIMPLE;
	    set_label (mlabel_red, "method: simple");
        }
  } else 
	error ("Please read the automaton first");
}



void mchange_green (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{

  if (green_wft_selected) { 
	    if (green_wft_unrolled) {
                free_autom_values (&wft_green, usize);
		green_wft_unrolled = FALSE;
            }
	wft_depth = 0;
        if (wft_green.method == SIMPLE) {
            wft_green.method = GENERAL; 
	    set_label (mlabel_green, "method: general");
	} else {
            wft_green.method = SIMPLE;
	    set_label (mlabel_green, "method: simple");
        }
  } else 
	error ("Please read the automaton first");
}

void mchange_blue (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{

  if (blue_wft_selected) { 
	    if (blue_wft_unrolled) {
                free_autom_values (&wft_blue, usize);
		blue_wft_unrolled = FALSE;
            }
	wft_depth = 0;
        if (wft_blue.method == SIMPLE) {
            wft_blue.method = GENERAL; 
	    set_label (mlabel_blue, "method: general");
	} else {
            wft_blue.method = SIMPLE;
	    set_label (mlabel_blue, "method: simple");
        }
  } else 
	error ("Please read the automaton first");
}

void set_mode_im (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  mode = IMAGE;
  set_label (mode_label, "IMAGE MODE");
}


void set_mode_wfa (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  mode = WFA;
  set_label (mode_label, "WFA MODE");
}


void set_mode_wft (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  mode = WFT;
  set_label (mode_label, "WFT MODE");
}


void unroll_wfa (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void	decode_wfa (), set_scaling_flag (), set_cutoff_flag ();

  char	    *edit_labels[] = {"DEPTH:"};
  char		**text, str[20];
  toggle_button    toggle_buttons[] = 
		{"Scale colors", set_scaling_flag, FALSE,
		 "Cut off large values", set_cutoff_flag, FALSE};

    if (wfa_selected || (red_wfa_selected && green_wfa_selected && 
        blue_wfa_selected) || color_wfa_selected) {

        toggle_buttons[0].set = scaling_flag = FALSE;
        toggle_buttons[1].set = cutoff_flag = FALSE;
	if (wfa_depth > 0) {
            text = (char **) malloc (sizeof (char *));
	    sprintf (str, "%d", wfa_depth); text[0] = str;
        } else {
	    text = NULL;
        } 
	create_dialog_box ("Decode WFA", edit_labels, text,
			   XtNumber(edit_labels), NULL, NULL, 0, NULL,
			   toggle_buttons, XtNumber(toggle_buttons),
			   decode_wfa);
    } else 
      error ("Please read the automaton first");
}


/******************************************************************
 * Function: compose_wfa                                          *
 * 	     Compose WFA with an WFT (in that order)              *
 *           Decode the resulting WFA                             *
 ******************************************************************/

void compose_wfa (wfa, wft)
automaton *wfa, *wft;
{
      if ((wfa->states)*(wft->states) <= MAX_STATES) {
	 printf ("Doing composition.."); fflush (stdout);
	 compose (wfa, wft);
	 printf ("Done\n");
	 set_label (wfa_label, "Loaded WFA : composed WFA");
      } else 
	    error ("The composed automaton is too big");
}


void composition (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void 	load_wft2();

 	action_type = COMPOSITION;
  	if (wft_selected)
      	    display_file_sb ("Load the second WFT", 
			     WFT_MASK, load_wft2);
  	else
     	  error ("Please load the first wft before composition");
}



/* User selected ok to read the second wft */

void load_wft2 (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmFileSelectionBoxCallbackStruct *call_data;
{
  char *name;
  FILE *infile;
  int  ntapes = 2, read_ok, new_method, nstates;

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &name)) read_ok = TRUE;
  else read_ok = FALSE;
  XtUnmanageChild (dialog_box);
  
  if (read_ok) {
	infile = fopen (name, "r");
	if (infile == NULL) { 
	    error ("Can't open input wft file"); return;
        }
        if (!read_automaton (infile, &wft_aux, ntapes)) { 
		error ("Error in reading the input WFT file");
		fclose (infile);
		return;
	}
	fclose (infile);
	printf ("The automaton %s has been read successfully.\n", name);

	switch (action_type) {
		case SUM:     nstates = wft.states + wft_aux.states;
			      break;
		case CONCAT:  nstates = wft.states + wft_aux.states+1;
			      break;
		case COMPOSITION:
                              nstates = wft.states * wft_aux.states;
        		      break;
	}
        if (nstates <= MAX_STATES) {
	     if (wft.method == SIMPLE && wft_aux.method == SIMPLE)
		   new_method = SIMPLE;
	     else new_method = GENERAL;
             if (wft_unrolled) free_autom_values (&wft, usize);
	     switch (action_type) {
		case SUM:
	     		printf ("Doing addition.."); fflush (stdout);
	     		addwfa (&wft, &wft_aux);
	                printf ("Done\n"); fflush (stdout);
	                set_label (wft_label, "Loaded WFT : WFT after addition");
			break;
		case COMPOSITION:
	     		printf ("Doing composition.."); fflush (stdout);
	     		compose (&wft, &wft_aux);
	                set_label (wft_label, "Loaded WFT : composed WFT");
	                printf ("Done\n"); fflush (stdout);
			break;
		case CONCAT:
	     		printf ("Doing concatenation.."); fflush (stdout);
	     		if (!concatenate (&wft, &wft_aux)) {
			  error ("The first WFT is not in the correct form");
		          return;
		        }

	                printf ("Done\n"); fflush (stdout);
	                set_label (wft_label, "Loaded WFT : concatenated WFT");
			break;
	     }
	       printf ("Number of states : %d\n", wft.states);
  	       wft_selected = TRUE;
	       wft_started = FALSE;
	       wft_unrolled = FALSE;
               wft.method = new_method;
	       switch (wft.method) {
		case SIMPLE:
	             set_label (mlabel, "method: simple"); break;
	        case GENERAL:
	             set_label (mlabel, "method: general"); break;
               }
	       usize = 0; wft_depth = 0;
	} else
	     error ("The composed automaton is too big");
   }
}



void load_first (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmFileSelectionBoxCallbackStruct *call_data;
{
  char *name;
  FILE *infile;
  int  ntapes = 1, nstates, read_ok;
  void load_wfa2 ();

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &name)) read_ok = TRUE;
  else read_ok = FALSE;
  XtUnmanageChild (dialog_box);
  
  if (read_ok) {
	infile = fopen (name, "r");
	if (infile == NULL) { 
	    error ("Can't open input wft file"); return;
        }
        if (!read_automaton (infile, &wfa1_aux, ntapes)) { 
		error ("Error in reading the input WFA file");
		fclose (infile);
		return;
	}
	fclose (infile);
	printf ("The automaton %s has been read successfully.\n", name);
        display_file_sb ("Load the second automaton", WFA_MASK, load_wfa2);
  }
}

/* User selected ok to read the second wfa */

void load_wfa2 (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmFileSelectionBoxCallbackStruct *call_data;
{
  char *name;
  FILE *infile;
  int  ntapes = 1, nstates, read_ok;
  automaton wfa_aux;

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &name)) read_ok = TRUE;
  else read_ok = FALSE;
  XtUnmanageChild (dialog_box);
  
  if (read_ok) {
	infile = fopen (name, "r");
	if (infile == NULL) { 
	    error ("Can't open input wft file"); return;
        }
        if (!read_automaton (infile, &wfa_aux, ntapes)) { 
		error ("Error in reading the input WFA file");
		fclose (infile);
		return;
	}
	fclose (infile);
	printf ("The automaton %s has been read successfully.\n", name);
        switch (action_type) {
		case SUM: nstates = wfa.states+wfa_aux.states; 
			  break;
		case CONCAT: nstates = wfa.states+wfa_aux.states+1; 
			     break;
		case CPROD: nstates = wfa1_aux.states*wfa_aux.states; 
			    break;
        }

        if (nstates <= MAX_STATES) {
           wfa_depth = 0;
           switch (action_type) {
		case SUM :
	     	    printf ("Doing addition.."); fflush (stdout);
	     	    addwfa (&wfa, &wfa_aux); 
	            set_label (wfa_label, "Loaded WFA : WFA after addition");
  	    	    wfa_selected = TRUE;
		    adj_flag = FALSE; syn_flag = FALSE;
	    	    printf ("Done\n"); fflush (stdout);
		    printf ("Number of states : %d\n", wfa.states);
		    break;
		case CONCAT :
	     	    printf ("Doing concatenation.."); fflush (stdout);
	     	    if (!concatenate (&wfa, &wfa_aux)) {
			  error ("The first WFA is not in the correct form");
		          return;
		    }

	            set_label (wfa_label, "Loaded WFA : concatenated WFA");
  	    	    wfa_selected = TRUE;
		    adj_flag = FALSE; syn_flag = FALSE;
	    	    printf ("Done\n"); fflush (stdout);
		    printf ("Number of states : %d\n", wfa.states);
		    break;
		case CPROD:
		     if (cart_product (&wfa1_aux, &wfa_aux, &wfa)) {
	                 set_label (wfa_label, 
			            "Loaded WFA : cartesian product of two WFA");
  	    	         wfa_selected = TRUE;
		         adj_flag = FALSE; syn_flag = FALSE;
			 printf ("Number of states : %d\n", wfa.states);
                      } else
		         error ("Both WFA must have the same dimension!"); 
	    }
	} else
	     error ("The resulting automaton is too big");
  }
}


/*******************************************************
 *  WFA minimalization
 *******************************************************/

void min_wfa (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct *call_data;
{

  void	   wfa_min_ok ();
  char	   *text[2], str1[10], str2[10];
  char	   *edit_labels[] = {"Orthogalization error :", 
			     "Edge error :"};
			     

 if (wfa_selected || red_wfa_selected || green_wfa_selected || 
     blue_wfa_selected) {
    	sprintf (str1, "%6.4f", ON_ERROR); 
    	sprintf (str2, "%6.4f", EDGE_ERROR); 
    	text[0] = str1;
    	text[1] = str2;

    	create_dialog_box ("WFA minimalization", edit_labels, text,
				XtNumber(edit_labels), NULL, NULL, 0,
				NULL, NULL, 0, wfa_min_ok);
 } else
   error ("Please load the automaton first");
}

void wfa_min_ok (widget, dialog, call_data)
Widget		widget;
Widget		dialog;
XmAnyCallbackStruct *call_data;
{
     char	*input_str1, *input_str2;
     float	ortho_error, edge_error;

     /* read the approximation error */

     input_str1 = XmTextGetString (editor[0]);  
     input_str2 = XmTextGetString (editor[1]);  

     if (input_str1 != NULL) 
	sscanf (input_str1, "%f", &ortho_error);
     else
	ortho_error = ON_ERROR;

     if (input_str2 != NULL) 
	sscanf (input_str2, "%f", &edge_error);
     else
	edge_error = EDGE_ERROR;

     XtDestroyWidget (dialog);

     printf ("Start of the WFA minimalization algorithm ...\n");
     fflush (stdout);

     if (color_flag) {
	 printf ("Minimizing red component...\n"); fflush (stdout);
         if (minimize_wfa (&wfa_red, ortho_error, edge_error)) {
               printf ("Done\n"); fflush (stdout);
         } else 
	       error("Singular matrix in the WFA minimalization algorithm");

	 printf ("Minimizing green component...\n"); fflush (stdout);
         if (minimize_wfa (&wfa_green, ortho_error, edge_error)) {
               printf ("Done\n"); fflush (stdout);
         } else 
	       error("Singular matrix in the WFA minimalization algorithm");

	 printf ("Minimizing blue component...\n"); fflush (stdout);
         if (minimize_wfa (&wfa_blue, ortho_error, edge_error)) {
               printf ("Done\n"); fflush (stdout);
         } else 
	       error("Singular matrix in the WFA minimalization algorithm");
     } else {

         if (minimize_wfa (&wfa, ortho_error, edge_error)) {
             printf ("Done\n"); fflush (stdout);
         } else 
	   error("Singular matrix in the WFA minimalization algorithm");
     }
	
}



/**************************************************
 * FUNCTION: decode_aut                           *
 *           Decode the automaton and display     *
 *           the resulting image                  *
 **************************************************/

  
  void decode_aut (d)
  int d;
  {
      int i, new_size, u_size;
      void scale_img ();

      if (color_aut) {
	free_crop_image ();

         decode_automaton (&wfa_red, d);
         decode_automaton (&wfa_green, d);
         decode_automaton (&wfa_blue, d);
         for (u_size = 1, i=0; i<d; i++, u_size *= 2);

         printf ("Writing the output image.."); fflush (stdout);
         if (!image_flag) {
             orig_depth = (d < DDEPTH) ? DDEPTH : d;
	     orig_flag = COLOR;
             output_image (&orig_image_red, &wfa_red, size, 
                           &orig_size, d);
             output_image (&orig_image_green, &wfa_green, size, 
                           &orig_size, d);
             output_image (&orig_image_blue, &wfa_blue, size, 
                           &orig_size, d);
      	     printf ("Done\n"); fflush(stdout);
      	     orig_rows = orig_columns = orig_size;

	     raw_image_red = alloc_image (orig_size, orig_size);
	     copy_image (orig_image_red, raw_image_red, orig_size, orig_size);
	     raw_image_blue = alloc_image (orig_size, orig_size);
	     copy_image (orig_image_blue, raw_image_blue, orig_size, orig_size);
	     raw_image_green = alloc_image (orig_size, orig_size);
	     copy_image (orig_image_green, raw_image_green, orig_size, orig_size);

             rows = orig_rows; columns = orig_columns; 
             size = orig_size; depth = orig_depth;
             xtop = ytop = 0;

            } else {

               depth = (d < DDEPTH) ? DDEPTH : d;
               output_image (&raw_image_red, &wfa_red, size, &size, d); 
               output_image (&raw_image_green, &wfa_green, size, &size, d); 
               output_image (&raw_image_blue, &wfa_blue, size, &size, d); 

      	       printf ("Done\n"); fflush(stdout);
      	       rows = columns = size;
	       xtop = ytop = 0;

	       if (adj_flag) { 
		   scale_img (1.0/ScaleV, 1.0/ScaleH);
               } 
            }

            free_autom_values (&wfa_red, u_size);
            free_autom_values (&wfa_green, u_size);
            free_autom_values (&wfa_blue, u_size);
            color_flag = TRUE;

            /* display the output image */
            if (xv_flag) {
		xv_display();
            } else {
               if (image_flag) XtUnmanageChild (pop_window);
  	       if (create_color_image (raw_image_red, raw_image_green,
			        raw_image_blue , rows, columns)) {
                   resize_drawing_area (rows, columns);
                   XPutImage (XtDisplay (canvas), XtWindow (canvas), 
			theGC, image, 0, 0, 0, 0, columns, rows);
                   xv_disp_flag = FALSE;
                }
            }
            image_flag = TRUE;

       } else {
	     
           free_crop_image (); 
           decode_automaton (&wfa, d);
           for (u_size = 1, i=0; i<d; i++, u_size *= 2);
           printf ("Writing the output image.."); fflush (stdout);
           if (!image_flag) {
                 orig_depth = (d < DDEPTH) ? DDEPTH : d;
	       orig_flag = GREY;
               output_image (&orig_image, &wfa, size, 
                             &orig_size, d);
      	       printf ("Done\n"); fflush(stdout);
      	       orig_rows = orig_columns = orig_size;

		raw_image = alloc_image (orig_size, orig_size);
		copy_image (orig_image, raw_image, orig_size, orig_size);
                rows = orig_rows; columns = orig_columns; 
                size = orig_size; depth = orig_depth;
                xtop = ytop = 0;
           } else {
                 depth = (d < DDEPTH) ? DDEPTH : d;
                 output_image (&raw_image, &wfa, size, &size, d); 
      	         printf ("Done\n"); fflush(stdout);
      	         rows = columns = size;
	         xtop = ytop = 0;
	         if (adj_flag) { 
		   scale_img (1.0/ScaleV, 1.0/ScaleH);
                 }
           }

           free_autom_values (&wfa, u_size);
	   color_flag = FALSE;
           image_flag = TRUE;
	   transf_done = FALSE;

           /* display the output image */
           if (xv_flag) {
		xv_display ();
	   } else {
               if (image_flag) XtUnmanageChild (pop_window);
               resize_drawing_area (rows, columns);
               build_grey_table (gtable);
               create_grey_image (gtable, raw_image, rows, columns);
               XPutImage (XtDisplay (canvas), XtWindow (canvas), 
			 theGC, image, 0, 0, 0, 0, columns, rows);
               xv_disp_flag = FALSE;
           }
     }
}

void decode_wfa (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmSelectionBoxCallbackStruct *call_data;
{
  char *input_string;
  int  d;

/*
  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &input_string)); 
*/

  if ((input_string = XmTextGetString (editor[0])) != NULL) {
      sscanf (input_string, "%d", &d);
      XtFree (input_string);
      XtDestroyWidget (dialog_box);
      if (d > 0) { decode_aut(d); wfa_depth = d; }
  } else
      XtDestroyWidget (dialog_box);
}


void save_prev_image ()
{

    if (!color_flag) {

        if (prev_image != NULL) 
           free_image (prev_image, psize);  
        
	prev_image = alloc_image (size, size);

        if (crop_flag) 
	    copy_image (crop_image, prev_image, size, size);
         else 
	    copy_image (raw_image, prev_image, size, size);


     } else {

        if (prev_image_red != NULL) 
           free_image (prev_image_red, psize); 

        if (prev_image_blue != NULL) 
           free_image (prev_image_blue, psize); 

        if (prev_image_green != NULL) 
           free_image (prev_image_green, psize); 

	prev_image_red = alloc_image (size, size);
	prev_image_blue = alloc_image (size, size);
	prev_image_green = alloc_image (size, size);

        if (crop_flag) {
	    copy_image (crop_image_red, prev_image_red, size, size);
	    copy_image (crop_image_blue, prev_image_blue, size, size);
	    copy_image (crop_image_green, prev_image_green, size, size);
        } else {
	    copy_image (raw_image_red, prev_image_red, size, size);
	    copy_image (raw_image_blue, prev_image_blue, size, size);
	    copy_image (raw_image_green, prev_image_green, size, size);
        }

      }

	psize = size; prows = rows; pcolumns = columns; 
	pdepth = depth;
}

/************************************************
 * FUNCTION: apply_wft                          *
 *           Apply the wft to an image          *
 ************************************************/


void apply_wft (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmSelectionBoxCallbackStruct *call_data;
{
  char *input_string;
  int  i, j, d, output_size;
  void shrink_image (), zoom_image (), adjust_image ();

/*
  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &input_string)); 
*/
  save_prev_image ();

  if ((input_string = XmTextGetString (editor[0])) != NULL) {
      sscanf (input_string, "%d", &d);
      XtFree (input_string);
      XtDestroyWidget (dialog_box);

      if (d > 0) {
         if (d <= MAXDEPTH) {
         if (wft_load_type == SINGLE) {
            if (!power_of_2 (rows) || !power_of_2(columns))
		adjust_image ();

	    (void) flush_Xbuffer ();

            if (d != wft_depth) {
	       if (wft_unrolled) free_autom_values (&wft, usize);
	       decode_automaton (&wft, d);
            }
	    (void) flush_Xbuffer ();

            printf ("Writing the output image.."); 
            for (usize = 1, i=0; i < d; i++, usize *= 2);
	    fflush (stdout);
            if (color_flag) {
	        if (crop_flag) {
                    output_image (&crop_image_red, &wft, 
			          size, &output_size, d);
                    output_image (&crop_image_green, &wft, 
				  size, &output_size, d);
                    output_image (&crop_image_blue, &wft, 
				  size, &output_size, d);
                 } else {
                    output_image (&raw_image_red, &wft, size, 
				  &output_size, d);
                    output_image (&raw_image_green, &wft, size,
				  &output_size, d);
                    output_image (&raw_image_blue, &wft, size, 
				  &output_size, d);
                  }
              } else {
	          if (crop_flag) {
                      output_image (&crop_image, &wft, size, 
			            &output_size, d);
                  } else {
                      output_image (&raw_image, &wft, size, 
				    &output_size, d);
                  }
              }
	      (void) flush_Xbuffer ();
              wft_unrolled = TRUE;
	
	 } else {
              if (d != wft_depth) {
	         (void) flush_Xbuffer ();
	         if (red_wft_unrolled) free_autom_values (&wft_red, usize);
	         (void) flush_Xbuffer ();
	         if (red_wft_selected) {
                        decode_automaton (&wft_red, d);
			red_wft_unrolled = TRUE;
                 }
	         (void) flush_Xbuffer ();

	         if (green_wft_unrolled) free_autom_values (&wft_green, usize);
	         (void) flush_Xbuffer ();
	         if (green_wft_selected) {
			decode_automaton (&wft_green, d);
			green_wft_unrolled = TRUE;
                 }
	         (void) flush_Xbuffer ();

	         if (blue_wft_unrolled) free_autom_values (&wft_blue, usize);
	         (void) flush_Xbuffer ();
	         if (blue_wft_selected) {
			decode_automaton (&wft_blue, d);
			blue_wft_unrolled = TRUE;
                 }
	         (void) flush_Xbuffer ();
              }

              printf ("Writing the output image.."); 
	      fflush (stdout);

              for (usize = 1, i=0; i < d; i++, usize *= 2);
	      if (crop_flag) {
                  if (red_wft_selected) { 
                      output_image (&crop_image_red, &wft_red, 
				    size, &output_size, d);
		  } else {
		      output_size = (usize > DSIZE) ? usize:DSIZE;
		      if (output_size > size) 
			 zoom_image (&crop_image_red, size, output_size);
		      if (output_size < size)
			 shrink_image (&crop_image_red, size, output_size);
		  }
			
                  if (green_wft_selected) {
                      output_image (&crop_image_green, &wft_green,
				    size, &output_size, d);
		  } else {
		      output_size = (usize > DSIZE) ? usize:DSIZE;
		      if (output_size > size) 
			 zoom_image (&crop_image_green, size, output_size);
		      if (output_size < size)
			 shrink_image (&crop_image_green, size, output_size);
		  }

                  if (blue_wft_selected) {
                      output_image (&crop_image_blue, &wft_blue, 
				    size, &output_size, d);
		  } else {
		      output_size = (usize > DSIZE) ? usize:DSIZE;
		      if (output_size > size) 
			 zoom_image (&crop_image_blue, size, output_size);
		      if (output_size < size)
			 shrink_image (&crop_image_blue, size, output_size);
		  }
                } else {
                  if (red_wft_selected) { 
                      output_image (&raw_image_red, &wft_red, 
				    size, &output_size, d);
		  } else {
		      output_size = (usize > DSIZE) ? usize:DSIZE;
		      if (output_size > size) 
			 zoom_image (&raw_image_red, size, output_size);
		      if (output_size < size)
			 shrink_image (&raw_image_red, size, output_size);
		  }
			
                  if (green_wft_selected) {
                      output_image (&raw_image_green, &wft_green,
				    size, &output_size, d);
		  } else {
		      output_size = (usize > DSIZE) ? usize:DSIZE;
		      if (output_size > size) 
			 zoom_image (&raw_image_green, size, output_size);
		      if (output_size < size)
			 shrink_image (&raw_image_green, size, output_size);
		  }

                  if (blue_wft_selected) {
                      output_image (&raw_image_blue, &wft_blue, 
				    size, &output_size, d);
		  } else {
		      output_size = (usize > DSIZE) ? usize:DSIZE;
		      if (output_size > size) 
			 zoom_image (&raw_image_blue, size, output_size);
		      if (output_size < size)
			 shrink_image (&raw_image_blue, size, output_size);
		  }
             }
        }
        printf ("Done\n"); fflush(stdout);
	size = output_size;
	wft_depth = d;
	depth = (d < DDEPTH) ? DDEPTH : d;
        rows = columns = size;
        /*printf ("rows = %d, columns = %d, size = %d, unrolled depth = %d, unrolled size = %d.\n", 
		rows, columns, size, wft_depth, usize);
        */
	if (image_flag && !xv_flag) XtUnmanageChild (pop_window);

   	/* display the output image */
         xtop = 0; ytop = 0; rows = columns = size;
         display_image (FALSE);
        
    } else
	 error ("Maximum depth (9) exceeded");
    }
   } else
      XtDestroyWidget (dialog_box);
}
			
void unroll_wft (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmAnyCallbackStruct call_data;
{
  char *input_string, *filename;
  int  i, d;
  FILE *outfile;

  input_string = XmTextGetString (editor[0]);
  filename = XmTextGetString (editor[1]);

  if (filename != NULL) {
      if ((outfile = fopen (filename, "w")) == NULL) {
         error ("Cannot open the output file, using standard output");
      }
  } else
	outfile = NULL;

  if (input_string != NULL) {
      sscanf (input_string, "%d", &d);
      XtFree (input_string);
      XtDestroyWidget (dialog_box);

      if (d > 0) {
         if (d <= MAXDEPTH) {

          if (d != wft_depth) {
	     if (wft_unrolled) free_autom_values (&wft, usize);
	     decode_automaton (&wft, d);
          }
          wft_unrolled = TRUE;

          for (usize = 1, i=0; i < d; i++, usize *= 2);
	  output_matrix (outfile, &wft, usize, usize, d);

          if (filename != NULL)
              printf ("Sucessfully written %s.\n", filename);
          if (outfile != NULL) fclose(outfile);

          wft_depth = d;
       } else
	 error ("Maximum depth (9) exceeded");
     }
   } else
      XtDestroyWidget (dialog_box);
}

void iter_wft (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct *call_data;
{
  int  output_size;

  if (wft_started) {
      save_prev_image ();
      printf ("Writing the output image.."); fflush (stdout);
      if (wft_load_type == SINGLE) {
          if (color_flag) {
	      if (crop_flag) {
                    output_image (&crop_image_red, &wft, 
			          size, &output_size, wft_depth);
                    output_image (&crop_image_green, &wft, 
				  size, &output_size, wft_depth);
                    output_image (&crop_image_blue, &wft, 
				  size, &output_size, wft_depth);
                 } else {
                    output_image (&raw_image_red, &wft, size, 
				  &output_size, wft_depth);
                    output_image (&raw_image_green, &wft, size,
				  &output_size, wft_depth);
                    output_image (&raw_image_blue, &wft, size, 
				  &output_size, wft_depth);
                  }
              } else {
	          if (crop_flag) 
                      output_image (&crop_image, &wft, size, 
			            &output_size, wft_depth);
                  else 
                      output_image (&raw_image, &wft, size, 
				    &output_size, wft_depth);
              }
	
	 } else {
	      if (crop_flag) {
                  if (red_wft_selected) 
                      output_image (&crop_image_red, &wft_red, 
				    size, &output_size, wft_depth);
                  if (green_wft_selected) 
                      output_image (&crop_image_green, &wft_green,
				    size, &output_size, wft_depth);
                  if (blue_wft_selected) 
                      output_image (&crop_image_blue, &wft_blue, 
				    size, &output_size, wft_depth);
                } else {
                  if (red_wft_selected) 
                            output_image (&raw_image_red, &wft_red,  
					  size, &output_size, wft_depth);
                  if (green_wft_selected) 
                            output_image (&raw_image_green, &wft_green,
					  size, &output_size, wft_depth);
                  if (blue_wft_selected) 
                            output_image (&raw_image_blue, &wft_blue, 
					  size, &output_size, wft_depth);
                }
      }
	
        printf ("Done\n"); fflush(stdout);
	size = output_size;
/*        printf ("rows = %d, columns = %d, size = %d\n", 
		rows, columns, size); */
	if (!xv_flag) XtUnmanageChild (pop_window);

   /* display the output image */
         xtop = 0; ytop = 0; rows = columns = size;
         display_image (FALSE);
   } else
	error ("Please apply the WFT once before iterating");
}
			


/*****************************************************************
 * TRANSFORM CALLBACKS
 ****************************************************************/

void fliph (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void apply_transformation ();
  int cut_flag;

  if (image_flag) {
     apply_transformation ("fliph.t", depth);
     if (!xv_flag) XtUnmanageChild (pop_window);
      if (rows < size || columns < size) cut_flag = TRUE;
      else cut_flag = FALSE;
     display_image (cut_flag);
  } else
     error ("Please read the image first");
}


void flipv (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void apply_transformation ();
  int cut_flag;

  if (image_flag) {
     apply_transformation ("flipv.t", depth);
     if (!xv_flag) XtUnmanageChild (pop_window);
      if (rows < size || columns < size) cut_flag = TRUE;
      else cut_flag = FALSE;
     display_image (cut_flag);
  } else
     error ("Please read the image first");
}

void rotate (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  int cut_flag;
  void apply_transformation ();

  if (image_flag) {
     apply_transformation ("rot.t", depth);
     if (!xv_flag) XtUnmanageChild (pop_window);
      if (rows < size || columns < size) cut_flag = TRUE;
      else cut_flag = FALSE;
     display_image (cut_flag);
  } else
     error ("Please read the image first");
}


void crop (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  XImage *tmp_image;
  double   sz;
  int     i, j;

  if (xv_flag) {
      error ("This operation cannot be used with xv display");
      return;
  }

  if (rect_selected) {
     rect_selected = FALSE;
     odepth = depth; osize = size;
     tmp_image = XSubImage (image, subimg.x, subimg.y, 
			    subimg.width, subimg.height); 
     XtUnmanageChild (pop_window);
     resize_drawing_area (subimg.height, subimg.width);
     free (image);
     conv_image = tmp_image->data;
     image = XCreateImage (theDisplay,
			   DefaultVisual (theDisplay, theScreen),
			   DefaultDepth (theDisplay, theScreen),
			   ZPixmap, 0, tmp_image->data, 
                           subimg.width, subimg.height, 8, 0);

	/* display the output image */
      
      XPutImage (XtDisplay (canvas), XtWindow (canvas), theGC, 
                 image, 0, 0, 0, 0, subimg.width, subimg.height);
      free(tmp_image);
      sz = (double) (( subimg.height >= subimg.width ) ? 
             	       subimg.height : subimg.width );
      depth = ceil (log (sz)/ log ((double) 2.0));
      for (size = 1, i = 0; i < depth; i++, size *= 2);
      printf("Size of the cropped image = %d  columns x %d rows\n",
              subimg.width, subimg.height);
      printf("Depth = %d, size = %d\n",depth, size);
      
      if (color_flag) {
          if (crop_flag) {
	      raw_image_red = crop_image_red;
	      raw_image_green = crop_image_green;
	      raw_image_blue = crop_image_blue;
          }

	crop_image_red = alloc_image (size, size);
	crop_image_blue = alloc_image (size, size);
	crop_image_green = alloc_image (size, size);
/*
          printf ("xs = %d, ys = %d\n", subimg.x, subimg.y);
          printf ("xtop = %d, ytop = %d\n", xtop, ytop);
*/
          for (i = 0; i < size; i++) {
               for(j = 0; j < size; j++) {
                   if (i < subimg.height && j < subimg.width) {
                      crop_image_red[i][j] = 
			raw_image_red[ytop+i+subimg.y][xtop+j+subimg.x];
                      crop_image_green[i][j] = 
			raw_image_green[ytop+i+subimg.y][xtop+j+subimg.x];
                      crop_image_blue[i][j] = 
			raw_image_blue[ytop+i+subimg.y][xtop+j+subimg.x];
	           } else {
	               crop_image_red[i][j] = 0;
		       crop_image_green[i][j] = 0;
		       crop_image_blue[i][j] = 0;
                   }
                }
              }
       } else {
           if (crop_flag) raw_image = crop_image;
	   crop_image = alloc_image (size, size);
/*
           printf ("xs = %d, ys = %d\n", subimg.x, subimg.y);
           printf ("xtop = %d, ytop = %d\n", xtop, ytop);
*/
           for (i = 0; i < size; i++) {
               for(j = 0; j < size; j++) {
                  if (i < subimg.height && j < subimg.width)
                      crop_image [i][j] = raw_image[ytop+i+subimg.y]
                                              [xtop+j+subimg.x];
	          else
		      crop_image[i][j] = 0;
               }
           }
       }
	  
      oheight = rows; owidth = columns;
      rows = subimg.height; columns = subimg.width;
      xtop = 0; ytop = 0;
      crop_flag = TRUE;
      set_label (crop_label, "");
  } else 
      error ("Please select the crop rectangle first");
}


void uncrop (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
   int	i;

  if (xv_flag) {
      error ("This operation cannot be used with xv display");
      return;
  }

   if (crop_flag) {
	crop_flag = FALSE;
	if (color_flag) {
	    for (i=0; i < size; i++) {
		free (crop_image_red[i]);
		free (crop_image_green[i]);
		free (crop_image_blue[i]);
	    }
            free(crop_image_red); crop_image_red = NULL;
            free(crop_image_green); crop_image_green = NULL;
            free(crop_image_blue); crop_image_blue = NULL;
	} else {
	    for (i=0; i < size; i++) free (crop_image[i]);
            free(crop_image); crop_image = NULL;
        }
	rows = oheight; columns = owidth; 
        depth = odepth; size = osize;
        if (!xv_flag) XtUnmanageChild (pop_window);
        display_image (FALSE);
    }
}


void zoom (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void zoom_image ();
  int	i, j;
  FILE  *outfile;

  if (image_flag) {
     save_prev_image ();
     printf ("Zooming on the image...");
     if (color_flag) {
         if (crop_flag) {
             zoom_image (&crop_image_red, size, 2*size);
             zoom_image (&crop_image_green, size, 2*size);
             zoom_image (&crop_image_blue, size, 2*size);
         } else {
             zoom_image (&raw_image_red, size, 2*size);
             zoom_image (&raw_image_green, size, 2*size);
             zoom_image (&raw_image_blue, size, 2*size);
         }
     } else {
         if (crop_flag)
             zoom_image (&crop_image, size, 2*size);
         else
             zoom_image (&raw_image, size, 2*size);
     }
     rows *= 2; columns *= 2;
     printf ("Done\n"); fflush(stdout);
     size *= 2; depth++;
     transf_done = TRUE;
     if (!xv_flag) XtUnmanageChild (pop_window);
     display_image (FALSE);
  } else
     error ("Please read the image first");
}


void shrink (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void shrink_image ();

  if (image_flag) {
     save_prev_image ();
     printf ("Shrinking the image...");
     if (color_flag) {
         if (crop_flag) {
             shrink_image (&crop_image_red, size, size/2);
             shrink_image (&crop_image_green, size, size/2);
             shrink_image (&crop_image_blue, size, size/2);
         } else {
             shrink_image (&raw_image_red, size, size/2);
             shrink_image (&raw_image_green, size, size/2);
             shrink_image (&raw_image_blue, size, size/2);
         }
     } else {
         if (crop_flag)
             shrink_image (&crop_image, size, size/2);
         else
             shrink_image (&raw_image, size, size/2);
     }
     rows /= 2; columns /= 2;
     printf ("Done\n"); fflush(stdout);
     size /= 2; depth--;

     if (!xv_flag) XtUnmanageChild (pop_window);
     transf_done = TRUE;
     display_image (FALSE);

  } else
     error ("Please read the image first");
}


void change_contrast (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void  contrast_ok ();

  if (image_flag) 
     display_dialog ("Please select the contrast ( > 1 - lighter, < 1 - darker)", contrast_ok);
  else
     error ("Please read the image first");
}

  
void rot_image (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void  rot_image_ok (), set_resize_flag ();
/*
  char	*edit_labels[] = {"Rotation angle:"};
  toggle_button   toggle_buttons[] = 
		{"Resize after rotation", set_resize_flag, FALSE};
*/


    if (image_flag) {
/*
        toggle_buttons[0].set = resize_flag;
	create_dialog_box ("Rotate Image", edit_labels, NULL,
			   XtNumber(edit_labels), NULL, NULL, 0, NULL,
			   toggle_buttons, XtNumber(toggle_buttons),
			   rot_image_ok);
*/
	display_dialog ("Enter the rotation angle:", rot_image_ok);
     } else
       error ("Please read the image first");
}

void scale (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  void  scale_ok ();

  if (image_flag) 
     display_dialog ("Please enter the scaling factor", scale_ok);
  else
     error ("Please read the image first");
}


void resize_img (widget, client_data, call_data)
Widget		widget;
caddr_t		client_data;
XmAnyCallbackStruct call_data;
{
  char *labels[] = {"HEIGHT:", "WIDTH:"};
  void  resize_ok ();


  if (image_flag) 
     create_dialog_box ("Resize Image", labels, NULL, 
			XtNumber(labels), NULL, NULL, 0, 
			NULL, NULL, 0, resize_ok);
  else
     error ("Please read the image first");
}

/************ change the contrast of an image *************/

void contrast_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmSelectionBoxCallbackStruct *call_data;
{
  int    output_size;
  FILE   *infile;
  char *input_string;
  float contrast;
  void  scale_values ();

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &input_string)); 
  XtDestroyWidget (dialog_box);
  scaling_flag = FALSE;
  cutoff_flag = FALSE;

  if (input_string != NULL) {

    save_prev_image ();

    /* adjust the init. distribution as selected by the user */
    sscanf (input_string, "%f", &contrast);

    printf ("Changing contrast of the image ...");
    if (color_flag) {
	if (crop_flag) {
            scale_values (crop_image_red, contrast, size);
            scale_values (crop_image_green, contrast, size);
            scale_values (crop_image_blue, contrast, size);
         } else {
             scale_values (raw_image_red, contrast, size);
             scale_values (raw_image_green, contrast, size);
             scale_values (raw_image_blue, contrast, size);
         }
    } else {
	if (crop_flag) 
            scale_values (crop_image, contrast, size);
         else 
            scale_values(raw_image, contrast, size);
    }

    transf_done = TRUE;
    printf ("Done\n"); fflush(stdout);
    if (!xv_flag) XtUnmanageChild (pop_window);
    display_image (FALSE);
  }
}

/************** rotate the image *****************/
void rot_image_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmSelectionBoxCallbackStruct *call_data;
{
  void rotate_image ();
  char *input_string;
  float	angle, scaleV, scaleH;
  int   i, output_size, out_rows, out_cols;
  double sz;
  int   interp_flag = TRUE, disp_flag = FALSE;


  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &input_string)); 

  /*input_string = XmTextGetString (editor[0]);*/
  XtDestroyWidget (dialog_box);

  if (input_string != NULL) {
  	sscanf (input_string, "%f", &angle);
	XtFree (input_string);

     	save_prev_image ();
	while (angle >= 360.0) angle -= 360.0;

     	printf ("Rotating the image by %5.2f degrees.\n", angle);
	angle *=  PI / 180;

  	/* find size of the rotated image */
   	out_rows = rows * sin(angle) + columns * cos(angle);
   	out_cols = rows * cos(angle) + columns * sin(angle); 
   	sz = (double) (out_rows > out_cols) ? out_rows : out_cols;
   	depth = ceil (log (sz) / log ((double) 2.0));
   	for (output_size = 1, i = 0; i < depth; output_size *= 2, i++); 

	/*printf ("Output image: rows = %d, columns = %d, size = %d, depth = %d\n", out_rows, out_cols, output_size, depth);
*/
     	if (color_flag) {
            if (crop_flag) {
             	rotate_image (&crop_image_red, size, rows, columns,  
				output_size, out_rows, out_cols,
				angle, interp_flag);
             	rotate_image (&crop_image_green, size, rows, columns,  
				output_size, out_rows, out_cols,
				angle, interp_flag);
             	rotate_image (&crop_image_blue, size, rows, columns,  
				output_size, out_rows, out_cols,
				angle, interp_flag);
            } else {
             	rotate_image (&raw_image_red, size, rows, columns,  
				output_size, out_rows, out_cols,
				angle, interp_flag);
             	rotate_image (&raw_image_green, size, rows, columns,  
				output_size, out_rows, out_cols,
				angle, interp_flag);
             	rotate_image (&raw_image_blue, size, rows, columns,  
				output_size, out_rows, out_cols,
				angle, interp_flag);
            }
     	} else {
            if (crop_flag)
             	rotate_image (&crop_image, size, rows, columns, 
				output_size, out_rows, out_cols,
				angle, interp_flag);
            else
             	rotate_image (&raw_image, size, rows, columns, 
				output_size, out_rows, out_cols,
				angle, interp_flag);
     	}

/*
	if (resize_flag) {
	   printf("Resizing back to the original size...\n");
	   scaleV = (float) rows / (float) out_rows;
	   scaleH = (float) columns / (float) out_cols;
           size = output_size; rows = out_rows; columns = out_cols;
	   scale_img (scaleV, scaleH);
        } else {
           size = output_size; rows = out_rows; columns = out_cols;
        }
*/

        size = output_size; rows = out_rows; columns = out_cols;
	xtop = ytop = 0;
	transf_done = TRUE;
        printf ("Done\n"); fflush(stdout);
        if (!xv_flag) XtUnmanageChild (pop_window);
        /*if (rows < size || columns < size) disp_flag = TRUE;*/
        display_image (TRUE);
   }
}

/************** scale the image ******************/
void scale_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmSelectionBoxCallbackStruct *call_data;
{
  void scale_img ();
  char *input_string;
  float	factor;

  if (XmStringGetLtoR (call_data->value, XmSTRING_DEFAULT_CHARSET,
      &input_string)); 
  XtDestroyWidget (dialog_box);

  if (input_string != NULL) {
  	sscanf (input_string, "%f", &factor);

     	save_prev_image ();
	scale_img (factor, factor);
	transf_done = TRUE;
        if (!xv_flag) XtUnmanageChild (pop_window);
        display_image (FALSE);
  }
}

/*********** resize the image ***************/

void resize_ok (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmSelectionBoxCallbackStruct *call_data;
{
  void scale_img ();
  char *input_strh, *input_strw;
  int  height, width;
  float	scaleV, scaleH;

  input_strh = XmTextGetString (editor[0]);
  input_strw = XmTextGetString (editor[1]);
  XtDestroyWidget (dialog_box);

  if (input_strh != NULL && input_strw != NULL) {
  	sscanf (input_strh, "%d", &height);
  	sscanf (input_strw, "%d", &width);
	XtFree (input_strh);
	XtFree (input_strw);

	scaleV = (float) height / (float) rows;
	scaleH = (float) width / (float) columns;

        printf ("height: %d, width: %d, scaleH = %5.2f, scaleV = %5.2f\n", 
			height, width, scaleH, scaleV);

     	save_prev_image ();
	scale_img (scaleV, scaleH);
	transf_done = TRUE;
        if (!xv_flag) XtUnmanageChild (pop_window);
        display_image (FALSE);
  }
}

void scale_img (scaleV, scaleH)
float scaleV, scaleH;
{
  int   interp_flag = TRUE;
  int	i, output_size, out_rows, out_cols;
  double sz;
  void  scale_image ();

     	printf ("Scaling the image by %5.2f (horizontal) and %5.2f (vertical). \n", 
		scaleV, scaleH);

  	/* Compute the output size and depth */
  	out_rows = rows * scaleV + 0.5;
  	out_cols = columns * scaleH + 0.5;
  	sz = (double) (out_rows > out_cols) ? out_rows : out_cols;
  	depth = ceil (log (sz) / log ((double) 2.0));
  	for (output_size = 1, i = 0; i < depth; output_size *= 2, i++); 
/*
	printf ("Output image: rows = %d, columns = %d, size = %d, depth = %d.\n",
	  	out_rows, out_cols, output_size, depth);
*/


     	if (color_flag) {
            if (crop_flag) {
             	scale_image (&crop_image_red, size, rows, columns, 
			     output_size, out_rows, out_cols,
			     scaleV, scaleH, interp_flag);
             	scale_image (&crop_image_green, size, rows, columns, 
			     output_size, out_rows, out_cols,
			     scaleV, scaleH, interp_flag);
             	scale_image (&crop_image_blue, size, rows, columns, 
			     output_size, out_rows, out_cols,
			     scaleV, scaleH, interp_flag);
            } else {
             	scale_image (&raw_image_red, size, rows, columns, 
			     output_size, out_rows, out_cols,
			     scaleV, scaleH, interp_flag);
             	scale_image (&raw_image_green, size, rows, columns, 
			     output_size, out_rows, out_cols,
			     scaleV, scaleH, interp_flag);
             	scale_image (&raw_image_blue, size, rows, columns, 
			     output_size, out_rows, out_cols,
			     scaleV, scaleH, interp_flag);
            }
     	} else {
            if (crop_flag)
             	scale_image (&crop_image, size, rows, columns,
			     output_size, out_rows, out_cols,
			     scaleV, scaleH, interp_flag);
            else
             	scale_image (&raw_image, size, rows, columns,
			     output_size, out_rows, out_cols,
			     scaleV, scaleH, interp_flag);
     	}

        size = output_size; rows = out_rows; columns = out_cols;
        printf ("Done.\n"); fflush(stdout);
}

/************************************************************
 * FUNCTION: rotate_image ()
 *
 * This function performs actual rotation
 ************************************************************/

void rotate_image (raw_image, input_size, input_rows, input_columns, 
		   output_size, out_rows, out_cols, angle, interp_flag)
imgtype ***raw_image;
int	input_size, input_rows, input_columns, output_size,
	out_rows, out_cols;
float   angle;
int	interp_flag;
{
  int i, j, ctr_row, ctr_col, row_num, col_num, max_rows, max_columns;
  float row_addr, col_addr, row_delta, col_delta;
  float sinAngle, cosAngle, x, y;
  imgtype **output_image, pixel_value;
  imgtype	interpolate ();

   /* precompute sin and cos for rotation */
   sinAngle = sin ((double) angle);
   cosAngle = cos ((double) angle);

  /* allocate and initialize the output image */
  output_image = alloc_image (output_size, output_size);

  for (i = 0; i < output_size; i++) 
       for (j = 0; j < output_size; j++)
	   output_image[i][j] = 0;
  

  /**************************************************************
   * Actual rotation
   *
   * Code adapted from the book "Practical image processing in C"
   * by Craig A. Lindley
   *************************************************************/


   /* center of rotation */
   ctr_row = ytop + input_rows/2;
   ctr_col = xtop + input_columns/2; 

   /* Calculations from destination perspective */

   max_rows = input_rows + ytop;
   max_columns = input_columns + xtop;

   for (i = 0; i < out_rows; i++) {
	y = (float) (i - out_rows/2); 

	for (j = 0; j < out_cols; j++) { 
	     x = (float) (j - out_cols/2); 

	     /* rotate */
	     col_addr = x * cosAngle - y * sinAngle;
	     row_addr = x * sinAngle + y * cosAngle;
/*
		printf ("i = %d, j = %d, x = %f, y = %f, row_addr = %f, col_addr = %f\n", i, j, x, y, row_addr, col_addr); fflush (stdout);
*/
	     /* convert from relative distance to pixel location */
	     col_addr += ctr_col;
	     row_addr += ctr_row;

	     row_num = (int) row_addr;
	     row_delta = row_addr - row_num;
	     col_num = (int) col_addr;
	     col_delta = col_addr - col_num;
/*
	printf ("row_num = %d, col_num = %d, row_addr = %f, col_addr = %f\n", 
		row_num, col_num, row_addr, col_addr); fflush (stdout);
*/

	     if (row_num >= ytop && row_num < max_rows && 
		 col_num >= xtop && col_num < max_columns) {
	         if (interp_flag) 
	            pixel_value = interpolate (*raw_image, row_num, col_num, 
					       max_rows, max_columns, 
					       row_delta, col_delta);
	         else
	   	    pixel_value = (*raw_image)[row_num][col_num];
	
        	output_image[i][j] = pixel_value;
             }
        }
    }
	   

  /* free the old image and return the new one */
  for (i=0; i < input_size; i++)
         free ((*raw_image)[i]);
  free (*raw_image);
  *raw_image = output_image;
}


/************************************************************
 * FUNCTION: scale_image ()
 *
 * This function performs actual scaling
 ************************************************************/

void scale_image (raw_image, input_size, input_rows, input_columns, 
		  output_size, out_rows, out_cols, 
		  scaleV, scaleH, interp_flag)
imgtype ***raw_image;
int	input_size, input_rows, input_columns, 
	output_size, out_rows,out_cols;
float   scaleV, scaleH;
int	interp_flag;
{
  int i, j, row_num, col_num;
  float row_addr, col_addr, row_delta, col_delta;
  imgtype **output_image, pixel_value;
  imgtype	interpolate ();


  /* allocate and initialize the output image */
  output_image = alloc_image (output_size, output_size);

  for (i = 0; i < output_size; i++) 
       for (j = 0; j < output_size; j++)
	   output_image[i][j] = 0;
  

  /**************************************************************
   * Actual scaling 
   *
   * Code adapted form the book "Practical image processing in C"
   * by Craig A. Lindley
   *************************************************************/

   /* Calculations from destination perspective */

   for (i = 0; i < out_rows; i++) {
	row_addr = ((float) i) / scaleV;
	row_num = (int) row_addr;
	row_delta = row_addr - row_num;
	row_num += ytop;

	for (j = 0; j < out_cols; j++) { 
	     col_addr = ((float) j) / scaleH;
	     col_num = (int) col_addr;
	     col_delta = col_addr - col_num;
	     col_num += xtop;

	     if (interp_flag) 
	        pixel_value = interpolate (*raw_image, row_num, col_num, 
					   input_rows + ytop, input_columns + xtop, 
					   row_delta, col_delta);
	     else
	   	pixel_value = (*raw_image)[row_num][col_num];
	
        	output_image[i][j] = pixel_value;
        }
    }
	   

  /* free the old image and return the new one */
  for (i=0; i < input_size; i++)
         free ((*raw_image)[i]);
  free (*raw_image);
  *raw_image = output_image;

  xtop = 0; ytop = 0;
}

/********************************************************
 * FUNCTION interpolate ()
 *
 * Performs interpolation to enhance th transformed image
 * by: Craig A. Lindley
 ********************************************************/

 imgtype interpolate (image, row, col, nrows, ncols, row_delta, 
			    col_delta)
 imgtype **image;
 int	row, col, nrows, ncols;
 float  row_delta, col_delta;
 {
   imgtype	ptA, ptB, ptC, ptD;
   float	contribAB, contribCD, pix_val;
   
/*
   printf ("Now processing row = %d and col = %d (nrows = %d, ncols = %d)\n", 
		row, col, nrows, ncols); fflush (stdout);
*/

   ptA = image[row][col];

   if (row + 1 < nrows && col + 1 < ncols) {
	ptB = image[row][col+1];
	ptC = image[row+1][col];
	ptD = image[row+1][col+1];
   } else
	ptB = ptC = ptD = ptA;

   /*
    * Interpolate to find brightness contibution of each
    * pixel in neighborhood. Done in both horizontal and
    * vertical directions.
    */

    contribAB = col_delta * ((float) ptB - ptA) + ptA;
    contribCD = col_delta * ((float) ptD - ptC) + ptC;
    pix_val = 0.5 + contribAB + (contribCD - contribAB) * row_delta;
    

  return (imgtype) pix_val;
}

   
/****** General function for applying the wft transformation ******/

 void apply_transformation (name, d)
 int  d;
 char *name;
 {
  int    ntapes = 2, output_size;
  float  ratio;
  FILE   *infile;

    save_prev_image ();
    wft_aux.method = SIMPLE;
    scaling_flag = FALSE;
    cutoff_flag = FALSE;
    infile = fopen (name, "r");
    if (infile == NULL) { 
        error ("Can't open input wft file"); return;
    }
    if (!read_automaton (infile, &wft_aux, ntapes)) { 
		error ("Error in reading the auxiliary WFT file");
		fclose (infile);
		return;
    }
    fclose (infile);
    printf ("The automaton %s has been read successfully.\n", name);
    decode_automaton (&wft_aux, d);
    printf ("Writing the output image.."); fflush (stdout);
    if (color_flag) {
	if (crop_flag) {
            output_image (&crop_image_green, &wft_aux, size, &output_size, d);
            output_image (&crop_image_red, &wft_aux, size, &output_size, d);
            output_image (&crop_image_blue, &wft_aux, size, &output_size, d);
         } else {
             output_image (&raw_image_red, &wft_aux, size, &output_size, d);
             output_image (&raw_image_green, &wft_aux, size, &output_size, d);
             output_image (&raw_image_blue, &wft_aux, size, &output_size, d);
         }
    } else {
	if (crop_flag) 
            output_image (&crop_image, &wft_aux, size, &output_size, d);
         else 
            output_image (&raw_image, &wft_aux, size, &output_size, d);
    }
    free_autom_values (&wft_aux, output_size);
    free_automaton (&wft_aux);
    transf_done = TRUE;

    ratio =  ((float) output_size)/size;
    rows *= ratio; columns *= ratio;
    size = output_size;
/*
    printf ("ratio = %f, rows = %d, columns = %d, size = %d\n", 
	     ratio, rows, columns, size);
*/
    printf ("Done\n"); fflush(stdout);
}




/****************************************************************
 * Function:	zoom_image					*
 *              zooms on the input image                        *
 ****************************************************************/

void zoom_image (raw_image, input_size, output_size)
imgtype ***raw_image;
int	input_size, output_size;
{
  int	i, j, k, l, ix, iy, step;
  imgtype **output_image;

  output_image = alloc_image (output_size, output_size);

  /*printf ("input size = %d, output size = %d\n", input_size, output_size);*/
  step = output_size/input_size;
  for (i = 0; i < input_size; i++) {
       for (j = 0; j < input_size; j++) {
	    k = step*i; l = step*j;
	    for (ix = 0; ix < step; ix++)
	         for (iy = 0; iy < step; iy++)
	             output_image[k+ix][l+iy] = (*raw_image)[i][j];
        }
  }

  for (i=0; i < input_size; i++)
         free ((*raw_image)[i]);
  free (*raw_image);
  *raw_image = output_image;

}


void shrink_image (raw_image, input_size, output_size)
imgtype ***raw_image;
int	input_size, output_size;
{
  int	i, j, k, l, step, ix, iy;
  float s;
  imgtype **output_image, pix;

  output_image = alloc_image (output_size, output_size);

  step = input_size/output_size;
  printf ("step = %d\n", step);

  for (i = 0; i < output_size; i++) {
       for (j = 0; j < output_size; j++) {
	    k = step*i; l = step*j; s = 0;
	    for (ix = 0; ix < step; ix++) {
	         for (iy = 0; iy < step; iy++) {
		      pix = (*raw_image)[k+ix][l+iy];
		      s += (float) pix;
		 }
	    }
	    s /= step * step;
	    output_image[i][j] = (unsigned char) s;
        }
  }

  for (i=0; i < input_size; i++)
         free ((*raw_image)[i]);
  free (*raw_image);
  *raw_image = output_image;
}
   

                             
/**********************  for debugging ***********************/
void save_grey (image, size)
unsigned char **image;
int	size;
{
  FILE   *outfile;
  int    i, j;

  outfile = fopen ("out", "w");
  fprintf (outfile, "P5\n");
  fprintf (outfile, "%d %d\n", size, size);
  fprintf (outfile, "%d\n", 255);
  for (i=0; i<size; i++)
       for (j=0; j<size; j++) 
	    fputc (imgtype_to_char(image[i][j]), outfile);
}

void save_color (name,image_red, image_green, image_blue, size)
char *name;
unsigned char **image_red, **image_green, **image_blue;
int	size;
{
  FILE   *outfile;
  int    i, j;

  outfile = fopen (name, "w");
  fprintf (outfile, "P6\n");
  fprintf (outfile, "%d %d\n", size, size);
  fprintf (outfile, "%d\n", 255);
  for (i=0; i<size; i++) {
       for (j=0; j<size; j++) {
	    fputc (imgtype_to_char(image_red[i][j]), outfile);
	    fputc (imgtype_to_char(image_green[i][j]), outfile);
	    fputc (imgtype_to_char(image_blue[i][j]), outfile);
       }
  }
}

int max (a,b,c)
int a,b,c;
{
   if (a >= b) {
      if (a >= c) return a;
      else return c;
   } else {
      if (b >= c) return b;
      else return c;
   }
}


/******* scaling flag and resize flag setting callback ******/

void set_scaling_flag (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmToggleButtonCallbackStruct *call_data;
{
  if (call_data->set) scaling_flag = TRUE;
  else scaling_flag = FALSE;
}

void set_cutoff_flag (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmToggleButtonCallbackStruct *call_data;
{
  if (call_data->set) cutoff_flag = TRUE;
  else cutoff_flag = FALSE;
}

void set_resize_flag (widget, dialog_box, call_data)
Widget		widget;
Widget		dialog_box;
XmToggleButtonCallbackStruct *call_data;
{
  if (call_data->set) resize_flag = TRUE;
  else resize_flag = FALSE;
}

/********** Method change toggle button callback *******/

void change_method (widget, method, call_data)
Widget		widget;
int		method;
XmToggleButtonCallbackStruct *call_data;
{
  if (call_data->set) {
     /* must reset wft depth to unroll WFT properly ! */
     if (wft.method != method) {
	 if (method == SIMPLE)
             set_label (mlabel,"method: simple");
         if (method == GENERAL)
             set_label (mlabel,"method: general");

	 if (wft_unrolled) {
		free_autom_values (&wft, usize);
		wft_unrolled = FALSE;
	 }
         wft_depth = 0;
         wft.method = method;
     }
  }
}




void free_crop_image ()
{
  int i;

         if (crop_flag && color_flag) {
	     crop_flag = FALSE;
	     for (i=0; i < size; i++) free (crop_image_red[i]);
	     for (i=0; i < size; i++) free (crop_image_green[i]);
	     for (i=0; i < size; i++) free (crop_image_blue[i]);
             free(crop_image_red); crop_image_red = NULL;
             free(crop_image_green); crop_image_green = NULL;
             free(crop_image_blue); crop_image_blue = NULL;
             size = osize;
         } else if (crop_flag && !color_flag) {
	       crop_flag = FALSE;
	       for (i=0; i < size; i++) free (crop_image[i]);
               free(crop_image); crop_image = NULL;
               size = osize;
         }
}

void scale_values (raw_image, sc_factor, size)
imgtype		**raw_image;
float		sc_factor;
int		size;
{
   int		i,j;
   float	x;

   for (i=0; i<size; i++) {
	for (j=0; j<size; j++) {
		raw_image[i][j] *= sc_factor;
	}
    }
}
		
void clear_wft_labels ()
{
    set_label (blue_label, ""); set_label (mlabel_blue, "");
    set_label (green_label, ""); set_label (mlabel_green, "");
    set_label (red_label, ""); set_label (mlabel_red, "");
}

void clear_wfa_labels ()
{
    set_label (blue_wfa_label, ""); 
    set_label (green_wfa_label, ""); 
    set_label (red_wfa_label, ""); 
}

void adjust_image ()
{
  int		sz, d;

  sz = size; d = depth;
  /*printf ("rows = %d, columns = %d, size = %d\n", rows, columns, size);*/
  find_size (rows, columns, &sz, &d);
  ScaleV = (float) sz/ (float) rows;
  ScaleH = (float) sz/ (float) columns;
  /*printf ("ScaleV = %6.4f, ScaleH = %6.4f\n", ScaleV, ScaleH);*/
  scale_img (ScaleV, ScaleH);
}
  

void find_size (rows, columns, size, depth)
int	rows, columns, *size, *depth;
{
  float  dx1, dx2, dy1, dy2;

  dx1 = fabs((float)rows - (float)*size/2);
  dx2 = *size - rows;
  dy1 = fabs((float)columns - (float)*size/2);
  dy2 = *size - columns;

  if (*size <= MAXSIZE) {
     if (dx1 < dx2 && dy1 < dy2) { *size /= 2; (*depth)--; }
  } else {
	*size = MAXSIZE; *depth = MAXDEPTH;
  }
}

int power_of_2 (num)
int	num;
{ 
  int   n;

  for (n = num; n > 1; n /= 2) {
       if (n%2) return (FALSE);
  }
  return (TRUE);
}

unsigned char imgtype_to_char (x)
imgtype x;
{

  if (x > 255) x = 255;
  if (x < 0) x = 0;
  /*printf ("x = %f, converted to %d\n", x,c);*/
  return((unsigned char) x);
}

